%% ========================== -*-Erlang-*- =============================
%% Core Erlang grammar for the Yecc parser generator
%%
%% Copyright (C) 1999-2001 Richard Carlsson
%%
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2 of the
%% License, or (at your option) any later version.
%%
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
%% Lesser General Public License for more details.
%%
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
%% USA
%%
%% Author contact: richardc@csd.uu.se
%%
%% $Id: cerl_parser.yrl,v 1.1.1.1 2010/02/26 15:26:33 svenolof Exp $
%%
%% ====================================================================

Nonterminals

annotated_clause annotated_clauses annotated_fun annotated_function_name
annotated_module annotated_single_expression
annotated_single_expressions annotated_value_list annotation_list
application atomic_literal attribute attribute_list binary
binary_segment binary_segments case_expr catch_expr clause constant
constants expression expressions fun_expr function_definition
function_definitions function_name function_names guard intermodule_call
let_expr letrec_expr list list_constant list_constant_tail list_pattern
list_pattern_tail list_tail module_attributes module_body
module_definition module_exports module_header nil other_pattern pattern
pattern_list patterns primop_call receive_expr sequencing
single_expression targets timeout try_expr tuple tuple_constant
tuple_pattern value_list variable variables.

Terminals

%% Separators

'(' ')' '{' '}' '[' ']' '<' '>' ',' ':' '|' '/' '=' '#' '->' '-|'

%% Keywords (atoms are assumed to always be single-quoted).

'after' 'apply' 'attributes' 'call' 'case' 'catch' 'do' 'end' 'fun' 'in'
'let' 'letrec' 'module' 'of' 'primop' 'receive' 'try' 'when'

%% Literal tokens (provided by the tokeniser):

atom char float integer string var.


%% Declare the start rule for parsing

Rootsymbol annotated_module.


%% Grammar

annotated_module ->
    module_definition:
	'$1'.
annotated_module ->
    '(' module_definition '-|' annotation_list ')':
        set_ann('$2', concrete_list('$5')).

annotation_list ->
    '[' ']': [].
annotation_list ->
    '[' constants ']': lists:reverse('$2').

module_definition ->
    'module' atom module_header module_body 'end':
        {Es, As} = '$3',
        c_module(c_atom(tok_val('$2')), Es, As, '$4').

%% Produced in reverse order.

constants ->
    constant: ['$1'].
constants ->
    constants ',' constant: ['$3' | '$1'].

module_header ->
    module_exports module_attributes:
        {'$1', '$2'}.

module_exports ->
    '[' ']': [].
module_exports ->
    '[' function_names ']': lists:reverse('$2').

%% Produced in reverse order.

function_names ->
    function_name: ['$1'].
function_names ->
    function_names ',' function_name: ['$3' | '$1'].

function_name ->
    atom '/' integer:
        c_fname(tok_val('$1'), tok_val('$3')).

module_attributes ->
    'attributes' '[' ']': [].
module_attributes ->
    'attributes' '[' attribute_list ']': lists:reverse('$3').

%% Produced in reverse order.

attribute_list ->
    attribute: ['$1'].
attribute_list ->
    attribute_list ',' attribute: ['$3' | '$1'].

attribute ->
    atom '=' constant:
        {c_atom(tok_val('$1')), '$3'}.

module_body ->
    '$empty': [].
module_body ->
    function_definitions: lists:reverse('$1').

%% Produced in reverse order.

function_definitions ->
    function_definition: ['$1'].
function_definitions ->
    function_definitions function_definition: ['$2' | '$1'].

function_definition ->
    annotated_function_name '=' annotated_fun:
        {'$1', '$3'}.

annotated_function_name ->
    function_name: '$1'.
annotated_function_name ->
    '(' function_name '-|' annotation_list ')':
        set_ann('$2', concrete_list('$4')).

annotated_fun ->
    fun_expr: '$1'.
annotated_fun ->
    '(' fun_expr '-|' annotation_list ')':
        set_ann('$2', concrete_list('$4')).

constant -> atomic_literal: '$1'.
constant -> tuple_constant: '$1'.
constant -> list_constant: '$1'.

tuple_constant ->
    '{' '}':
        c_tuple([]).
tuple_constant ->
    '{' constants '}':
        c_tuple(lists:reverse('$2')).

list_constant ->
    '[' constant list_constant_tail:
	c_cons('$2', '$3').

list_constant_tail ->
    ']':
        c_nil().
list_constant_tail ->
    '|' constant ']':
        '$2'.
list_constant_tail ->
    ',' constant list_constant_tail:
        c_cons('$2', '$3').

atomic_literal -> integer: c_int(tok_val('$1')).
atomic_literal -> float: c_float(tok_val('$1')).
atomic_literal -> atom: c_atom(tok_val('$1')).
atomic_literal -> nil: '$1'.
atomic_literal -> char: c_char(tok_val('$1')).
atomic_literal -> string: c_string(tok_val('$1')).

nil ->
    '[' ']': c_nil().

variable ->
    var: c_var(tok_val('$1')).
variable ->
    '(' var '-|' annotation_list ')':
        set_ann(c_var(tok_val('$2')), concrete_list('$4')).

pattern ->
    variable: '$1'.
pattern ->
    other_pattern: '$1'.
pattern ->
    '(' other_pattern '-|' annotation_list ')':
        set_ann('$2', concrete_list('$4')).

other_pattern -> atomic_literal: '$1'.
other_pattern -> tuple_pattern: '$1'.
other_pattern -> list_pattern: '$1'.
other_pattern -> variable '=' pattern: c_alias('$1', '$3').

%% Produced in reverse order.

patterns ->
    pattern: ['$1'].
patterns ->
    patterns ',' pattern: ['$3' | '$1'].

tuple_pattern ->
    '{' '}':
        c_tuple([]).
tuple_pattern ->
    '{' patterns '}':
        c_tuple(lists:reverse('$2')).

list_pattern ->
    '[' pattern list_pattern_tail:
	c_cons('$2', '$3').

list_pattern_tail ->
    ']':
        c_nil().
list_pattern_tail ->
    '|' pattern ']':
        '$2'.
list_pattern_tail ->
    ',' pattern list_pattern_tail:
        c_cons('$2', '$3').

expression ->
    annotated_value_list: '$1'.
expression ->
    annotated_single_expression: '$1'.

annotated_value_list ->
    value_list: '$1'.
annotated_value_list ->
    '(' value_list '-|' annotation_list ')':
        set_ann('$2', concrete_list('$4')).

value_list ->
    '<' '>':
	c_values([]).
value_list ->
    '<' annotated_single_expressions '>':
	c_values(lists:reverse('$2')).

%% Produced in reverse order.

annotated_single_expressions ->
    annotated_single_expression:
	['$1'].
annotated_single_expressions ->
    annotated_single_expressions ',' annotated_single_expression:
	['$3' | '$1'].

annotated_single_expression ->
    single_expression: '$1'.
annotated_single_expression ->
    '(' single_expression '-|' annotation_list ')':
        set_ann('$2', concrete_list('$4')).

single_expression -> atomic_literal: '$1'.
single_expression -> var: c_var(tok_val('$1')).
single_expression -> function_name: '$1'.
single_expression -> tuple: '$1'.
single_expression -> list: '$1'.
single_expression -> binary: '$1'.
single_expression -> let_expr: '$1'.
single_expression -> case_expr: '$1'.
single_expression -> fun_expr: '$1'.
single_expression -> letrec_expr: '$1'.
single_expression -> application: '$1'.
single_expression -> intermodule_call: '$1'.
single_expression -> primop_call: '$1'.
single_expression -> try_expr: '$1'.
single_expression -> receive_expr: '$1'.
single_expression -> sequencing: '$1'.
single_expression -> catch_expr: '$1'.

%% Produced in reverse order.

expressions ->
    expression: ['$1'].
expressions ->
    expressions ',' expression: ['$3' | '$1'].

tuple ->
    '{' '}':
        c_tuple([]).
tuple ->
    '{' expressions '}':
        c_tuple(lists:reverse('$2')).

list ->
    '[' expression list_tail:
	c_cons('$2', '$3').

list_tail ->
    ']':
        c_nil().
list_tail ->
    '|' expression ']':
        '$2'.
list_tail ->
    ',' expression list_tail:
        c_cons('$2', '$3').

binary ->
    '#' '{' binary_segments '}':
	c_binary(lists:reverse('$2')).

%% Produced in reverse order.

binary_segments ->
    binary_segment: ['$1'].
binary_segments ->
    binary_segments binary_segment: ['$2' | '$1'].

binary_segment ->
    expression ':' expression '[' ']':
	{'$1', '$3', []}.
binary_segment ->
    expression ':' expression '[' expressions ']':
	{'$1', '$3', '$5'}.

let_expr ->
    'let' targets '=' expression 'in' expression:
        c_let('$2', '$4', '$6').

targets ->
    variable: ['$1'].
targets ->
    '<' '>': [].
targets ->
    '<' variables '>': lists:reverse('$2').

%% Produced in reverse order.

variables ->
    variable: ['$1'].
variables ->
    variables ',' variable: ['$3' | '$1'].

case_expr ->
    'case' expression 'of' annotated_clauses 'end':
        c_case('$2', lists:reverse('$4')).

%% Produced in reverse order.

annotated_clauses ->
    annotated_clause: ['$1'].
annotated_clauses ->
    annotated_clauses annotated_clause: ['$2' | '$1'].

annotated_clause ->
    clause: '$1'.
annotated_clause ->
    '(' clause '-|' annotation_list ')':
        set_ann('$2', concrete_list('$4')).

clause ->
    pattern_list guard '->' expression:
        c_clause('$1', '$2', '$4').

pattern_list ->
    pattern: ['$1'].
pattern_list ->
    '<' '>': [].
pattern_list ->
    '<' patterns '>': lists:reverse('$2').

guard ->
    'when' expression: '$2'.

fun_expr ->
    'fun' '(' ')' '->' expression:
        c_fun([], '$5').
fun_expr ->
    'fun' '(' variables ')' '->' expression:
        c_fun(lists:reverse('$3'), '$6').

letrec_expr ->
    'letrec' 'in' expression:
        c_letrec([], '$3').
letrec_expr ->
    'letrec' function_definitions 'in' expression:
        c_letrec(lists:reverse('$2'), '$4').

application ->
    'apply' expression '(' ')':
        c_apply('$2', []).
application ->
    'apply' expression '(' expressions ')':
        c_apply('$2', lists:reverse('$4')).

intermodule_call ->
    'call' expression ':' expression '(' ')':
        c_call('$2', '$4', []).
intermodule_call ->
    'call' expression ':' expression '(' expressions ')':
        c_call('$2', '$4', lists:reverse('$6')).

primop_call ->
    'primop' atom '(' ')':
        c_primop(c_atom(tok_val('$2')), []).
primop_call ->
    'primop' atom '(' expressions ')':
        c_primop(c_atom(tok_val('$2')), lists:reverse('$4')).

%% The only correct form for "try" is with exactly two variables,
%% but we let the parser accept any number.
%% try_expr ->
%%     'try' expression 'catch' '(' ')' '->' expression:
%%         c_try('$2', [], '$7').
%% try_expr ->
%%     'try' expression 'catch' '(' variables ')' '->' expression:
%%         c_try('$2', lists:reverse('$5'), '$8').

try_expr ->
    'try' expression 'of' '<' variables '>' '->' expression
        'catch' '<' variables '>' '->' expression:
        c_try('$2', lists:reverse('$5'), '$8', '$11', '$14').


receive_expr ->
    'receive' timeout:
        {T, A} = '$2',
        c_receive([], T, A).
receive_expr ->
    'receive' annotated_clauses timeout:
        {T, A} = '$3',
        c_receive(lists:reverse('$2'), T, A).

timeout ->
    'after' expression '->' expression: {'$2', '$4'}.

sequencing ->
    'do' expression expression:
        c_seq('$2', '$3').

catch_expr ->
    'catch' expression:
        c_catch('$2').


%% =====================================================================

Erlang code.

%% =====================================================================
%% Core Erlang parser, generated by the Yecc parser generator.
%%
%% Copyright (C) 1999-2001 Richard Carlsson
%%
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2 of the
%% License, or (at your option) any later version.
%%
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
%% Lesser General Public License for more details.
%%
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
%% USA
%%
%% Author contact: richardc@csd.uu.se
%%
%% $Id: cerl_parser.yrl,v 1.1.1.1 2010/02/26 15:26:33 svenolof Exp $
%%
%% =====================================================================

-import(cerl, [c_atom/1, c_var/1, c_fname/2, c_char/1, c_float/1,
	       c_int/1, c_string/1, c_nil/0, c_tuple/1, c_cons/2,
	       c_alias/2, c_let/3, c_letrec/2, c_case/2, c_clause/3,
	       c_fun/2, c_apply/2, c_call/3, c_try/5, c_primop/2,
	       c_receive/3, c_module/4, concrete/1, c_seq/2,
	       c_catch/1, c_values/1, c_binary/1, set_ann/2]).

tok_val(T) ->
    element(3, T).

concrete_list(Xs) ->
    [concrete(X) || X <- Xs].


%% =====================================================================
