%%% File    : analyze.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : New, brave user interface to analysis
%%% Created : 27 Jan 2000 by Sven-Olof Nystr|m <svenolof@it.uu.se>

-module(analyze).
-author('svenolof@harpo.it.uu.se').

-compile(export_all).
-export([check_module/1, check_module/2, check_function/3,  set_path/1]).
-export([type_module/1, type_module/2, type_core/1]).

-define(prop, proplists). % This library module is either called
                           % proplists or property_lists,

% Typechecker assumes that there are files Module.erl and Module.spec.
% For the modules in ProgramModuleList there should of course be
% corresponding .erl modules and for the spec modules .spec files.

% Need to set path first. Something like this...
% analyze:set_path(["../Examples", "../StandardSpec",
%                   "/home/harpo/csd/svenolof/ErlangExamples"]).

check_module(Module) ->
    analyze([Module], spec, {module, Module}, [optimistic, time]).

check_module(Module, Options) ->
    analyze([Module], spec, {module, Module}, 
            Options ++ [optimistic, time]).
     
check_function(M,N,A) ->
    Fnid = program:make_fnid(M, N, A),
    analyze([M], spec, {function, Fnid}, [time]).


% type_module computes type information from a module, without the aid
% of specifications.

type_module(Module) ->
    analyze([Module], default, {module, Module}, 
            [{safe, false}, {external_specs, false}, time]).
type_module(Module, Options) ->
    ExternalSource = ?prop:get_value(external_source, 
					 Options++[{external_source, []}]),
    analyze([Module|ExternalSource], default, {module, Module}, 
	    Options ++ [{safe, false}, {external_specs, false}, time]).

    

% analyze, the generic analysis

analyze(ModuleList, Mode, Root, Options) ->
    program:init(),
    context:init(),
    warn_filter:init(),
    Include = ?prop:get_value(include, 
				  Options++[{include, []}]),
    program:load_erl_module_list(ModuleList, Include),
    load_standard_spec_modules(),
    AS = as:make_state(),
    AS1 = insert_options(Options ++ 
                         [safe, spec, external_specs, 
                          {external_source, []}, 
                          {polyvariance, default}, {max_contexts, 100}],
			 AS),
    time(init,AS1),

    AS2 = analyze_root(Root, Mode, AS1),
    io:format("~n~s",[time(after_analysis, AS2)]),
    report(AS2),
    present(Root, AS2),
    io:format("~n~s",[time(after_presentation, AS2)]),
%     lists:foreach(fun({Key,Val}) ->
% 			  io:format("~n~p~n~p~n", [Key, Val])
% 		  end,
% 		  gb_trees:to_list(element(2, AS2))),
    true.
%    precision:save_precision(hd(ModuleList), AS2).
       

analyze_root({function, Fnid}, Mode, AS) ->
    analyze_function(Fnid, Mode, AS);
analyze_root({module, M}, Mode, AS) ->
    analyze_module(M, Mode, AS).

analyze_module(M, Mode, AS) ->
    io:format("~nAM ~w", [get(polyvariance)]),

    FnidList = module_fnid_list(M, Mode),
    lists:foldl(fun(Fnid, BS) ->
			io:format("~nAnalyzing Function ~s",[program:show_fnid(Fnid)]),
			analyze_function(Fnid, Mode, BS)
		end,
		AS,
		FnidList).

module_fnid_list(Module, spec) ->
    program:load_mod(Module, spec),
    SpecList = program:module_spec_list(Module),
    lists:map(fun({Fnid,_}) -> Fnid end, SpecList);
module_fnid_list(Module, default) ->
    F = fun(Fref) ->
		program:make_fnid(Module, 
				  kinds:fref_name(Fref), 
				  kinds:fref_arity(Fref))
	end,
    lists:map(F, program:export_list(Module)).

analyze_function(Fnid, spec, AS) ->
    Context =  context:call(Fnid, start, context:empty()),
    {_, AS1} = work:call_top_level(Fnid, start, Context, AS),
    work:iter(AS1);
analyze_function(Fnid, default, AS) ->
    ArgList = arglist(program:fnid_arity(Fnid)),
    {Context, AS0} = work:call_analysis(Fnid, 0, context:empty(), AS),
    AS1 = param:insert_arg_list(Context, ArgList, AS0),
    Self = term:make_pid(Context),
    AS2 = as:insert(tag:self(Context), Self, AS1),
    work:iter(AS2).


%% Given a Core Erlang module, produce type information

type_core(Core) ->
    program:init(),
    context:init(),
    warn_filter:init(),
    Module = program:install_core(Core),
    FnidList = module_fnid_list(Module, default),
    load_standard_spec_modules(),
    AS = as:make_state(),
    AS1 = insert_options([{safe, false}, {external_specs, false}], AS),
    AS2 = lists:foldl(fun(Fnid, BS) ->
			      analyze_function(Fnid, default, BS)
		      end,
		      AS1,
		      FnidList),
    present:generate_internal(AS2).

%% reporting


report(AS) ->
    report:present(AS).

present({module, M}, AS) ->
    present:present(M, AS);
present({function, F}, AS) ->
    present:present_function(F, AS).


insert_options(Options, AS) ->
    Safety = ?prop:get_bool(safe, Options),
    UseSpec = ?prop:get_bool(spec, Options),
    Time = ?prop:get_bool(time, Options),
    ExternalSpecs = ?prop:get_bool(external_specs, Options),
    Optimistic = ?prop:get_bool(optimistic, Options),
    Quiet = ?prop:get_bool(quiet, Options),
    Poly = ?prop:get_value(polyvariance, Options),
    CFA_K = ?prop:get_value(cfa_k, Options),
    Max = ?prop:get_value(max_contexts, Options),

    AS1 = as:dict_put(use_spec, UseSpec, AS),
    AS2 = as:dict_put(safe, Safety, AS1),
    AS3 = as:dict_put(time, Time, AS2),
    AS4 = as:dict_put(optimistic, Optimistic, AS3),
    AS5 = as:dict_put(quiet, Quiet, AS4),

    put(polyvariance, Poly),
    put(cfa_k, CFA_K),
    put(max_contexts, Max),
    put(external_specs, ExternalSpecs),
    AS5.

time(init, _) ->
    statistics(runtime);
time(_, AS) ->
    case as:dict_get(time, AS) of
	true ->
	    {_, Time} = statistics(runtime),
	    io_lib:format("Time:~6.1f seconds", [Time/1000]);
	false ->
	    true
    end.
    
load_standard_spec_modules() ->
    program:load_mod(erlang, spec),
    program:load_mod(math, spec),
    program:load_mod('This is not the primop module', spec).

arglist(N) when N > 0 ->
    [term:universal() | arglist(N-1)];
arglist(0) ->
    [].

%% Set a path.

set_path(Path0) ->
    program:set_path(Path0).



%% Read a string. A function such as this would be present in a
%% well-designed standard library.

read_string(String) ->
    {ok, Tokens, _} = 
	erl_scan:string(String),
    {ok, Term} = erl_parse:parse_term(Tokens ++ [{dot,1}]),
    Term.

    
