%%% File    : as.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@csd.uu.se>
%%% Purpose : Analysis State -- Represent the state of the analysis
%%% Created : 15 Feb 1999 by Sven-Olof Nystr|m <svenolof@csd.uu.se>

-module(as).
-author('svenolof@harpo.csd.uu.se').

%-define(trace(AS, AS1), trace_update(AS, AS1)).
-define(trace(AS, AS1), AS1).

-export([make_state/0, lookup/3, lookup/2, lookup_cont/4,
	 insert/3, get_work/1, add_visited/3,  make_context_table/2,
	 dict_put/3, dict_get/2, report/3, get_reports/1]).

make_state() ->
    AS = {state, 
	  store:empty_store(),
	  gb_trees:empty(), % table of known activation points
	  wl:empty(), 
	  gb_trees:empty()},
    trace_init(AS),
    init_state(AS).

init_state(AS) ->
    S = tag:any_string(),
    AS1 = insert(S, term:join2(term:make_nil(), 
			       term:make_cons(term:make_int_any(), S)), AS),
    dict_key(report, gb_trees:empty(), AS1).
    

lookup(IP, Context, AS) ->
    Term = store:lookup_ip_term(IP, as_store(AS)),
    Work = work:current_work(),
    AS1 = depend(IP, Work, AS),
    ?trace(AS, AS1),
    {Term, AS1}.

lookup(IP, AS) ->
    store:lookup_ip_term(IP, as_store(AS)).


% Run continuation Cont for each term found in IP.

lookup_cont(IP, Cont, Context, AS) ->
    {Term, AS1} = lookup(IP, Context, AS),
    case term:term_tag(Term) of 
	join ->
	    L = term:join_arg_list(Term),
	    lists:foldl(fun(Term, BS)-> Cont(Term, BS) end, AS1, L);
	_ ->
	    Cont(Term, AS1)
    end.


as_store({state, Store, _, _, _}) ->
    Store.

insert(IP, Term, AS) ->
    {state, Store, Pending, Worklist, Dict} = AS,
    {Store1, WL} = store:insert(IP, Term, Store),
    KeyList = lists:map(fun (W) -> work:work_index(W) end, WL),
    Worklist1 = wl:insert_list(KeyList, WL, Worklist),
    AS1 = {state, Store1, Pending, Worklist1, Dict},
    ?trace(AS, AS1).

depend(IP, Work, {state, Store, Pending, Worklist, Dict}) ->
    Store1 = store:depend(IP, Work, Store),
    {state, Store1, Pending, Worklist, Dict}.

% Handling of work

get_work(AS={state, Store, Pending, Worklist, Dict}) ->
    case wl:get_work(Worklist) of
	undefined ->
	    undefined;
	{ok, Work, WL} ->
	    AS1 =  {state, Store, Pending, WL, Dict},
	    ?trace(AS, AS1),
	    {ok, Work, AS1}
    end.


add_visited(AP, Context, AS={state, Store, Visited, Worklist, Dict}) ->
    Index = work:work_index(AP),
    case gb_trees:is_defined(Index, Visited) of
	true ->
	    {state, Store, Visited, Worklist, Dict};
	false ->
	    Visited1 = gb_trees:insert(Index, true, Visited),
	    WL = wl:insert(Index, AP, Worklist),
	    AS1 = {state, Store, Visited1, WL, Dict},
	    ?trace(AS, AS1)		
    end.

%% Presenting the contents of the store

make_context_table({state, Store, _, _, _}, Interesting) ->
    store:make_context_table(Store, Interesting).

%% Report

report(Fnid, Message, AS) ->
    ReportTable = dict_get(report, AS),
    case gb_trees:lookup(Fnid, ReportTable) of
	{value, MessageTable} ->
	    case gb_trees:is_defined(Message, MessageTable) of
		true ->
		    AS;
		false ->
		    MessageTable1 = gb_trees:insert(Message, true, MessageTable),
		    ReportTable1 = gb_trees:update(Fnid, MessageTable1, ReportTable),
		    dict_put(report, ReportTable1, AS)
	    end;
	none ->
	    MessageTable = gb_trees:insert(Message, true, gb_trees:empty()),
	    ReportTable1 = gb_trees:insert(Fnid, MessageTable, ReportTable),
	    dict_put(report, ReportTable1, AS)
    end.

get_reports(AS) ->
    dict_get(report, AS).
       
%% Dictionary. A general dictionary in each analysis state.

dict_key(Name, InitVal, AS={state, Store, Visited, Worklist, Dict}) ->
    Dict1 = gb_trees:insert(Name, InitVal, Dict),
    AS1={state, Store, Visited, Worklist, Dict1},
    ?trace(AS, AS1).
    
dict_get(Name, {state, Store, Visited, Worklist, Dict}) ->
    {value, Value} = gb_trees:lookup(Name, Dict),
    Value.

dict_put(Name, Value, AS={state, Store, Visited, Worklist, Dict}) ->
    Dict1 = gb_trees:enter(Name, Value, Dict),
    AS1={state, Store, Visited, Worklist, Dict1},
    ?trace(AS, AS1).
	    

%%%
trace_init(AS) ->
    put(trace_state, AS).

trace_update(AS0, AS) ->
    ASb = put(trace_state, AS),
    case AS0 == ASb of
	true ->
	    AS;
	false ->
	    io:format("~nBefore:~n~p~nReplace:~n~p~nAfter:~n~p",[AS0, ASb, AS]),
	    exit(0)
    end.
