%%% File    : call.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : Various calls
%%% Created : 17 Aug 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(call).
-author('svenolof@harpo.csd.uu.se').

-export([call_exp/4, local_call/4, call_fnid/5, treat_escaped_list/3, treat_escaped/3]).

call_exp(Exp, Env, Context, AS) ->
    {Module, AS1} = walker:exp(kinds:call_module(Exp), Env, Context, AS),
    {Name, AS2} = walker:exp(kinds:call_name(Exp), Env, Context, AS1),
%    io:format("~nCall: ~w~w",[Module, Name]),
    {ArgList, AS3} = walker:exp_list(kinds:call_arg_list(Exp), Env, Context, AS2),
    Self = env:lookup('$self', Env),
    Except = env:lookup('$except', Env),
    Pos = kinds:pos(Exp),
    Result_IP = tag:call_result(Pos, Context),
    AS4 = call ({Module, Name, ArgList}, {Result_IP, Self, Except}, Pos, Context, AS3),
    {Result_IP, AS4}.

call({Module, Name, ArgList}, {ResultIP, Self, Except}, Pos, Context, AS) ->
    F =	fun ([Module1, Name1], BS) ->
		case term:is_universal(Module1) or term:is_universal(Name1) of
		    false ->
			ModuleName = term:atom_val(Module1),
			NameName = term:atom_val(Name1),
			{Result, BS1} =
			    general_call({ModuleName, NameName, ArgList}, 
					 Self, Except, Pos, Context, BS),
			as:insert(ResultIP, Result, BS1);
		    true ->
			BS1 = treat_escaped_list(ArgList, Context, BS),
			as:insert(ResultIP, term:universal(), BS1)
		end
	end,
    wait:wait([Module, Name], [atom, atom], F, Context, AS).

general_call({Module, Name, ArgList}, Self, Except, Pos, Context, AS) ->
    Arity = length(ArgList),
    Fnid = program:make_fnid(Module, Name, Arity),
    case program:spec_mode(Module) of
	local ->
	    call_analysis(Fnid, ArgList, Self, Except, Pos, Context, AS);
	extern ->
	    Pos1 = {Pos, walker:current_fnid()},
	    call_specification(Fnid, ArgList, Self, Except, Pos1, Context, AS);
	unknown ->
	    AS1 = report:report_unknown_module(Module, AS),
	    F = fun (Arg, BS) ->
			treat_escaped(Arg, Context, BS)
		end,
	    AS2 = lists:foldl(F, AS1, ArgList),
	    {term:universal(), AS2}
    end.

call_fnid(Fnid, ArgList, {Result, Self, Except}, Context, AS) ->
    Module = program:fnid_module(Fnid),
    Pos = 0, 
% Will this work? The current position should be encoded in current context.
    case program:spec_mode(Module) of
	local ->
	    {Term, AS1} =
		call_analysis(Fnid, ArgList, Self, Except, Pos, Context, AS),
	    as:insert(Result, Term, AS1);
	extern -> 
	    Pos1 = {Pos, walker:current_fnid()},
	    {Term, AS1} =
		call_specification(Fnid, ArgList, Self, Except, Pos1, Context, AS),
	    as:insert(Result, Term, AS1)
    end.


%%% local call

local_call(Exp, Env, Context, AS) ->
    Pos = kinds:pos(Exp), 
    Self = env:lookup('$self', Env),
    Except = env:lookup('$except', Env),
    ExpList = kinds:apply_arg_list(Exp),
    Operator = kinds:apply_op(Exp),
    {ArgList, AS0} = walker:exp_list(ExpList, Env, Context, AS),
    Module = context:current_module(Context),
    Name = kinds:fref_name(Operator),
    Arity = kinds:fref_arity(Operator),
    case length(ExpList) == Arity of
	true ->
	    Fnid = program:make_fnid(Module, Name, Arity),
	    call_analysis(Fnid, ArgList,  Self, Except, Pos, Context, AS0)
    end.



call_analysis(Fnid, ArgList, Self, Except, Pos, Context, AS) ->
    {AP, AS1} = work:call_analysis(Fnid, Pos, Context, AS),
    AS2 = param:insert_arg_list(AP, ArgList, AS1),
    AS3 = as:insert(Except, tag:except(AP), AS2),
    AS4 = as:insert(tag:self(AP), Self, AS3),
    Term = tag:result(AP),
    {Term, AS4}.

call_specification(Fnid, ArgList, Self, Except, Pos, Context, AS) ->
    Module = program:fnid_module(Fnid),
    AS0 = record_call(context:current_fnid(Context), Fnid, Pos, ArgList, AS),
    case program:try_module(Module, spec) or as:dict_get(use_spec, AS0) of
	true ->
	    {NewContext, AS1} = work:call_specification(Fnid, Pos, Context, AS0),
	    AS2 = param:insert_arg_list(NewContext, ArgList, AS1),
	    AS3 = as:insert(Except, tag:except(NewContext), AS2),
	    AS4 = as:insert(tag:self(NewContext), Self, AS3),
	    Term = tag:result(NewContext),
	    {Term, AS4};
	false ->
	    AS2 = treat_escaped(ArgList, Context, AS0),
	    {term:universal(), AS0}
    end.

% Record args used for precision measurements.

record_call(Caller, Fnid, Pos, ArgList, AS) ->
    Module = program:fnid_module(Fnid),
    case lists:member(Module, [erlang, math]) of 
	true ->
	    record_args(1, Caller, Fnid, Pos, ArgList, AS);
	false ->
	    AS
    end.

record_args(_, _, _, _, [], AS) ->
    AS;
record_args(N, Caller, Fnid, Pos, [Term | TermList], AS) ->
    AS1 = as:insert(tag:record_bif(Caller, Fnid, Pos, N), Term, AS),
    record_args(N+1, Caller, Fnid, Pos, TermList, AS1).

%%%

treat_escaped_list(TermList, Context, AS) ->
    F = fun (T, BS) ->
		treat_escaped(T, Context, BS)
	end,
    lists:foldl(F, AS, TermList).

treat_escaped(Term, Context, AS) ->
    Visited = gb_trees:empty(),
    treat_escaped(Term, Visited, Context, AS).

treat_escaped(Term, Visited, Context, AS) ->
    case gb_trees:is_defined(Term, Visited) of
	true ->
	    AS;
	false ->
	    V1 = gb_trees:insert(Term, true, Visited),
	    case term:term_tag(Term) of
		ip ->
		    {Term1, AS1} = as:lookup(Term, Context, AS),
		    treat_escaped(Term1, V1, Context, AS1);
		universal ->
		    AS;
		join ->
		    TermList = term:join_arg_list(Term),
		    F = fun (T, BS) ->
				treat_escaped(T, V1, Context, BS)
			end,
		    lists:foldl(F, AS, TermList);
		functor ->
		    TermList = term:term_functor_arg_list(Term),
		    F = fun (T, BS) ->
				treat_escaped(T, V1, Context, BS)
			end,
		    lists:foldl(F, AS, TermList);
		'fun' ->
		    Arity = term:fun_arity(Term),
		    Pos = term:fun_pos(Term),
		    FunContext = term:fun_ap(Term),
		    ArgList = unilist(Arity),
		    param:insert_fun_arg_list(FunContext, Pos, ArgList, AS);
		X ->
		    io:format("~nEscape:~w",[Term]),
		    AS
	    end
    end.

unilist(0) ->
     [];
unilist(N) ->
    [term:universal() | unilist(N-1)].
	
