%%% File    : context.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@csd.uu.se>
%%% Purpose : Implement information points
%%% Created : 15 Feb 1999 by Sven-Olof Nystr|m <svenolof@csd.uu.se>

-module(context).
-author('svenolof@csd.uu.se').

-export([init/0, empty/0, call/3, 
	 current_fnid/1, set_current_fnid/2, report_point/1, 
	 current_module/1, current_tyid/1, caller/1,
	 context_proper/1]).

% A context: current function and (abstraction of) path leading to
% it. An analysis point: Synonymous concept.
%
% Representation of context: <Fnid, <Stratum, Pos>*>.

% Macros 'polyvariance' and 'cfa_k' used in benchmarking only. Two
% setups, the first for normal use, the second for benchmarking.

% -define(DefaultMaxContexts,100).
% -define(polyvariance, default).
% -define(cfa_k, 0).

-define(DefaultMaxContexts,get(max_contexts)).
-define(polyvariance, get(polyvariance)).
-define(cfa_k, get(cfa_k)).


empty() ->
    {context, nil, nil, find_context([])}.

init() ->
    put(context_table, gb_trees:empty()),
    put(context_index, gb_trees:empty()),
    true.
    
call(Fnid, Pos, Context) ->
    case ?polyvariance of
	default ->
	    call_default(Fnid, Pos, Context);
	cfa ->
	    call_cfa(Fnid, Pos, Context);
	ddr ->
	    call_ddr(Fnid, Pos, Context);
	ddr_lim ->
	    call_ddr_lim(Fnid, Pos, Context);
	ddr_sdr ->
	    call_ddr_sdr(Fnid, Pos, Context);
	ddr_sdr_lim ->
	    call_ddr_sdr_lim(Fnid, Pos, Context);
	sl_special ->
	    call_sl_special(Fnid, Pos, Context)	
end.


call_default(Fnid, Pos, {context, Current, _Caller, Context}) ->
    Stack = context_data(Context),
    case gb_trees:lookup(Fnid, get(strata_table)) of
	{value, Stratum} -> 
	    true;
	none ->
	    Stratum = Fnid
    end,

    FPos = {fpos,Stratum, Pos},
    case find_previous(FPos, Stack) of
	{ok, Stack1} ->
	    PC = find_context(Stack1);			  
	undefined -> 
	    case poly(Fnid, Stratum) of
		true ->
		    PC = find_context([FPos | Stack]);
		false ->
		    PC = find_context([FPos])
	    end	    
    end,
    {context, Fnid, Current, PC}.

poly(Fnid, Stratum) ->
    case gb_trees:lookup(Stratum, get(weight_table)) of
	{value, W} ->
	    case W < ?DefaultMaxContexts of 
		true -> 
		    true;
		false ->
		    in_standard_module(Fnid) 
	    end;
	none -> 
	    true
    end.

in_standard_module(Fnid) ->
    lists:member(program:fnid_module(Fnid), [erlang, math]).

context_data(ContextID) ->
    {value, ContextData} = 
	gb_trees:lookup(ContextID, get(context_table)),
    ContextData.

find_context(Stack) ->
    Index = get(context_index),
    case gb_trees:lookup(Stack, Index) of
	{value, Context} ->
	    Context;
	none ->
	    ID = gb_trees:size(Index),
	    Context = ID,

	    Index1 = gb_trees:insert(Stack, Context, Index),
	    put(context_index, Index1),
	    Table = get(context_table),
	    Table1 = gb_trees:insert(ID, Stack, Table),
	    put(context_table, Table1),
	    Context
    end.

current_fnid({context, Fnid, _, _}) ->
    Fnid.

current_tyid(Context) ->
    current_fnid(Context).

current_module(Context) ->
    Deid = current_fnid(Context),
    case program:deid_type(Deid) of
	fnid ->
	    program:fnid_module(Deid);
	tyid ->
	    program:tyid_module(Deid)
    end.

set_current_fnid(Fnid, {context, _, Caller, C}) ->
    {context, Fnid, Caller, C}.


report_point(Context) ->
    current_fnid(Context).

find_previous(_FPos, []) ->
    undefined;
find_previous({fpos, Stratum,  _}, Stack = [{fpos, Stratum, _} | _]) ->
    {ok, Stack};
find_previous(FPos, [_ | Stack]) ->
    find_previous(FPos, Stack).

caller({context, _, Fnid, _}) ->
    Fnid.


context_proper({context, _, _, S}) ->
    S.


%%% Alternative definitions of call. For benchmarking purposes only.

% call_cfa. My impression of Shiver's classic.

call_cfa(Fnid, Pos, {context, Current, _Caller, Context}) ->
    Stack = context_data(Context),
    FPos = {fpos,Fnid, Pos},
    Stack1 = [FPos | Stack],
    case in_standard_module(Fnid) of
	true ->
	    Stack2 = prefix(?cfa_k, Stack1); %Stack2 = Stack1;
	false ->
	    Stack2 = prefix(?cfa_k, Stack1)
    end,
    PC = find_context(Stack2),
    {context, Fnid, Current, PC}.

% call_ddr. Let call stack grow freely. Only cut when recursion is
% dynamically detected.

call_ddr(Fnid, Pos, {context, Current, _Caller, Context}) ->
    Stack = context_data(Context),
    FPos = {fpos,Fnid, Pos},
    case find_previous(FPos, Stack) of
	{ok, Stack1} ->
	    PC = find_context(Stack1);			  
	undefined -> 
	    PC = find_context([FPos | Stack])
    end,
    {context, Fnid, Current, PC}.

% call_ddr_sdr. Like call_ddr but identifies mutually recursive
% functions. Should be much faster when mutual recursion occurs.

call_ddr_sdr(Fnid, Pos, {context, Current, _Caller, Context}) ->
     case gb_trees:lookup(Fnid, get(strata_table)) of
	{value, Stratum} -> 
	    true;
	none ->
	    Stratum = Fnid
    end,
    Stack = context_data(Context),
    FPos = {fpos,Stratum, Pos},
    case find_previous(FPos, Stack) of
	{ok, Stack1} ->
	    PC = find_context(Stack1);			  
	undefined -> 
	    PC = find_context([FPos | Stack])
    end,
    {context, Fnid, Current, PC}.

% call_ddr_lim. Like call_ddr but limits the length of a context. 

call_ddr_lim(Fnid, Pos, {context, Current, _Caller, Context}) ->
    Stack = context_data(Context),
    FPos = {fpos,Fnid, Pos},
    case find_previous(FPos, Stack) of
	{ok, Stack1} ->
	    PC = find_context(Stack1);			  
	undefined -> 
	    Stack1 = [FPos | Stack],
            Stack2 = prefix(?cfa_k, Stack1),
	    PC = find_context(Stack2)
    end,
    {context, Fnid, Current, PC}.

% call_ddr_sdr_lim. Like call_ddr_sdr but limits.

call_ddr_sdr_lim(Fnid, Pos, {context, Current, _Caller, Context}) ->
     case gb_trees:lookup(Fnid, get(strata_table)) of
	{value, Stratum} -> 
	    true;
	none ->
	    Stratum = Fnid
    end,
    Stack = context_data(Context),
    FPos = {fpos,Stratum, Pos},
    case find_previous(FPos, Stack) of
	{ok, Stack1} ->
	    PC = find_context(Stack1);			  
	undefined -> 
	    Stack1 = [FPos | Stack],
            Stack2 = prefix(?cfa_k, Stack1),
	    PC = find_context(Stack2)
    end,
    {context, Fnid, Current, PC}.

% Simple list operation. Couldn't find anything like it in lists.

prefix(0, _L) ->
     [];
prefix(_, []) ->
    [];
prefix(N, [X|L]) ->
    [X|prefix(N-1, L)].



call_sl_special(Fnid, Pos, {context, Current, _Caller, Context}) ->
    Stack = context_data(Context),
    FPos = {fpos,Fnid, Pos},
    Stack1 = [FPos | Stack],
    case program:fnid_name(Fnid) == mk_label of
	true ->
	    Stack2 = Stack1;
	false ->
	    Stack2 = prefix(?DefaultMaxContexts, Stack1)
    end,
    PC = find_context(Stack2),
    {context, Fnid, Current, PC}.

