-module(draw).

% -compile(export_all).
-export([draw_struct/1]).

% Give type structures a representation that can be written to file.

draw_struct(S) ->
    NameList = present_struct:struct_name_list(S),
    ArgList = get_arglist(NameList),
    F = fun(Name) ->
		Term = present_struct:lookup_name(Name, S),
		draw_term(Term)
	end,
    Header = {lists:map(F, ArgList), 
	      case lists:member(result, NameList) of 
		  true -> F(result);
		  false -> draw_term(term:join0())
	      end},
    IP_list = lists:sort(lists:map(fun(Name) ->
					   {term:numbered_ip_number(Name), Name}
				   end,
				   present_struct:struct_ip_list(S))),
    G = fun({_, IP}) ->
		IP_name = draw_ip_ref_term(IP),
		Term = present_struct:lookup_ip(IP, S),
		TermTree = draw_term(Term),
		Tree = draw_bind(IP_name, TermTree),
		Tree
	end,
    WList = lists:map(G, IP_list),
    {Header, WList}.

get_arglist(NameList) ->
    lists:sort(lists:filter(fun({arg, _}) ->
				    true;
			       (_) ->
				    false
			    end,
			    NameList)).

draw_term(Term) ->
    case term:term_tag(Term) of 
	functor ->
	    draw_functor_term(Term);
	'fun' ->
	    draw_fun_term(Term);
	ip ->
	    draw_ip_ref_term(Term);
	join ->
    	    draw_join_term(Term);
	universal ->
    	    draw_any();
	parameter ->
    	    draw_parameter_term(Term);
	pid ->
	    draw_pid_term(Term)	   
    end.

draw_functor_term(Term) ->
    case term:is_formal(Term) of
	true ->
	    {any};
	false ->
	    Functor = term:term_functor(Term),
	    List = term:term_functor_arg_list(Term),
	    draw_functor(Functor, List)
    end.

draw_functor(atomc, [Term]) ->
    case term:term_tag(Term) of
	universal ->
	    {atom};
	functor ->
	    {atoma, Atom} = term:term_functor(Term),
	    {atom, Atom}
    end;
draw_functor(intc, [Term]) ->
    case term:term_tag(Term) of
	universal ->
	    {int};
	functor ->
	    {inta, Int} = term:term_functor(Term),
	    {int, Int}
    end;
draw_functor(floatc, [Term]) ->
    case term:term_tag(Term) of
	universal ->
	    {float}
    end;
draw_functor(tuple, [Term]) ->
    case is_clean_list(Term) of
	true ->
	    {tuple, draw_clean_list(Term)};
	false ->
	    {tuple, draw_term(Term)}
    end;
draw_functor(cons, TermList) ->
    [Hd, Tl] = draw_list(TermList),
    {cons, Hd, Tl};
draw_functor(nil, []) ->
    {nil};
draw_functor(FunctorName, List) -> % This should not happen! 
    List1 = draw_list(List),
    {call, 0, {atom, 0, FunctorName},List1}.

is_clean_list(Term) ->
    case term:term_tag(Term) of
	functor ->
	    case term:term_functor(Term) of
		cons ->
		    [_, Tl] = term:term_functor_arg_list(Term),
		    is_clean_list(Tl);
		nil ->
		    true;
		_ ->
		    false
	    end;
	_ ->
	    false
    end.

draw_clean_list(Term) ->
    case term:term_tag(Term) of
	functor ->
	    case term:term_functor(Term) of
		cons ->
		    [Hd, Tl] = term:term_functor_arg_list(Term),
		    [draw_term(Hd) |
		     draw_clean_list(Tl)];
		nil ->
		    []
	    end
    end.


	

draw_fun_term(_Fun) ->
    {'fun'}.


draw_bind(Var, Val) ->
    {'=', Var, Val}.

draw_ip_ref_term(IP) ->
    N = term:numbered_ip_number(IP),
    Name = show:nice_var_name(N),
    {var, list_to_atom(Name)}.


draw_join_term(Term) ->
    case term:join_arg_list(Term) of
	[] ->
	     {none};
	[T] ->
	    draw_term(T);
	TermList ->
	    L = lists:map(fun(Term1) ->
				  draw_term(Term1)
			  end,
			  TermList), 
	    {join, L}
    end.

draw_list(TermList) ->
    lists:map(fun draw_term /1, TermList).

draw_parameter_term(_) ->
    draw_any().

draw_pid_term(_) ->
    {pid}.

draw_any() ->
    {any}.

% Reading structs; No point in writing structs if we cannot read them!
% Put it here for now, despite the name of the module.

%% read(ModuleName, MFA) ->
%%     Path = program:path(),
%%     {ok, FileName} = sa_file:find_module(Path, ModuleName, present),
%%     {ok, L} = file:consult(FileName),
%%     {value, {function, _, Struct}} = lists:keysearch(MFA, 2, L),
%%     Struct.


    
