%%% File    : flow.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : Build flow graph. Compute cycles.
%%% Created : 29 Oct 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(flow).
-author('svenolof@harpo.csd.uu.se').

-export([module_flow_graph/1,multi_module_flow_graph/1]).


module_flow_graph(M) ->
    G0 = graph:empty_graph(),
    DefList0 = gb_trees:to_list(program:program_fnid_table()),
    DefList = lists:filter(fun({Fnid, _}) ->
				   program:fnid_module(Fnid) == M
			   end,
			   DefList0),
    build_graph(DefList, G0).

multi_module_flow_graph(MList) ->
    G0 = graph:empty_graph(),
    DefList0 = gb_trees:to_list(program:program_fnid_table()),
    DefList = lists:filter(fun({Fnid, _}) ->
				   lists:member(program:fnid_module(Fnid), MList)
			   end,
			   DefList0),
    build_graph(DefList, G0).

build_graph([{Fnid, Def} | List], G) ->
    G0 = graph:new_node(Fnid, G),
    G1 = case kinds:is_fdef(Def) of
	     true ->
		 traverse(Fnid, kinds:fdef_body(Def), G0);
	     false  ->
		 G0
	 end,	     
    build_graph(List, G1);
build_graph([], G) ->
    G.

traverse(Fnid, E, G) ->
    case kinds:exp_type(E) of
	var ->
	    G;
	constructor ->
	    L = kinds:constructor_exp_arg_list(E),
	    traverse_list(Fnid, L, G);
	control ->
	    traverse_control(Fnid, E, G);
	apply ->
	    traverse_apply(Fnid, E, G);
	call ->
	    traverse_call(Fnid, E, G);
	fref ->
	    traverse_fref(Fnid, E, G);
	'fun' ->
	    traverse_fun(Fnid, E, G)
    end.

traverse_list(Fnid, [E | EList], G) ->
    G1 = traverse(Fnid, E, G),
    traverse_list(Fnid, EList, G1);
traverse_list(Fnid, [], G) ->
    G.

traverse_control(Fnid, E, G) ->
    case kinds:control_exp_type(E) of
	'let' ->
	    Arg = kinds:let_exp_arg(E),
	    Body = kinds:let_exp_body(E),
	    traverse_list(Fnid, [Arg, Body], G);
	letrec ->
	    DefList = kinds:letrec_def_list(E),
	    Body = kinds:letrec_body(E),
	    traverse_list(Fnid, 
			  [Body | lists:map(fun({_, Fun}) -> Fun end, DefList)], G);
	'case' ->
	    Arg = kinds:case_exp_arg(E),
	    ClauseList = kinds:case_exp_clause_list(E),
	    G1 = traverse(Fnid, Arg, G),
	    traverse_clause_list(Fnid, ClauseList, G1);
	'receive' ->
	    ClauseList = kinds:receive_exp_clause_list(E),
	    Timeout = kinds:receive_exp_timeout(E),
	    Action = kinds:receive_exp_action(E),
	    G1 = traverse_clause_list(Fnid, ClauseList, G),
	    traverse_list(Fnid, [Timeout, Action], G1);
	seq ->
	    Arg = kinds:seq_arg(E),
	    Body = kinds:seq_body(E),
	    traverse_list(Fnid, [Arg, Body], G);
	'catch' ->
	    Body = kinds:catch_body(E),
	    traverse(Fnid, Body, G);
	'try' ->
	    Body = kinds:try_body(E),
	    traverse(Fnid, Body, G)
    end.

traverse_clause(Fnid, Clause, G) ->
    Guard = kinds:clause_guard(Clause),
    Body = kinds:clause_body(Clause),
    traverse_list(Fnid, [Guard, Body], G).

traverse_clause_list(Fnid, [Clause | ClauseList], G) ->
    G1 = traverse_clause(Fnid, Clause, G),
    traverse_clause_list(Fnid, ClauseList, G1);
traverse_clause_list(Fnid, [], G) ->
    G.

traverse_apply(Fnid, E, G) ->	    
    Op = kinds:apply_op(E),
    ArgList = kinds:apply_arg_list(E),
    G1 =  traverse_list(Fnid, [Op| ArgList], G),
    case kinds:is_fref(Op) of
	true ->
	    Module = program:fnid_module(Fnid),
	    Name = kinds:fref_name(Op),
	    Arity = kinds:fref_arity(Op),
	    Son = program:make_fnid(Module, Name, Arity),
	    graph:insert_son(Fnid, Son, G1);
	_ ->
	    G1
    end.

traverse_call(Fnid, E, G) ->
    Module = kinds:call_module(E),
    Name = kinds:call_name(E),
    ArgList = kinds:call_arg_list(E),
    G1 = traverse_list(Fnid, [Module, Name | ArgList], G),
    case locate_dest(E) of
	{ok, Son} ->
	    graph:insert_son(Fnid, Son, G1);
	not_found ->
	    G1
    end.

locate_dest(E) ->
    Module =  kinds:call_module(E),
    Name =  kinds:call_name(E),
    case kinds:is_atom(Module) and kinds:is_atom(Name) of
	true ->
	    Arity = length(kinds:call_arg_list(E)),
	    Son = program:make_fnid(kinds:concrete(Module), kinds:concrete(Name), Arity),
	    {ok, Son};
	false ->
	    not_found
    end.

traverse_fref(Fnid, E, G) ->
    Module = program:fnid_module(Fnid),
    Name = kinds:name_atom(E),
    Arity = kinds:fref_arity(E),
    Son = program:make_fnid(Module, Name, Arity),
    graph:insert_son(Fnid, Son, G).

traverse_fun(Fnid, E, G) ->
    Body = kinds:fun_body(E),
    traverse(Fnid, Body, G).


	










