%%% File    : generate_special.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Description : Handling of specials in specifications
%%% Created : 20 Aug 2002 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(generate_special).

-export([generate_special/6]).

-define(max_reasonable_index, 10).

generate_special(nth, 2, [N_Term, L_Term], Pos, Context, AS) ->
    case is_simple_nth(N_Term, L_Term, Context, AS) of
	{none, AS1} ->
	    {term:join0(), AS1};
	{any, AS1} ->
	    Result = tag:nth_result(Context, Pos),
	    AS2 = generate_general_nth(L_Term, Result, gb_trees:empty(), Pos, Context, AS1),
	    {Result, AS2};
	{{N, L}, AS1} ->
	    {lists:nth(N, L), AS1}
    end;

generate_special(set_nth, 3, [N_Term,L_Term,New_Term], Pos, Context, AS) ->
    case is_simple_nth(N_Term, L_Term, Context, AS) of
	{none, AS1} ->
	    {term:join0(), AS1};
	{{_, _}, AS1} ->
	    generate_precise_set_nth(N_Term, L_Term, New_Term, Pos, Context, AS1);
	{any, AS1} ->
	    ListElement = tag:nth_result(Context, Pos),
	    AS2 = generate_general_nth(L_Term, ListElement, gb_trees:empty(), Pos, Context, AS1),
	    AS3 = as:insert(ListElement, New_Term, AS2),
	    Result = tag:set_nth_result(Context, Pos),
	    AS4 = as:insert(Result, term:make_nil(), AS3),
	    AS5 = as:insert(Result, term:make_cons(ListElement, Result), AS4),
	    {Result, AS5}
    end;

generate_special(apply, 3, [M_Term, N_Term, A_Term], _, Context, AS) ->
    RSE = special_definition:rse(Context), 
    Result = special_definition:result(Context),
    AS1 = special_definition:call(M_Term, N_Term, A_Term, RSE, Context, AS),
    {Result, AS1}.

is_simple_nth(Index, List, Context, AS) ->
    case is_fixed_integer(Index, Context, AS) of
	{none, AS1} ->
	    {none, AS1};
	{any, AS1} ->
	    {any, AS1};
	{{int, N}, AS1} when N < ?max_reasonable_index ->
	    case is_fixed_list(List, Context, AS1) of
		{none, AS2} ->
		    {none, AS2};
		{any, AS2} ->
		    {any, AS2};
		{{list, L}, AS2} when length(L) < ?max_reasonable_index ->
		    {{N, L}, AS2};
		{{list, _}, AS2} ->
		    {any, AS2}
	    end;
	{_, AS2} ->
	    {any, AS2}
    end.

is_fixed_integer(N, Context, AS) ->
    {N1, AS1} = spec_match:deref(N, Context, AS),
    case term:is_empty(N1) of
	true ->
	    {none, AS1};
	false ->
	    case term:term_tag(N1) of
		functor ->
		    case term:is_int(N1) of
			true ->
			    Arg = hd(term:term_functor_arg_list(N1)),
			    {Arg1, AS2} = spec_match:deref(Arg, Context, AS1),
			     case term:term_tag(Arg1) of
				 functor ->
				     case term:term_functor(Arg1) of
					 {inta, Integer} ->
					     {{int, Integer}, AS2};
					 _ ->
					     {any, AS2}
				     end;
				 _ ->
				     {any, AS2}
			    end;
			false ->
			    {any, AS1}
		    end;
		_ ->
		    {any, AS1}
	    end
    end.

is_fixed_list(List, Context, AS) ->
    {List1, AS1} = spec_match:deref(List, Context, AS),
    case term:is_empty(List1) of
	true ->
	    {none, AS1};
	false ->
	    case term:term_tag(List1) of
		functor ->
		    case term:term_functor(List1) of
			nil ->
			    {{list, []}, AS1};
			cons ->
			    Cdr = term:cons_cdr(List1),
			    case is_fixed_list(Cdr, Context, AS1) of
				{none, AS2} ->
				    {none, AS2};
				{any, AS2} ->
				    {any, AS2};
				{{list, L}, AS2} ->
				    {{list, [term:cons_car(List1) | L]}, AS2}
			    end;
			_ ->
			    {any, AS1}
		    end;
		_ ->
		    {any, AS1}
	    end
    end.
    
generate_general_nth(Term, Result, Visited, Pos, Context, AS) ->
    case gb_trees:is_defined(Term, Visited) of
	true ->
	    AS;
	false ->
	    Visited1 = gb_trees:insert(Term, true, Visited),
	    Result = tag:nth_result(Context, Pos),
	    case term:term_tag(Term) of
		ip ->
		    {Term1, AS1} = as:lookup(Term, Context, AS),
		    generate_general_nth(Term1, Result, Visited1, Pos, Context, AS1);
		join ->
		    TermList = term:join_arg_list(Term),
		    F = fun (Term1, BS) ->
				generate_general_nth(Term1, Result, Visited1, Pos, Context, BS)
			end,
		    lists:foldl(F, AS, TermList);
		universal ->
		    as:insert(Result, Term, AS);
		functor ->
		    case term:is_cons(Term) of
			true ->
			    [Car, Cdr] = term:term_functor_arg_list(Term),
			    AS1 = as:insert(Result, Car, AS),
			    generate_general_nth(Cdr, Result, Visited1, Pos, Context, AS1);
			false ->
			    AS
		    end;
		_ ->
		    AS
	    end
    end.

generate_precise_set_nth(N_Term, L_Term, New_Term, Pos, Context, AS) ->    
    Result = tag:set_nth_result(Context, Pos),
    G = fun(Res, CS) ->
		as:insert(Result, Res, CS)
	end,
    F = fun(N, BS) ->
		wait_set_nth(N, L_Term, New_Term, G, Context, BS)
	end,
    AS1 = wait_int(N_Term, F, Context, AS),
    H = fun(BS) ->
		wait_set_any(L_Term, New_Term, G, Context, BS)
	end,
    AS2 = wait_int_any(N_Term, H, Context, AS1),
    {Result, AS2}.

% simple_set_nth assumes that N is an int an L a list, without any
% complications, such as unions or ip's.

wait_set_nth(N, List, X, F, Context, AS) ->
    G = fun (Term, BS) ->
		case term:is_cons(Term) of
		    true when N > 1 ->
			Hd = term:cons_car(Term),
			Tl = term:cons_cdr(Term),
			F1 = fun(Tl1, CS) ->
				     F(term:make_cons(Hd, Tl1), CS)
			     end,
			wait_set_nth(N-1, Tl, X, F1, Context, BS);
		    true when N == 1 ->
			Tl = term:cons_cdr(Term),
			F(term:make_cons(X,Tl), BS);
		    _ ->
			case term:is_universal(Term) of
			    true ->
				F(Term, BS);
			    false ->
				BS
			end
		end
	end,
    term:do_all(List, G, Context, AS).

wait_set_any(List, X, F, Context, AS) ->
    G = fun (Term, BS) ->
		case {term:is_cons(Term), term:is_universal(Term)} of
		    {true, _}  ->
			Hd = term:cons_car(Term),
			Tl = term:cons_cdr(Term),
			BS1 = F(term:make_cons(X,Tl), BS),
			F1 = fun(Tl1, CS) ->
				     F(term:make_cons(Hd, Tl1), CS)
			     end,
			wait_set_any(Tl, X, F1, Context, BS1);
		    {_, true} ->
			BS1 = F(term:universal(), BS),
			BS1;
		    _ ->
			BS
		end
	end,
    term:do_all(List, G, Context, AS).

wait_int(Term0, F, Context, AS) ->
    G = fun (Term, BS) ->
		case term:is_int(Term) of
 		    true ->
			[Term1] = term:term_functor_arg_list(Term),
			wait_int_1(Term1, F, Context, BS);
		    false ->
			BS
		end
	end,
    term:do_all(Term0, G, Context, AS).

wait_int_1(Term0, F, Context, AS) ->
    G = fun (Term, BS) ->
		case term:term_tag(Term) of
		    functor ->
			case term:term_functor(Term) of
			    {inta, N} when N < ?max_reasonable_index->
				F(N, BS);
			    _ ->
				BS
			end;
		    _ ->
			BS
		end
	end,
    term:do_all(Term0, G, Context, AS).

wait_int_any(Term0, F, Context, AS) ->
    H = fun (Term, CS) ->
		case term:is_universal(Term) of
		    true ->
			F(CS);
		    false ->
			G = fun (Term1, BS) ->
				    case term:is_int(Term1) of
					true ->
					    [Term2] = term:term_functor_arg_list(Term1),
					    wait_int_any_1(Term2, F, Context, BS);
					false ->
					    BS
				    end
			    end,
                        term:do_all(Term, G, Context, CS)
		end
	end,
    term:do_all(Term0, H, Context, AS).
	

wait_int_any_1(Term0, F, Context, AS) ->
    G = fun (Term, BS) ->
		case term:term_tag(Term) of
		    universal ->
			F(BS);
		    functor ->
			case term:term_functor(Term) of
			    {inta, N} when N >= ?max_reasonable_index->
				F(BS);
			    _ ->
				BS
			end;
		    _ ->
			BS
		end
	end,
    term:do_all(Term0, G, Context, AS).
