%%% File    : match.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : 
%%% Created :  9 Jul 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(match).
-author('svenolof@harpo.csd.uu.se').

-export([clause/5]).


%% Walker. Pattern matching


clause(Clause, Key, Env, Context, AS) ->
    case is_empty_term(Key, Context, AS) of
        {false, AS1} ->
            clause_nonempty(Clause, Key, Env, Context, AS1);
	{true, AS1} ->
	    {term:join0(), term:join0(), AS1}
    end.

clause_nonempty(Clause, Key, Env, Context, AS) ->
    PatternList = kinds:clause_pattern_list(Clause),
    DestList = tag:vector(Context, length(PatternList), kinds:pos(Clause)),
    Guard   = kinds:clause_guard(Clause),
    Body    = kinds:clause_body(Clause),
    Env1 = create_environment_list(PatternList, Env, Context),
    {Diff, Table} = guard_test(Guard, {simple, gb_trees:empty()}),
    Binder = binder(Env1, Context),
    AS1 = wait:wait_vector_r(Key, DestList, Context, AS),
    AS2 = match_list(pattern_list(PatternList),
                     DestList, [], Binder, {Env1, Table}, Context, AS1),
    {Result, AS3} = walker:exp(Body, Env1, Context, AS2),
    {FailKey, AS4}
        = fail(PatternList, {Diff, Table}, Key, Context, AS3),
    {Result, FailKey, AS4}.


%% To be added: If key is empty and clause body is non-trivial, report
%% that clause is unused.

is_empty_term(Term, Context, AS) ->
    is_empty_term(Term, e_history(), Context, AS).

is_empty_term(Term, History, Context, AS) ->
    case term:term_tag(Term) of
        join ->
	    case e_history(History, join) of
		halt ->
		    term:is_empty(Term);
		History1 ->
		    TermList = term:join_arg_list(Term),
		    is_empty_all_term(TermList, History1, Context, AS)
	    end;
	ip ->
	    case e_history(History, ip) of
		halt ->
		    {false, AS};
		History1 ->
		    {Term1, AS1} = as:lookup(Term, Context, AS),
		    is_empty_term(Term1, History1, Context, AS1)
	    end;
        functor ->
	    case e_history(History, functor) of
		halt ->
		    term:is_empty(Term);
		History1 ->
		    TermList = term:term_functor_arg_list(Term),
		    is_empty_any_term(TermList, History1, Context, AS)
	    end;
        _ ->
            {false, AS}
    end.

is_empty_all_term([Term|TermList], Context, History, AS) ->
    case is_empty_term(Term, Context, History, AS) of
	{true, AS1} ->
	    is_empty_all_term(TermList, Context, History, AS1);
	{false, AS1} ->
	    {false, AS1}
    end;
is_empty_all_term([], _Context, _History, AS) ->
    {true, AS}.

is_empty_any_term([Term|TermList], Context, History, AS) ->
    case is_empty_term(Term, Context, History, AS) of
	{true, AS1} ->
	    {true, AS1};
	{false, AS1} ->
	    is_empty_any_term(TermList, Context, History, AS1)
    end;
is_empty_any_term([], _Context, _History, AS) ->
    {false, AS}.

e_history() ->
    0.

e_history(0, ip) ->
    1;
e_history(1, ip) ->
    halt;
e_history(H, _) ->
    H.



fail(Pattern, {simple, Table}, Key, Context, AS) ->
    fail_pattern(Pattern, Key, Table, Context, AS);
fail(_, {hard, _}, Key, _, AS) ->
    {Key, AS}.

create_environment(Pattern, Env, Context) ->
    case kinds:pattern_type(Pattern) of
	var ->
	    Pos = kinds:pos(Pattern),
	    Name = kinds:name_atom(Pattern),
	    IP = tag:var(Name, Pos, Context),
	    env:bind(Name, IP, Env);
	alias ->
	    Var = kinds:alias_var(Pattern),
	    Pattern1 = kinds:alias_pattern(Pattern),
	    Pos = kinds:pos(Var),
	    Name = kinds:name_atom(Var),
	    IP = tag:var(Name, Pos, Context),
	    Env1 = env:bind(Name, IP, Env),
	    create_environment(Pattern1, Env1, Context);
	constructor ->
	    PatternList = kinds:constructor_pattern_arg_list(Pattern),
	    create_environment_list(PatternList, Env, Context)
    end.

create_environment_list([Pattern | PatternList], Env, Context) ->
    Env1 = create_environment(Pattern, Env, Context),
    create_environment_list(PatternList, Env1, Context);
create_environment_list([], Env, _) ->
    Env.


match2(Pattern, Term, Save, Cont, {Env, Table}, Context, AS) ->
    case term:term_tag(Pattern) of
	universal ->
	    Cont(Save, AS);
	pattern_var ->
	    Name = term:pattern_var_name(Pattern),
	    IP = env:lookup(Name, Env),
	    case gb_trees:lookup(Name, Table) of
		none ->
		    Cont([{IP, Term} | Save], AS);
		{value, Test} ->
		    {Term1, AS1} = filter(Test, Term, Context, AS),
		    Cont([{IP, Term1}|Save], AS1)
	    end;
	alias_pattern ->
	    Name = term:alias_pattern_var_name(Pattern),
	    Pattern1 = term:alias_pattern_pattern(Pattern),
	    IP = env:lookup(Name, Env),
	    Save1 = [{IP, Term} | Save],
	    match2(Pattern1, Term, Save1, Cont, {Env, Table}, Context, AS);
	functor ->
	    match_functor(Pattern, Term, Save, Cont, {Env, Table}, Context, AS)
    end.



match_functor(Pattern, Key, Save, Cont, Env, Context, AS) ->
    case term:term_tag(Key) of 
	ip ->
	    F = fun(Term1, AS1) ->
			match_functor(Pattern, Term1, Save, Cont, Env, Context, AS1)
		 end,
	    as:lookup_cont(Key, F, Context, AS);
	join ->
	    match_join(Pattern, Key, Save, Cont, Env, Context, AS);
	universal ->
	    PatternList = term:term_functor_arg_list(Pattern),
	    KeyList = lists:map(fun(_) -> term:universal() end, PatternList),
	    match_list(PatternList, KeyList, Save, Cont, Env, Context, AS);
	parameter ->
	    AS;
	'fun' ->
	    AS;
	functor ->
	    PC = term:term_functor(Pattern),
	    TC = term:term_functor(Key),
	    case PC == TC of
		true ->
		    PatternList = term:term_functor_arg_list(Pattern),
		    KeyList = term:term_functor_arg_list(Key),
		    match_list(PatternList, KeyList, Save, Cont, Env, Context, AS);
		false ->
		    AS
	    end;
	pid->
	    AS
    end.


match_list([P|Patterns], [T|Terms], Save, Cont, Env, Context, AS) ->
    Cont1 = 
	fun(Save1, AS1) ->
		match_list(Patterns, Terms, Save1, Cont, Env, Context, AS1)
	end,
    match2(P, T, Save, Cont1, Env, Context, AS);
match_list([], [], Save, Cont, _, _Context, AS) ->
    Cont(Save, AS).

match_join(Pattern, Term, Save, Cont, Env, Context, AS) ->
    TermList = term:join_arg_list(Term),
    match_join_list(Pattern, TermList, Save, Cont, Env, Context, AS).

match_join_list(Pattern, [Term | TermList], Save, Cont, Env, Context, AS) ->
    AS1 = match2(Pattern, Term, Save, Cont, Env, Context, AS),
    match_join_list(Pattern, TermList, Save, Cont, Env, Context, AS1);
match_join_list(_Pattern, [], Save, Cont, _, _Context, AS) ->
    Cont(Save, AS).


    
binder(Env, Context) ->
    fun (Save, AS) ->
	    save_terms(Save, Env, Context, AS)
    end.


save_terms([S1 |  S2], Env, Context, AS) ->
    AS1 = save_terms(S1, Env, Context, AS),
    save_terms(S2, Env, Context, AS1);
save_terms({IP, Term}, _Env, _Context, AS) ->
    as:insert(IP, Term, AS);
save_terms([], _, _, AS) ->
    AS.
			    
%%% What used to be module fail

fail_pattern(PatternList, Key, Table, Context, AS) ->
    PTerm = term:make_vector(pattern_list(PatternList)),
    remove(PTerm, Key, Table, Context, AS).

remove(PTerm, Key, Table, Context, AS) ->
    H = history_new(),
    remove(PTerm, Key, H, Table, Context, AS).

remove(PTerm, Key, H, Table, Context, AS) ->
    case term:term_tag(PTerm) of
	pattern_var ->
	    Name = term:pattern_var_name(PTerm),
	    case gb_trees:lookup(Name, Table) of
		none ->
		    {term:join0(), AS};
		{value, Test} ->
		    remove_test(Test, Key,  Context, AS)
	    end;
	alias_pattern ->
	    remove(term:alias_pattern_pattern(PTerm), Key, H, Table, Context, AS);
	functor ->	
            remove_functor(PTerm, Key, H, Table, Context, AS);
	universal ->
	    {term:join0(), AS}
    end.

remove_functor(Pattern, Key, H, Table, Context, AS) ->
    case term:term_tag(Key) of 
	ip ->
            {Key1, AS1} = as:lookup(Key, Context, AS),
            remove_functor(Pattern, Key1, H, Table, Context, AS1);
	universal ->
	    {Key, AS};
	join ->
            TermList = term:join_arg_list(Key),
            {TermList1, AS1} = 
                remove_join_list(Pattern, TermList, H, Table, Context, AS),
            {term:join(TermList1), AS1};
	functor ->
	    PC = term:term_functor(Pattern),
	    TC = term:term_functor(Key),
	    case PC == TC of
		true ->
		    PatternList = term:term_functor_arg_list(Pattern),
		    KeyList = term:term_functor_arg_list(Key),
		    {RemoveList, AS1} = 
			remove_list(PatternList, KeyList, H, Table, Context, AS),
		    TermListList = combine_arg_lists(RemoveList, KeyList),
		    TermList = lists:map(fun(ArgList) ->
						 term:apply_functor(TC, ArgList)
					 end,
					 TermListList),
		    {term:join(TermList), AS1};			
		false ->
		    {Key, AS}
	    end;
	_ ->
	    {Key, AS}
    end.

remove_list([Spec | SpecList], [Term | TermList], H, Table, Context, AS) ->
    {Term1, AS1} = remove(Spec, Term,  H, Table, Context, AS),
    {TermList1, AS2} = remove_list(SpecList, TermList, H, Table, Context, AS1),
    {[Term1 | TermList1], AS2};
remove_list([], [], _, _, _, AS) ->
    {[], AS}.

remove_join_list(Spec, [Term | TermList], H, Table, Context, AS) ->
    {Term1, AS1} = remove(Spec, Term, H, Table, Context, AS),
    {TermList1, AS2} = remove_join_list(Spec, TermList, H, Table, Context, AS1),
    {[Term1 | TermList1], AS2};
remove_join_list(_, [], _, _, _, AS) ->
    {[], AS}.

combine_arg_lists([Remove | RemoveList], [Term | TermList]) ->
    TermListList = combine_arg_lists(RemoveList, TermList),
    TermListList1 = 
	lists:map(fun(List) ->
			  [Term | List]
		  end,
		  TermListList),
    [[Remove | TermList] | TermListList1];
combine_arg_lists([], []) ->
    [].
    
history_new() ->
    {0,0}.


%% Pattern. Builds term-like representation of pattern

pattern(Pattern) ->
    case kinds:pattern_type(Pattern) of
	var ->
	    Var = kinds:name_atom(Pattern),
	    term:make_pattern_var(Var);
	alias ->
	    Var =kinds:name_atom(kinds:alias_var(Pattern)),
	    Pattern1 = kinds:alias_pattern(Pattern),
	    Pattern2 = pattern(Pattern1),
	    term:make_alias_pattern(Var, Pattern2);
	constructor ->
	    constructor_pattern(Pattern)
    end.

pattern_list([]) ->
    [];
pattern_list([E|EList]) ->
    Term = pattern(E),
    TermList = pattern_list(EList),
    [Term|TermList].

    
constructor_pattern(Pattern) ->
    C = kinds:pattern_constructor(Pattern),
    case kinds:constructor_type(C) of
	atomic ->
	    atomic_constructor_pattern(C);
	tuple ->
	    tuple_pattern(Pattern);
	vector ->
	    tuple_pattern(Pattern);
	cons ->
	    cons_pattern(Pattern)
	end.

atomic_constructor_pattern(C) ->
    case kinds:atomic_constructor_type(C) of
	atom -> 
	    term:make_atom(kinds:atom_val(C));
	char  ->
    	    term:make_int(kinds:char_val(C));
	int ->
    	    term:make_int(kinds:int_val(C));
	float ->
    	    term:make_float(kinds:float_val(C)); 
	nil ->
	    term:make_nil();
	string  ->
	    term:make_string(kinds:string_val(C))
    end.

tuple_pattern(Pattern) ->
    ElementList = kinds:constructor_pattern_arg_list(Pattern),
    TermList = pattern_list(ElementList),
    term:make_tuple(TermList).

cons_pattern(Cons) ->
    [Head, Tail] = kinds:constructor_pattern_arg_list(Cons),
    [TermHead, TermTail] = pattern_list([Head, Tail]),
    term:make_cons(TermHead, TermTail).
    

%%% What used to be module guard_test

guard_test(Exp, {Diff, Table}) ->
    case kinds:exp_type(Exp) of
	call ->
	    Module = kinds:call_module(Exp),
	    Name = kinds:call_name(Exp),
	    ArgList = kinds:call_arg_list(Exp),
	    case kinds:is_atom(Module) and kinds:is_atom(Name) of
		true ->
		    guard_call(atom_val(Module), 
			       atom_val(Name), 
			       length(ArgList), ArgList, {Diff, Table});
		false ->
		    {hard, Table}
	    end;
	constructor ->
	    case kinds:guard_type(Exp) of
		true ->
		    {Diff, Table};
		_ ->
		    {hard, Table}
	    end;
	control ->
	    {hard, Table}			    
    end.

atom_val(Exp) ->
    C = kinds:concrete(Exp),
    C.


guard_call(erlang,'and',2,[Exp1, Exp2],{Diff, Table}) ->
    {Diff1, Table1} = guard_test(Exp1, {Diff, Table}),
    guard_test(Exp2, {Diff1, Table1});
guard_call(erlang,is_list,1,[Var],{Diff, Table}) ->
    case kinds:exp_type(Var) of
	var ->
	    Name = kinds:var_name(Var),
	    Test = {'or', {p,term:make_cons(term:universal(),term:universal())},
		    {p, term:make_nil()}},
	    {Diff, add_test(Name, Test, Table)};
	_ ->
	    {hard, Table}
    end;
guard_call(erlang,is_integer,1,[Var],{Diff, Table}) ->
    case kinds:exp_type(Var) of
	var ->
	    Name = kinds:var_name(Var),
	    Test = {p, term:make_int_any()},
	    {Diff, add_test(Name, Test, Table)};
	_ ->
	    {hard, Table}
    end;
guard_call(_, _,_,_, {_Diff, Table}) ->
    {hard, Table}.
			     
add_test(Name, Test, Table) ->
      case gb_trees:lookup(Name, Table) of 
	  none ->
	      gb_trees:insert(Name, Test, Table);
	  {value, Test1} ->
	      gb_trees:update(Name, {'and', Test,Test1}, Table)
      end.


filter({p, Pattern}, Key, Context, AS)  ->
    filter_pattern(Pattern, Key, Context, AS);
filter({'and', Test1, Test2}, Key, Context, AS)  ->
    {Key1, AS1} =  filter(Test1, Key, Context, AS),
    filter(Test2, Key1, Context, AS1);
filter({'or', Test1, Test2}, Key, Context, AS)  ->
    {Term1, AS1} = filter(Test1, Key, Context, AS),
    {Term2, AS2} = filter(Test2, Key, Context, AS1),
    {term:join2(Term1, Term2), AS2}.


filter_pattern(Pattern, Term, Context, AS) ->
    case term:term_tag(Pattern) of
	universal ->
	    {Term, AS};
	functor ->
	    filter_functor(Pattern, Term,  Context, AS)
    end.

filter_functor(Pattern, Key, Context, AS) ->
    case term:term_tag(Key) of 
	ip ->
	    {Term1, AS1} = as:lookup(Key, Context, AS),
	    filter_functor(Pattern, Term1, Context, AS1);
	join ->
	    filter_join(Pattern, Key, Context, AS);
	universal ->
	    Functor = term:term_functor(Pattern),
	    PatternList = term:term_functor_arg_list(Pattern),
	    KeyList = lists:map(fun(_) -> term:universal() end, PatternList),
	    {TermList, AS1} = 
		filter_list(PatternList, KeyList, Context, AS),
	    {term:apply_functor(Functor, TermList), AS1};
	parameter ->
	    {term:join0(), AS};
	functor ->
	    PC = term:term_functor(Pattern),
	    TC = term:term_functor(Key),
	    case PC == TC of
		true ->
	   	    PatternList = term:term_functor_arg_list(Pattern),
		    KeyList = term:term_functor_arg_list(Key),
		    {TermList, AS1} = 
			filter_list(PatternList, KeyList, Context, AS),
		    {term:apply_functor(PC, TermList), AS1};
		false ->
		    {term:join0(), AS}
	    end;
	pid->
	    {term:join0(), AS}
    end.

filter_join(Pattern, Key, Context, AS) ->
    KeyList = term:join_arg_list(Key),
    F = fun (Term, {List, BS}) ->
		{Term1, BS1} = filter_pattern(Pattern, Term, Context, BS),
		{[Term1|List], BS1}
	end,
    {TermList, AS1} = lists:foldl(F, {[], AS}, KeyList),
    {term:join(TermList), AS1}.

filter_list([Pattern|PatternList],[Key|KeyList], Context, AS) ->
    {Term, AS1} = filter_pattern(Pattern, Key, Context, AS),
    {TermList, AS2} = filter_list(PatternList, KeyList, Context, AS1),
    {[Term|TermList], AS2};
filter_list([], [], _Context, AS) ->
    {[], AS}.
		 
remove_test({p, Pattern}, Key, Context, AS)  ->
    remove(Pattern, Key, gb_trees:empty(), Context, AS);
remove_test({'and', Test1, Test2}, Key, Context, AS)  ->
    {Key1, AS1} = remove_test(Test1, Key, Context, AS),
    {Key2, AS2} = remove_test(Test2, Key, Context, AS1),
    {term:join([Key1, Key2]),AS2};
remove_test({'or', Test1, Test2}, Key, Context, AS)  ->
    {Key1, AS1} = remove_test(Test1, Key, Context, AS),
    remove_test(Test2, Key1, Context, AS1).





