%%% File    : present_simp.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : Simplifying structures. Part of the presentation interface
%%% Created : 27 Nov 2001 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(present_simp).
-author('svenolof@harpo.it.uu.se').

% Simplify presentation structures


-export([fuse/1, equal/1, eq/1]).
-export([cleanup/1, purify/1, reify/1]).
-export([replace_ips/2]).
-export([cut/2]).

lookup_ip(Name, Struct) ->
    present_struct:lookup_ip(Name, Struct).


%%% Fuse

fuse(Struct) ->
    {struct, RootTable, IPTable} = Struct,
    List = lists:map(fun({Key,Term}) ->
			     {Key, fuse_term(Term)}
		     end,
		     gb_trees:to_list(IPTable)),
    RootList = lists:map(fun({Key,Term}) ->
			     {Key, fuse_term(Term)}
		     end,
		     gb_trees:to_list(RootTable)),
    {struct, list_to_table(RootList), list_to_table(List)}.
		     
fuse_term(Term) ->
    case term:term_tag(Term) of
	join ->
	    fuse_join(Term);
	_ ->
	    Term
    end.

list_to_table(List) ->
    list_to_table(List, gb_trees:empty()).

list_to_table([{Key, Value} | List], Table) ->
    list_to_table(List, gb_trees:insert(Key, Value, Table));
list_to_table([], Table) ->
    Table.


%%% Equal

equal(Struct) ->
    simp_equal(Struct).

simp_equal(Struct) ->
    TermList = present_struct:struct_term_list(Struct),
    {_N, R0} = build_minimal_relation(TermList),
    L0 = gb_trees:to_list(R0),
    Rn = iter_equal(TermList, L0, R0, Struct),
    rename_struct(Struct, Rn).

build_minimal_relation(TermList) ->
    NewIP = term:make_numbered_ip(1),
    {1, lists:foldl(fun(Term, Table) ->
			    gb_trees:insert(Term, NewIP, Table)
		    end,
		    gb_trees:empty(),
		    TermList)}.
    
iter_equal(TermList, L0, R, Struct) ->
    {_N, R1} = new_relation(TermList, R, Struct),
    L1 = gb_trees:to_list(R1),
    case L1 =/= L0 of
	true ->
	    iter_equal(TermList, L1, R1, Struct);
	false ->
	    R1
    end.

new_relation(TermList, R, Struct) ->
    {N2, R2, _} = 
	lists:foldl(fun (Term, {N, R0, Table0}) ->
			    Term1 = signature(Term, R, Struct),
			    case gb_trees:lookup(Term1, Table0) of
				{value, IP1} ->
				    R1 = gb_trees:insert(Term, IP1, R0),
				    {N, R1, Table0};
				none ->
				    IP1 = term:make_numbered_ip(N),
				    R1 = gb_trees:insert(Term, IP1, R0),
				    Table = gb_trees:insert(Term1, IP1, Table0),
				    {N+1, R1, Table}
			    end
		    end,
		    {0, gb_trees:empty(), gb_trees:empty()},
		    TermList),
    {N2, R2}.


rename_struct(Struct, R) ->
    present_simp:replace_ips(Struct, R).



signature(Term, R, Struct) ->
    signature(Term, gb_trees:empty(), R, Struct).
    
signature(Term, Memory, R, Struct) ->
    Term1 = 
    case term:term_tag(Term) of
	functor ->
	    List = term:term_functor_arg_list(Term),
	    List1 = lists:map(fun(Term0) -> rep(Term0, R) end, List),
	    term:apply_functor(term:term_functor(Term), List1);
	'fun' ->
	    Term;
	join  -> 
	    List = term:join_arg_list(Term),  	    
	    List1 = lists:map(fun(Term0) -> signature(Term0, Memory, R, Struct) end, List),
	    Term2 = term:join(List1),
	    case foo of %term:term_tag(Term2) of 
		join -> fuse_join(Term2);
		_ -> Term2
	    end;	
	universal ->
	    Term;
	parameter ->
	    Term;
	pid ->
	    Term;
	ip ->
	    case gb_trees:is_defined(Term, Memory) of
		true ->
		    exit({should_not_happen,Term});
		false ->
		    Memory1 = gb_trees:insert(Term, true, Memory),
		    signature(present_struct:lookup_ip(Term, Struct), Memory1, R, Struct)    
	    end
    end,
%    io:format("~nSignature of ~w is ~w",[Term, Term1]),
    Term1.

rep(Term, R) ->   
    case gb_trees:lookup(Term, R) of
	{value, T} ->
	    T;
	none ->
	    exit({what_the_heck,'?',Term, R})
    end.


%% Eq unifies ip's that refer to the same term. May be iterated.

eq({struct, NameTable, IP_table}) ->
    {_N, Term_to_NIP} = 
	simp_eq_collect(lists:map(fun({_, Term}) -> Term end, 
				  gb_trees:to_list(IP_table)),
			0,
			gb_trees:empty()), 
    IP_to_NIP = 
	lists:foldl(fun({IP, Term}, Table) -> 
			    {value, NIP} = gb_trees:lookup(Term, Term_to_NIP),
			    case gb_trees:is_defined(IP, Table) of
				true ->
				    Table;
				false ->
				    gb_trees:insert(IP, NIP, Table)
			    end
		    end,
		    gb_trees:empty(),
		    gb_trees:to_list(IP_table)),
    replace_ips({struct, NameTable, IP_table}, IP_to_NIP).
    
simp_eq_collect(TermList, N, Table) ->
    lists:foldl(fun simp_eq_collect_term/2, {N, Table}, TermList).
    
simp_eq_collect_term(Term, {N, Table}) ->
    case gb_trees:is_defined(Term, Table) of
	true ->
	    {N, Table};
	false ->
	    Table_1 = gb_trees:insert(Term, term:make_numbered_ip(N), Table),
	    {N+1, Table_1}
    end.


replace_ips({struct, NameTable, IP_table}, IP_to_NIP) ->
    F = fun({Name, Term}, NameTable_0) ->
		gb_trees:insert(Name, replace_ips_term(Term, IP_to_NIP), NameTable_0)
	end,
    NameTable_1 = lists:foldl(F, gb_trees:empty(), gb_trees:to_list(NameTable)),
    G = fun({IP, Term}, IP_table_0) ->
		{value, IP_1} = gb_trees:lookup(IP, IP_to_NIP),
		case gb_trees:is_defined(IP_1, IP_table_0) of
		    true ->
			IP_table_0;
		    false ->
			gb_trees:insert(IP_1, replace_ips_term(Term, IP_to_NIP), IP_table_0)
		end
	end,
    IP_table_1 = lists:foldl(G, gb_trees:empty(), gb_trees:to_list(IP_table)),
    {struct, NameTable_1, IP_table_1}.

replace_ips_term(Term, Table) ->
    case term:term_tag(Term) of
	functor ->
	    L = lists:map(fun (Term0) ->
				  replace_ips_term(Term0, Table)
			  end, 
			  term:term_functor_arg_list(Term)),
	    term:apply_functor(term:term_functor(Term), L);
	'fun' ->
	    Term;
	join  ->
	    L = lists:map(fun (Term0) ->
				  replace_ips_term(Term0, Table)
			  end,
			  term:join_arg_list(Term)),
	    term:join(L);
	universal ->
	    Term;
	parameter ->
	    Term;
	ip ->
	    {value, Term_1} = gb_trees:lookup(Term, Table),
	    Term_1;
	pid ->
	    MBox = term:pid_mbox(Term),
	    MBox1 = replace_ips_term(MBox, Table),
	    term:make_pid_with_mbox(MBox1)
    end.
		

%%% Reification. Give each union an ip.

reify(Struct) ->
    {struct, RootTable, IPTable} = Struct,
    NextIP = max_ip(Struct) +1,
    {RootList, {NextIP1, New1}} = 
	lists:mapfoldl(fun({Key, Term}, {NextIPx, Newx}) ->
			       {Term1, {NextIPy, Newy}} =  reify_term_0(Term, {NextIPx, Newx}),
			       {{Key, Term1}, {NextIPy, Newy}}
		       end,
		       {NextIP, []},
		       gb_trees:to_list(RootTable)),
    {IPList, {_, New2}} = 
	lists:mapfoldl(fun({Key, Term}, {NextIPz, Newz}) ->
			       {Term1, {NextIPw, Neww}} =  reify_term_0(Term, {NextIPz, Newz}),
			       {{Key, Term1}, {NextIPw, Neww}}
		       end,
		       {NextIP1, New1},
		       gb_trees:to_list(IPTable)),
    {struct, list_to_table(RootList), list_to_table(IPList ++ New2)}.


			 
reify_term_0(Term, {NextIP, IPList}) ->
    case term:term_tag(Term) of
	functor ->
	    reify_functor_term(Term, {NextIP, IPList});
	'fun' ->
	    {Term, {NextIP, IPList}};
	join  -> 
	    reify_join_term_0(Term, {NextIP, IPList});
	universal ->
	    {Term, {NextIP, IPList}};
	parameter ->
	    {Term, {NextIP, IPList}};
	ip ->
	    {Term, {NextIP, IPList}};
	pid ->
	    {Term, {NextIP, IPList}}
    end.

reify_join_term_0(Term, {NextIP, IPList}) ->
    List = term:join_arg_list(Term),
    {List1, {NextIP1, IPList1}} = 
	lists:mapfoldl(
	  fun(Termx, {NextIPx, IPListx}) ->
		  reify_term(Termx, {NextIPx, IPListx})
	  end,
	  {NextIP, IPList},
	  List),
    {term:join(List1),  {NextIP1, IPList1}}.


reify_term(Term, {NextIP, IPList}) ->
    case term:term_tag(Term) of
	functor ->
	    reify_functor_term(Term, {NextIP, IPList});
	'fun' ->
	    {Term, {NextIP, IPList}};
	join  -> 
	    reify_join_term(Term, {NextIP, IPList});
	universal ->
	    {Term, {NextIP, IPList}};
	parameter ->
	    {Term, {NextIP, IPList}};
	ip ->
	    {Term, {NextIP, IPList}};
	pid ->
	    MBox = term:pid_mbox(Term),
	    {MBox1, {NextIP1, IPList1}} = reify_term(MBox, {NextIP, IPList}),
	    {term:make_pid_with_mbox(MBox1), {NextIP1, IPList1}}
    end.

reify_functor_term(Term, {NextIP, IPList}) ->
    Functor = term:term_functor(Term),
    List = term:term_functor_arg_list(Term),
    {List1, {NextIP1, IPList1}} = 
	lists:mapfoldl(
	  fun(Termx, {NextIPx, IPListx}) ->
		  reify_term(Termx, {NextIPx, IPListx})
	  end,
	  {NextIP, IPList},
	  List),
    {term:apply_functor(Functor, List1), {NextIP1, IPList1}}.

reify_join_term(Term, {NextIP, IPList}) ->
    IP = term:make_numbered_ip(NextIP),
    List = term:join_arg_list(Term),
    {List1, {NextIP1, IPList1}} = 
	lists:mapfoldl(
	  fun(Termx, {NextIPx, IPListx}) ->
		  reify_term(Termx, {NextIPx, IPListx})
	  end,
	  {NextIP+1, IPList},
	  List),
    {IP,  {NextIP1, [{IP, term:join(List1)} | IPList1]}}.

%%% Purify makes sure that each root has an IP.

purify(Struct) ->
    {struct, RootTable, IPTable} = Struct,
    NextIP = max_ip(Struct) + 1,
    {RootList, {_, IPTable1}} =
	lists:mapfoldl(
	  fun({Key, Root}, {NextIPx, IPTablex}) ->
		  case term:term_tag(Root) == ip of
		      true ->
			  {{Key, Root}, {NextIPx, IPTablex}};
		      false ->
			  IP = term:make_numbered_ip(NextIPx),
			  IPTabley = gb_trees:insert(IP, Root, IPTablex),
			  {{Key, IP}, {NextIPx+1, IPTabley}}
		  end
	  end,
	  {NextIP, IPTable},
	  gb_trees:to_list(RootTable)),
    {struct, list_to_table(RootList), IPTable1}.
	  
			      
%%% Cleanup makes sure that the value of each IP is a non-ip

cleanup(Struct) ->
%    io:format("~n Cleanup ...",[]),
    {struct, RootTable, IPTable} = Struct,
    IPList = lists:map(fun({IP, Term}) ->
			       Term1 = cleanup_term_0(Term, Struct),
			       {IP, Term1}
		       end,
		       gb_trees:to_list(IPTable)),
%    io:format(" .. done",[]),
    {struct, RootTable, list_to_table(IPList)}.

cleanup_term_0(Term, Struct) ->
    case term:term_tag(Term) of
	ip -> 
	    Term1 = lookup_ip(Term, Struct),
	    cleanup_term_0(Term1, Struct);
	_ ->
	    cleanup_term(Term, Struct)
    end.

cleanup_term(Term, Struct) ->
    case term:term_tag(Term) of
	functor ->
	    cleanup_functor_term(Term, Struct);
	'fun' ->
	    Term;
	join  ->
	    cleanup_join_term(Term, Struct);
	universal ->
	    Term;
	parameter ->
	    Term;
	ip -> 
	    cleanup_ip(Term, Struct);
	pid ->
	    Term
    end.

cleanup_functor_term(Term, Struct) ->
    Functor = term:term_functor(Term),
    List = term:term_functor_arg_list(Term),
    List1 = lists:map(fun(Term1) ->
			      cleanup_term(Term1, Struct)
		      end,
		      List),
    term:apply_functor(Functor, List1).

cleanup_join_term(Term, Struct) ->
    List = term:join_arg_list(Term),
    List1 = lists:map(fun(Term1) ->
			      cleanup_term_0(Term1, Struct)
		      end,
		      List),
    term:join(List1).

cleanup_ip(Term, Struct) ->
    Term1 = lookup_ip(Term, Struct),
    case is_simple_term(Term1) of 
	true ->
	    cleanup_term(Term1, Struct);
	false ->
	    Term
    end.

is_simple_term(Term) ->
    case term:term_tag(Term) of
	functor ->
	    check_depth(Term, 2);
	'fun' ->
	    true;
	join  -> 
	    term:join_arg_list(Term) == [];
	universal ->
	    true;
	parameter ->
	    true;
	ip ->
	    true;
	pid ->
	    true
    end.

check_depth(Term, N) when N> 0 ->
     case term:term_tag(Term) of
	functor ->
	     List = term:term_functor_arg_list(Term),
	     lists:all(fun(Term1) ->
			       check_depth(Term1, N-1)
		       end,
		       List);
	'fun' ->
	    true;
	join  -> 
	     term:join_arg_list(Term) == [];
	universal ->
	    true;
	parameter ->
	    true;
	ip ->
	    false;
	pid ->
	    false
     end;
check_depth(_, _) ->
    false.

% max_ip assumes that all ips are numbered

max_ip(Struct) ->
    {struct, _, IPTable} = Struct,
    lists:foldl(fun({Key, _}, Max) ->
			max(Max, term:numbered_ip_number(Key))
		end,
		-1,
		gb_trees:to_list(IPTable)).
			
max(X, Y) when X >= Y ->
    X;
max(_, Y) ->
    Y.


%%% Fuse

fuse_join(Term) ->
    List = term:join_arg_list(Term),
    fuse_join_list(List, []).

fuse_join_list([], AccList) ->
    term:join(AccList);
fuse_join_list([Term |List], ListAcc) ->
    case term:is_universal(Term) of
	true ->
	    term:universal();
	false ->
	    case find_compatible(Term, ListAcc) of
		{found, Term1, Rest} ->
		    NewTerm = fuse_terms(Term, Term1),
		    fuse_join_list([NewTerm | List], Rest);
		not_found ->
		    fuse_join_list(List, [Term | ListAcc])
	    end
    end.

find_compatible(Term, [Candidate | List]) ->
    case is_compatible(Term, Candidate) of
	true ->
	    {found, Candidate, List};
	false ->
	    case find_compatible(Term, List) of
		{found, Compatible, Rest} ->
		    {found, Compatible, [Candidate | Rest]};
		not_found ->
		    not_found
	    end
    end;
find_compatible(_, []) ->
    not_found.


is_compatible(X, Y) ->
    case disc(X) and disc(Y) of
	false ->
	    true;
	true ->
	    case {term:term_tag(X), term:term_tag(Y)} of
		{functor, functor} ->
		    case term:term_functor(X) == term:term_functor(Y) of
			true ->
			    XList = term:term_functor_arg_list(X),
			    YList = term:term_functor_arg_list(Y),
			    is_compatible_list(XList, YList);
			false ->
			    false
		    end;
		{parameter, parameter} ->
		    term:parameter_name(X) == term:parameter_name(Y);
		{pid, pid} ->
		    true;
		_ ->
		    false
	    end
    end.

is_compatible_list([X | XList], [Y | YList]) ->
    case is_compatible(X, Y) of
	true ->
	    is_compatible_list(XList, YList);
	false ->
	    false
    end;
is_compatible_list([],[]) ->
    true.

fuse_terms(X, Y) ->
    case term:is_universal(X) or term:is_universal(Y) of
	true ->
	    term:universal();
	false ->
	    case (term:term_tag(X)==functor) 
		and (term:term_tag(Y) == functor) of
		true ->
		    F = term:term_functor(X),
		    XList = term:term_functor_arg_list(X),
		    YList = term:term_functor_arg_list(Y),
		    List = fuse_term_lists(XList, YList),
		    term:apply_functor(F, List);
		_ ->
		    term:join2(X, Y)
	    end
    end.

fuse_term_lists([X | XList], [Y | YList]) ->
    Term = fuse_terms(X, Y),
    List = fuse_term_lists(XList, YList),
    [Term | List];
fuse_term_lists([],[]) ->
    [].


disc(X) ->    
    case term:term_tag(X) of
	functor ->
	    true;
	parameter ->
	    true;
	pid ->
	    true;
	_ ->
	    false
    end.
 
% Sometimes a struct is just too big to be resented.
% cut/2 makes sure that at most N IPs are bound to anything but any()

cut(Struct, N) ->
    Table = depth(Struct),
    InvTable = invert_table(Table),
    Depth = find_level(-1, N, gb_trees:to_list(InvTable)),
    cut_struct(Depth, Struct, Table).


depth(Struct) ->
    RootList = present_struct:struct_root_list(Struct),
    RootList1 = lists:map(fun(Root) -> {Root, 0} end, RootList),
    Table = lists:foldl(fun(Root, Table0) ->
				gb_trees:enter(Root, 0, Table0)
			end,
			gb_trees:empty(),
			RootList),				
    iter_depth(RootList1, Table, Struct).

iter_depth(List, Table, Struct) ->
    Table1 = depth_list(List, Table, Struct),
    case updated(Table1, Table) of
	[] ->
	    Table;
	List1 ->
	    iter_depth(List1, Table1, Struct)
    end.

depth_list([{IP, Depth}|List], Table, Struct) ->
    Term = present_struct:lookup_ip(IP, Struct),
    Table1 = depth_term(Term, Depth+1, Table, Struct),
    depth_list(List, Table1, Struct);
depth_list([], Table, _) ->
    Table.

depth_term(Term, Depth, Table, Struct) ->
    case term:term_tag(Term) of
	functor ->
	    List = term:term_functor_arg_list(Term),
	    lists:foldl(fun(Term1, Table1) ->
				depth_term(Term1, Depth, Table1, Struct)
			end,
			Table,
			List);
	'fun' ->
	    Table;
	join  -> 
	    List = term:join_arg_list(Term),
	    lists:foldl(fun(Term1, Table1) ->
				depth_term(Term1, Depth, Table1, Struct)
			end,
			Table,
			List);
	universal ->
	    Table;
	parameter ->
	    Table;
	ip ->
	    case gb_trees:lookup(Term, Table) of
		none ->
		    gb_trees:insert(Term, Depth, Table);
		{value, D} when D =< Depth ->
		    Table;
		{value, _} ->
		    gb_trees:update(Term, Depth, Table)
	    end;
	pid ->
	    Table
    end.

% A list of keys of NewTable whose values differ from those of
% OldTable.

updated(NewTable, OldTable) ->
    List = gb_trees:to_list(NewTable),
    updated_list(List, [], OldTable).

updated_list([{Key, Value}| List], Acc, Table) ->
    case gb_trees:lookup(Key, Table) of
	none ->
	    updated_list(List, [{Key, Value} | Acc], Table);
	{value, Value1} when Value < Value1 ->
	    updated_list(List, [{Key, Value} | Acc], Table);
	_ ->
	    updated_list(List, Acc, Table)
    end;
updated_list([], Acc, _) ->
    Acc.

% Builds inverted table mapping from depth to set of IPs.

invert_table(Table) ->
    List = gb_trees:to_list(Table),
    invert_list(List, gb_trees:empty()).

invert_list([{IP, Depth} | List], Table) ->
    case gb_trees:lookup(Depth, Table) of
	none ->
	    Set0 = gb_trees:empty(),
	    Set1 = gb_trees:insert(IP, true, Set0),
	    Table1 = gb_trees:insert(Depth, Set1, Table),
	    invert_list(List, Table1);
	{value, Set} ->
	    Set1 = gb_trees:insert(IP, true, Set),
	    Table1 = gb_trees:update(Depth, Set1, Table),
	    invert_list(List, Table1)
    end;
invert_list([], Table) ->
    Table.

% Find maximum depth which leaves at most N nodes

find_level(D, N, [{Key, Set} | List]) ->
    N1 = N - gb_trees:size(Set),
    if N1 < 0 ->
	    D;
       true ->
	    find_level(Key, N1, List)
    end;
find_level(D, _, []) ->
    D.

% Remove everything below D.

cut_struct(Depth, Struct, Table) ->
    {struct, NameTable, IP_table} = Struct,
    IP_table1 = 
	lists:foldl(fun({IP, D}, IP_table0) when D > Depth ->
			    gb_trees:update(IP, term:universal(), IP_table0);
		       (_, IP_table0) ->
			    IP_table0
		    end,
		    IP_table,
		    gb_trees:to_list(Table)),
    {struct, NameTable, IP_table1}.
