%%% File    : present_struct.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : Represent analysis information before presentation
%%% Created : 13 Nov 2001 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(present_struct).
-author('svenolof@harpo.it.uu.se').

-export([ collect_roots/3]).
-export([lookup_name/2, lookup_ip/2]).
-export([struct_name_list/1, struct_ip_list/1, struct_term_list/1]).
-export([struct_root_list/1]).
-export([rename/1]).

%% A structure is a pair <Name -> IP, IP -> term>, where the two
%% mappings are gb-trees. 

lookup_name(Name, {struct, NameTable, _}) ->
    {value, Term} = gb_trees:lookup(Name, NameTable),
    Term.

lookup_ip(IP, {struct, _, IP_table}) ->
    {value, Term} = gb_trees:lookup(IP, IP_table),
    Term.

struct_name_list({struct, NameTable, _}) ->
    lists:map(fun({Name, _}) ->
		      Name
	      end,
	      gb_trees:to_list(NameTable)).


struct_root_list({struct, NameTable, _}) ->
    lists:map(fun({_, Root}) ->
		      Root
	      end,
	      gb_trees:to_list(NameTable)).



struct_ip_list({struct, _, IP_table}) ->
    lists:map(fun({IP, _}) ->
		      IP
	      end,
	      gb_trees:to_list(IP_table)).

struct_term_list(Struct = {struct, _, IP_table}) ->
    Table0 = 
	gather_list(struct_ip_list(Struct), gb_trees:empty()),
    Table1 = 
	gather_list(struct_root_list(Struct), Table0),
    Table = 
	gather_list(lists:map(fun ({_, Term0}) -> Term0 end, gb_trees:to_list(IP_table)), Table1),
    lists:map(fun({Term, _}) -> Term end, gb_trees:to_list(Table)).

%%%

collect_roots(NameList, RootList, AS) ->
    TermTable = 
	collect_root_list(RootList, gb_trees:empty(), AS),
    F = fun({Term,_}, {IP_table, N, Term_IP}) ->
		case gb_trees:is_defined(Term, Term_IP) of
		    true ->
			{IP_table, N, Term_IP};
		    false ->
			NewIP = term:make_numbered_ip(N),
			New_IP_table = gb_trees:insert(NewIP, Term, IP_table),
			New_Term_IP = gb_trees:insert(Term, NewIP, Term_IP),
			{New_IP_table, N+1, New_Term_IP}
		end
	end,
    {IP_table, _, Term_IP} =
	lists:foldl(F, 
                    {gb_trees:empty(), 0, gb_trees:empty()}, 
                    gb_trees:to_list(TermTable)),
    H = fun({IP, Term}, New_ip_table) ->
		TermVal = as:lookup(Term, AS),
		gb_trees:insert(IP, rename(TermVal, Term_IP, AS), New_ip_table)
	end,
    New_ip_table = 
	lists:foldl(H, gb_trees:empty(), gb_trees:to_list(IP_table)),
    G = fun({Name, Root}, NameTable) ->
		NewRoot = rename(Root, Term_IP, AS),
		gb_trees:enter(Name, NewRoot, NameTable)
	end,
    NameTable = lists:foldl(G, gb_trees:empty(), zip(NameList, RootList)),
    {struct, NameTable, New_ip_table}.

zip([A|X],[B|Y]) ->
    [{A,B} | zip(X,Y)];
zip([], []) ->
    [].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Collect builds a table of all terms reachable from a root (or list
%% of roots).

collect_root_list([Term | TermList], Table, AS) ->
    Table2 = collect_term(Term, Table, AS),
    collect_root_list(TermList, Table2, AS);
collect_root_list([], Table, _) ->
    Table.


collect_term(Term, Table, AS) ->
    case term:term_tag(Term) of
	functor ->
	    collect_functor_term(Term, Table, AS);
	'fun' ->
	    Table;
	open ->
	    Table;
	join  -> 
	    collect_join_term(Term, Table, AS);
	universal ->
	    Table;
	parameter ->
	    Table;
	ip ->
	    case gb_trees:is_defined(Term, Table) of
		true ->
		    Table;
		false ->
		    Table1 = gb_trees:insert(Term, true, Table),
		    collect_ip(Term, Table1, AS)
	    end;
	pid ->
	    collect_pid(Term, Table, AS)
    end.

collect_functor_term(Term, Table, AS) ->
    List = term:term_functor_arg_list(Term),
    collect_term_list(List, Table, AS).

collect_join_term(Term, Table, AS) ->
    L = term:join_arg_list(Term),
    collect_term_list(L, Table, AS).

collect_ip(Term, Table, AS)  ->
    Term1 = as:lookup(Term, AS),
    collect_term(Term1, Table, AS).

collect_pid(Term, Table, AS) ->
    MBox = term:pid_mbox(Term),
    collect_term(MBox, Table, AS).


collect_term_list([Term | TermList], Table, AS) ->
    Table1 = collect_term(Term, Table, AS),
    collect_term_list(TermList, Table1, AS);
collect_term_list([], Table, _) ->
    Table.


%% Rename


rename(Term, Table, AS) ->
    case term:term_tag(Term) of
		functor ->
		    rename_functor_term(Term, Table, AS);
		'fun' ->
		    Term;
		join  ->
		    rename_join_term(Term, Table, AS);
		universal ->
		    Term;
		parameter ->
		    Term;
 		ip -> 
 		    rename_ip(Term, Table, AS);
		pid ->
		    rename_pid(Term, Table, AS)
    end.



rename_functor_term(Term, Table, AS) ->
    Functor = term:term_functor(Term),
    List = term:term_functor_arg_list(Term),
    List1 = rename_term_list(List, Table, AS),
    term:apply_functor(Functor, List1).
    

rename_join_term(Term, Table, AS) ->
    L = term:join_arg_list(Term),
    L1 = rename_term_list(L, Table, AS),
    term:join(L1).
    

rename_pid(Term, Table, AS) ->
    MBox = term:pid_mbox(Term),
    MBox1=rename(MBox, Table, AS),
    term:make_pid_with_mbox(MBox1).    


rename_ip(Term, Table, _AS) ->
    {value, IP} =  gb_trees:lookup(Term, Table),
    IP.

rename_term_list([Term | TermList], Table, AS) ->
    Term1 = rename(Term, Table, AS),
    TermList1 = rename_term_list(TermList, Table, AS),
    [Term1 | TermList1];
rename_term_list([], _Table, _AS) ->
    [].

%%% Rename names all nodes in a nice depth-first order

rename(Struct) ->
    {struct, RootTable, _IPTable} = Struct,
    NextIP = 0,
    Rename = gb_trees:empty(),
    WL = wl:empty(),
    Table = gb_trees:empty(),
    {RootList, {NextIP1, WL1, Rename1, Table1}} = 
	lists:mapfoldl(
	  fun({Key, Root}, {NextIPx, WLx, Renamex, Tablex}) ->
		  {Rooty, {NextIPy, WLy, Renamey, Tabley}} = 
		      rename_term(Root, {NextIPx, WLx, Renamex, Tablex}),
		  {{Key, Rooty}, {NextIPy, WLy, Renamey, Tabley}}
	  end,
	  {NextIP, WL, Rename, Table},
	  gb_trees:to_list(RootTable)),
    {_, _, _, Table2} = iter_rename({NextIP1, WL1, Rename1, Table1}, Struct),
    {struct, list_to_table(RootList), Table2}.


rename_term(Term, {NextIP, WL,Rename, Table}) ->
    case term:term_tag(Term) of
	functor ->
	    rename_functor_term(Term, {NextIP, WL,Rename, Table});
	'fun' ->
	    {Term, {NextIP, WL,Rename, Table}};
	join  ->
	    rename_join_term(Term, {NextIP, WL,Rename, Table});
	universal ->
	    {Term, {NextIP, WL,Rename, Table}};
	parameter ->
	    {Term, {NextIP, WL,Rename, Table}};
	ip -> 
	    rename_ip(Term, {NextIP, WL,Rename, Table});
	pid ->
	    rename_pid(Term, {NextIP, WL,Rename, Table})
    end.

rename_functor_term(Term, {NextIP, WL,Rename, Table}) ->
    Functor = term:term_functor(Term),
    List = term:term_functor_arg_list(Term),
    {List1, {NextIP1, WL1,Rename1, Table1}} = 
	rename_term_list(List, {NextIP, WL,Rename, Table}),
    {term:apply_functor(Functor, List1), {NextIP1, WL1,Rename1, Table1}}.

rename_join_term(Term, {NextIP, WL,Rename, Table}) ->
    List = term:join_arg_list(Term),
    {List1, {NextIP1, WL1,Rename1, Table1}} = 
	rename_term_list(List, {NextIP, WL,Rename, Table}),
    {term:join(List1), {NextIP1, WL1,Rename1, Table1}}.
    

rename_pid(Term, {NextIP, WL,Rename, Table}) ->
    MBox = term:pid_mbox(Term),
    {MBox1, {NextIP1, WL1, Rename1, Table1}} = 
	rename_term(MBox, {NextIP, WL,Rename, Table}),
    {term:make_pid_with_mbox(MBox1),  {NextIP1, WL1, Rename1, Table1}}.


rename_ip(Term, {NextIP, WL, Rename, Table}) ->
    case gb_trees:lookup(Term, Rename) of
	{value, NewIP} ->
	    {NewIP, {NextIP, WL, Rename, Table}};
	none ->
	    Rename1 = gb_trees:insert(Term, term:make_numbered_ip(NextIP), Rename),
	    WL1 = wl:insert(Term, Term, WL),
	    NextIP1 = NextIP+1,
	    {term:make_numbered_ip(NextIP), {NextIP1, WL1, Rename1, Table}}
    end.

rename_term_list(List, {NextIP, WL,Rename, Table} ) ->
    lists:mapfoldl(
      fun(Term0, {NextIPx, WLx,Renamex, Tablex}) ->
	      rename_term(Term0, {NextIPx, WLx,Renamex, Tablex})
      end,
      {NextIP, WL,Rename, Table},
      List).


iter_rename({NextIP, WL,Rename, Table}, Struct) ->
    case wl:get_work(WL) of
	undefined ->
	    {NextIP, WL,Rename, Table};
	{ok, IP, WL1} ->
	    Term = lookup_ip(IP, Struct),
	    {value, IP1} = gb_trees:lookup(IP, Rename),
	    {Term1, {NextIP2, WL2, Rename2, Table2}} =
		rename_term(Term, {NextIP, WL1, Rename, Table}),
	    Table3 = gb_trees:insert(IP1, Term1, Table2),
	    iter_rename({NextIP2, WL2, Rename2, Table3}, Struct)
    end.
    

list_to_table(List) ->
    list_to_table(List, gb_trees:empty()).

list_to_table([{Key, Value} | List], Table) ->
    list_to_table(List, gb_trees:insert(Key, Value, Table));
list_to_table([], Table) ->
    Table.

%%% Gather collects all subterms of a term in a table

gather(Term, Table) ->
  case gb_trees:is_defined(Term, Table) of
      true ->
	  Table;
      false ->
	  Table1 = gb_trees:insert(Term, true, Table),
	  case term:term_tag(Term) of
	      functor ->
		  List = term:term_functor_arg_list(Term),
		  Table2 = gather_list(List, Table1),
		  Table2;
	      'fun' ->
		  Table1;
	      join  -> 
		  List = term:join_arg_list(Term),  
		  Table2 = gather_list(List, Table1),
		  Table2;
	      universal ->
		  Table1;
	      parameter ->
		  Table1;
	      pid ->
		  Table1,
		  MBox = term:pid_mbox(Term),
		  gather(MBox, Table1);
	      ip ->
		  Table1
	  end
  end.

    
gather_list(TermList, Table) ->
    lists:foldl(fun gather/2, Table, TermList).
