%%% File    : program.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@csd.uu.se>
%%% Purpose : Internal representation of an Erlang program during analysis
%%% Created : 18 Feb 1999 by Sven-Olof Nystr|m <svenolof@csd.uu.se>

%%% Unimplemented: Checking that a function called by a remote call is
%%% indeed exported.

%% {program, ModuleList, DefTable}
%% {module, ModuleName, ExportList, ImportList, FunctionDefList}
%% {fnid, ModuleName, FunctionName, Arity}

-module(program).
-author('svenolof@csd.uu.se').

-export([init/0, program_fnid_table/0]).
-export([def/1, spec/1, typedef/1, attribute/2]).
-export([module_spec_list/1, export_list/1]).
-export([load_mod/2, load_erl_module_list/2, try_module/2]).
-export([make_fnid/3, fnid_module/1, fnid_name/1, fnid_arity/1, show_fnid/1]).
-export([make_tyid/3, tyid_module/1, tyid_name/1, tyid_arity/1, show_tyid/1, display_tyid/1]).
-export([deid_type/1]).
-export([spec_mode/1]).
-export([set_path/1, path/0]).
-export([install_core/1]).
-export([arity_list/2]).


% init/0 -- Initializes the program data base
% def/1 -- Given fnid, returns function definition
% spec/1 -- Given function identifier, return function specification
% typedef/1 -- Given type identifier, return type definition
% program_fnid_table/1 -- only used in module flow
% make_fnid/1, ... -- Create and dissect fnids

% The program data type

empty_program() ->
    {program, {gb_trees:empty(), gb_trees:empty()}, gb_trees:empty(), gb_trees:empty()}.

add_erl_module(MDef, {program, {ETable, STable}, FnidTable, SpecTable}) ->
    MDef1 = annot_mdef(MDef),
    DefList = kinds:mdef_body(MDef1),
    Name = kinds:name_atom(MDef1),
    ETable1 = gb_trees:insert(Name, MDef, ETable),
    FnidTable1 = insert_def_list(DefList, Name, FnidTable),
    {Name, {program, {ETable1, STable}, FnidTable1, SpecTable}}.

add_spec_module(Module, {program, {ETable, STable}, FnidTable, SpecTable}) ->
    {specmodule, ModuleName, Env, SpecList} = Module,
    F = fun(Spec) -> 
		case spec:spec_type(Spec) of
		    funspec ->
			Name = spec:funspec_name(Spec),
			Arity = spec:funspec_arity(Spec),
			Fnid = make_fnid(ModuleName, Name, Arity),
			Fnid;
		    mfunspec ->
			Name = spec:mfunspec_name(Spec),
			Arity = spec:mfunspec_arity(Spec),
			Fnid = make_fnid(ModuleName, Name, Arity),
			Fnid;
		    special ->
			Name = spec:special_name(Spec),
			Arity = spec:special_arity(Spec),
			Fnid = make_fnid(ModuleName, Name, Arity),
			Fnid;
		    function_attribute ->
			none;
		    module_attribute ->
			none;
		    dict ->
			io:format("~n sorry, dictionaries not implemented...",[])
		end
	end,
    FnidList = lists:filter(fun (none) -> false; (_) -> true end,
			    lists:map(F, SpecList)),
    STable1 = gb_trees:insert(ModuleName, true, STable),
    STable2 = gb_trees:insert({fnid_list, ModuleName}, FnidList, STable1), 
    SpecTable1 = insert_spec_list(SpecList, Env, ModuleName, SpecTable),
    SpecTable2 = insert_environment(gb_trees:to_list(Env), ModuleName, SpecTable1),
    {program, {ETable, STable2}, FnidTable, SpecTable2}.

insert_spec_list([Spec | SpecList], Env, ModuleName, Table) ->
    case spec:spec_type(Spec) of
	funspec ->
	    Name = spec:funspec_name(Spec),
	    Arity = spec:funspec_arity(Spec),
	    Fnid = make_fnid(ModuleName, Name, Arity),
	    Table1 = gb_trees:insert(Fnid, {spec,Spec, Env}, Table),
	    insert_spec_list(SpecList, Env, ModuleName, Table1);
	mfunspec ->
	    Name = spec:mfunspec_name(Spec),
	    Arity = spec:mfunspec_arity(Spec),
	    Fnid = make_fnid(ModuleName, Name, Arity),
	    Table1 = gb_trees:insert(Fnid, {spec,Spec, Env}, Table),
	    insert_spec_list(SpecList, Env, ModuleName, Table1);
	special ->
	    Name = spec:special_name(Spec),
	    Arity = spec:special_arity(Spec),
	    Fnid = make_fnid(ModuleName, Name, Arity),
	    Table1 = gb_trees:insert(Fnid, {spec,Spec, Env}, Table),
	    insert_spec_list(SpecList, Env, ModuleName, Table1);
	module_attribute ->
	    Tag = spec:attribute_tag(Spec),
	    Value = spec:attribute_value(Spec),
	    Table1 = gb_trees:insert({module_attribute, ModuleName, Tag}, Value, Table),
	    insert_spec_list(SpecList, Env, ModuleName, Table1);
	function_attribute ->
	    Tag = spec:attribute_tag(Spec),
	    Value = spec:attribute_value(Spec),
	    {Name, Arity} = spec:function_attribute_name_arity(Spec),
	    Fnid = make_fnid(ModuleName, Name, Arity),
	    Table1 =  gb_trees:insert({function_attribute, Fnid, Tag}, Value, Table),
	    insert_spec_list(SpecList, Env, ModuleName, Table1)
    end;

insert_spec_list([], _, _, Table) ->
    Table.
    
insert_environment([{Name, Typedef} | List], Module, Table) ->
    Arity = spec:typedef_arity(Typedef),
    Tyid = make_tyid(Module, Name, Arity),
    Table1 = gb_trees:insert(Tyid, Typedef, Table),
    insert_environment(List, Module, Table1);
insert_environment([], _, Table) ->
    Table.

% Inintialize the program; i.e., throw away all old modules.

init() ->
    P = empty_program(),
    set_current_program(P),
    init_spec_mode(),
    strata:init().
    

% {program, {ErlTable, SpecTable}, FunctionDefinitionTable, SpecificationTable}


% Installing an Erlang module into current program

install_erl_module(MDef) ->
    Program = current_program(),
    {Name, Program1} = add_erl_module(MDef, Program),
    set_current_program(Program1),
    strata:setup_module(Name).

% Install a specification module.

install_spec_module(Module) ->
    Program = current_program(),
    Program1 = add_spec_module(Module, Program),
    set_current_program(Program1).

% Looking up a definition

def(Fnid) ->
    Module = fnid_module(Fnid),
    case is_loaded(Module, erl) of
	true ->
	    true;
	false ->
	    load_mod(Module, erl)
    end,
    Program = current_program(),
    def_1(Fnid, Program).



def_1(Fnid, Program) ->
    Table = program_fnid_table(Program),
    case gb_trees:lookup(Fnid, Table) of
	{value, Def} ->
	    Def;
	none ->
	    exit({unknown_erlang_function,display_fnid(Fnid)})
    end.

% Looking up a specification

spec(Fnid) ->
    Module = fnid_module(Fnid),
    case is_loaded(Module, spec) of
	true ->
	    true;
	false ->
	    load_mod(Module, spec)
    end,
    Program = current_program(),
    {_, STable} = program_module_table(Program),
    case gb_trees:is_defined(Module, STable) of
	true ->
	    spec_1(Fnid, Program);
	false ->
	    case load_mod(Module, spec) of
		true ->
		    spec_1(Fnid, current_program())
	    end
    end.


spec_1(Fnid, Program) ->
    Table = program_spec_table(Program),
    case gb_trees:lookup(Fnid, Table) of
	{value, Spec} ->
	    Spec;
	none ->
	    exit({function_not_specified,display_fnid(Fnid)})
    end.


%%% Looking up a type definition

typedef(Tyid) ->
    Module = tyid_module(Tyid),
    Program = current_program(),
    {_, STable} = program_module_table(Program),
    case gb_trees:is_defined(Module, STable) of
	false ->
	    load_mod(Module, spec),
	    Program1 = current_program();
	_ ->
	    Program1 = Program
    end,
    typedef_1(Tyid, Program1).
	

typedef_1(Tyid, Program) ->
    STable = program_spec_table(Program),
    case gb_trees:lookup(Tyid, STable) of
	{value, Form} ->
	    Form;
	none ->
	    exit({type_not_specified, display_tyid(Tyid)})
    end.


%%% Looking up an attribute

attribute(Tag, Fnid) ->
    Module = fnid_module(Fnid),
    {program, _, _, SpecTable} = current_program(),
    case gb_trees:lookup({function_attribute, Fnid, Tag}, SpecTable) of
	{value, Spec} ->
	    {value, Spec};
	none ->
	    gb_trees:lookup({module_attribute, Module, Tag}, SpecTable)
    end.


%%% Dissecting a program

program_module_table({program, ModuleTable, _FnidTable, _SpecTable}) ->
    ModuleTable.

program_fnid_table() ->
    program_fnid_table(current_program()).

program_fnid_table({program, _ModuleTable, FnidTable, _SpecTable}) ->
    FnidTable.

program_spec_table({program, _ModuleTable, _FnidTable, SpecTable}) ->
    SpecTable.

export_list(Module) ->
    Program = current_program(),
    {ETable, _} = program_module_table(Program),
    {value, MDef} = gb_trees:lookup(Module, ETable),
    kinds:mdef_export_list(MDef).

			  
insert_def_list([], _ModuleName, Table) ->
    Table;
insert_def_list([Def|DefList], ModuleName, Table) ->
    FunctionName = kinds:name_atom(Def),
    Arity = kinds:fdef_arity(Def), 
    Fnid = make_fnid(ModuleName, FunctionName, Arity),
    Table1 = gb_trees:insert(Fnid, Def, Table),
    insert_def_list(DefList, ModuleName, Table1).

% Spec 
% There is some code in file spec_front.erl that should be moved here.

module_spec_list(M) ->
    P = current_program(),
    module_spec_list(M, P).

module_spec_list(M, P) ->
    {_, STable} = program_module_table(P),
    {value, FnidList} = gb_trees:lookup({fnid_list, M}, STable),
    lists:map(fun(Fnid) ->
		      {Fnid, spec(Fnid)}
	      end,
	      FnidList).

%% Fnid (Function identifier)

make_fnid(Module, Function, Arity) 
  when is_atom(Module), is_atom(Function), is_integer(Arity) ->
    {fnid, Module, Function, Arity}.

fnid_module({_, Module, _Function, _Arity}) ->
    Module.

fnid_name({_, _Module, Function, _Arity}) ->
    Function.

fnid_arity({fnid, _Module, _Function, Arity}) ->
    Arity.

show_fnid({_, Module, Function, Arity}) ->
    io_lib:format("~w:~w/~w",[Module, Function, Arity]).

display_fnid({fnid, Module, Function, Arity}) ->
    {Module, Function, Arity}.

is_fnid({fnid, _, _, _}) ->
    true;
is_fnid(_) ->
    false.

%% Tyid (TYpe IDentifier)

make_tyid(Module, Name, Arity) 
  when is_atom(Module), is_atom(Name), is_integer(Arity) ->
    {tyid, Module, Name, Arity}.

tyid_module({tyid, Module, _, _}) ->
    Module.

tyid_name({tyid, _Module, Name, _}) ->
    Name.

tyid_arity({tyid, _Module, _Name, Arity}) ->
    Arity.

show_tyid({tyid, Module, Function, Arity}) ->
    io_lib:format("~w:~w/~w",[Module, Function, Arity]).

display_tyid(Tyid) ->
    {tyid_module(Tyid), tyid_name(Tyid), tyid_arity(Tyid)}.

%% Deid (DEfinition IDentifier)

deid_type({fnid, _, _, _}) ->
    fnid;
deid_type({tyid, _, _, _}) ->
    tyid.

%% Storing and recalling current program

set_current_program(P) ->
    put(current_program, P),
    true.

current_program() ->
    case get(current_program) of
	undefined ->
	    init(),
	    current_program();
	P ->
	    P
    end.
    

%%% Modes
%
% local -- use Erlang source
% extern -- use specification

init_spec_mode() ->
    put(spec_mode, gb_trees:empty()),
    true.


set_spec_mode_local(Module) ->
    Table = get(spec_mode),
    put(spec_mode, gb_trees:insert(Module, local, Table)),
    true.

spec_mode(Module) ->
    Table = get(spec_mode),
    case gb_trees:is_defined(Module, Table) of
	true -> local;
	false -> 
	    case get(external_specs) of
		true ->
		    extern;
		false ->
		    case is_loaded(Module, spec) of
			true ->
			    extern;
			false ->
			    unknown
		    end
	    end
    end.


%% Finding list of arities for given module and function name.

arity_list(Module, Name) ->
    P = current_program(),
    arity_list(Module, Name, P).

arity_list(Module, Name, P) ->
    L = gb_trees:to_list(program_fnid_table(P)) ++ 
	gb_trees:to_list(program_spec_table(P)),
    F = fun({Fnid, _}) ->
		case is_fnid(Fnid) of
		    true ->
			(fnid_module(Fnid) == Module)
			    and (fnid_name(Fnid) == Name);
		    false ->
			false
		end
	end,
    L1 = lists:filter(F, L),
    G = fun({Fnid, _}) ->
		fnid_arity(Fnid)
	end,
    lists:map(G, L1).
			
    
%%% Annotations

annot_mdef(MDef) ->
    MDef1 = number_mdef(MDef),
    find_strings(MDef1).

%%% number node.

number_mdef(MDef) ->
    Pred = fun number_node_p/1,
    F = fun number_node/1,
    trav:trav_mdef(MDef, {Pred, F}).
 
number_node_p(Node) ->
    case kinds:type(Node) of
	mdef ->
	    true;
	fdef ->
	    true;
	clause ->
	    true;
	exp ->
	    case kinds:exp_type(Node) of
		var ->
		    true;
		alias ->
		    true;
		apply ->
		    true;
		local_call ->
		    true;
		call ->
		    true;
		fref ->
		    true;
		control ->
		    case kinds:control_exp_type(Node) of
			'case' ->
			    true;
			'let' ->
			    true;
			letrec ->
			    true;
			'try' ->
			    true;
			'receive' ->
			    true;
			_ ->
			    false
		    end;			
		'fun' ->
		    true;
		_ ->
		    false
	    end;
	_ ->
	    false
    end.

    
number_node(Node) ->
    kinds:add_ann(Node, {number, next_number()}).    

next_number() ->
    case get(node_counter) of
	undefined -> 
	    put(node_counter, 2),
	    1;
	N ->
	    put(node_counter, N+1),
	    N
    end.

find_strings(MDef) ->
    Pred = fun find_strings_p/1,
    F = fun mark_string/1,
    trav:trav_mdef(MDef, {Pred, F}).
 

find_strings_p(Node) ->
    kinds:is_string(Node).

mark_string(Node) ->				     
    kinds:add_ann(Node, {string, true}).

%%%% Interface to the file system.
%
% It seems reasonable to let the program module contain the interface
% to the file system, since it already contains so much functionality
% that is related to the interaction with the file system.

set_path(Path0) ->
    Path = sa_file:normalize(Path0),
    case sa_file:check_path(Path) of
	true ->
	    put(analysis_path, Path);
	Error ->
	    Error
    end.

path() ->
    case get(analysis_path) of
	undefined ->
	    ['.'];
	Path ->
	    Path
    end.


   
load_mod(M, Ext) ->
    Path = path(),
    case sa_file:find_module(Path, M, Ext) of
	{ok, FileName} ->
	    Rep = %sa_loader:
		read_module(FileName, M, Ext),
	    install_module(Rep, Ext),
	    true;
	not_found ->
	    exit({file_not_found,M,Ext})	    
    end.

load_erl_module_list(ModuleList, Include) ->
    Path = path(),
    program:init(),
    Program = current_program(),
    NameList = leml(ModuleList, Path, Program, Include),
    lists:foreach(fun set_spec_mode_local/1, NameList),
    strata:setup_module_list(NameList).

leml([Module|ModuleList], Path, Program, Include) ->
    Ext = erl,
    case sa_file:find_module(Path, Module, Ext) of
	{ok, FileName} ->
	    MDef = read_module_i(FileName, Module, Ext, Include),
	    {Name, Program1} = add_erl_module(MDef, Program),
	    [Name|leml(ModuleList, Path, Program1, Include)];
	not_found ->
	    exit({file_not_found,Module,Ext})	
    end;
leml([], _, Program, _) ->
    set_current_program(Program),
    [].

install_module(Rep, erl) ->
    install_erl_module(Rep);
install_module(Rep, spec) ->
    install_spec_module(Rep).


is_loaded(M, erl) ->
    Program = current_program(),
     {ETable, _} = program_module_table(Program),
    gb_trees:is_defined(M, ETable);
is_loaded(M, spec) ->
    Program = current_program(),
     {_, STable} = program_module_table(Program),
    gb_trees:is_defined(M, STable).

try_module(M, Ext) ->
    case is_loaded(M, Ext) of
	true ->
	    true;
	false ->
	    Path = path(),
	    case sa_file:find_module(Path, M, Ext) of
		{ok, FileName} ->
		    Rep = %sa_loader:
			read_module(FileName, M, Ext),
		    install_module(Rep, Ext),
		    true;
		not_found ->
		    false
	    end
    end.

% This used to be a separate module...
% -module(sa_loader).
% -author('svenolof@harpo.it.uu.se').
%
% -export([read_module/3]).

read_module(FileName, Module, erl) ->
    io:format("~nLoading Erlang module ~s ", [FileName]),
    kinds:read_module(FileName, Module);
read_module(FileName, _Module, spec) ->
 io:format("~nLoading Specification module ~s ", [FileName]),
    M = spec_front:read_file(FileName),
    case M of
	{specmodule, _Name, _Env, _SpecList} ->
	    M
    end.

read_module_i(FileName, Module, erl, Include) ->
    io:format("~nLoading Erlang module ~s ", [FileName]),
    kinds:read_module(FileName, Module, Include).
    
    
%%%

install_core(Core) ->
    install_module(Core, erl),
    list_to_atom(kinds:mdef_name(Core)).    
