%%% File    : report.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : Error reporting
%%% Created : 18 Jan 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(report).
-author('svenolof@harpo.it.uu.se').

%% When the analysis discovers any potential errors, they must be
%% reported. Since the analysis repeatedly traverses different parts
%% of the program, we collect the errors and present them in a single
%% traversal.

-export([report_exception/3, report_bad_arg/3, report_bad_arg/4, report_bad_fun_arg/2]).
-export([report_bad_result/3,report_unused_clause/3, report_unknown_module/2,
	report_bad_dict_entry/3, report_bad_dict_key/3]).
-export([present/1, present/2, number_of_errors/1]).


report_exception(Fnid, Exception, AS) ->
    case is_quiet(AS) of
	true ->
	    AS;
	false ->
	    M = {exception, Fnid, Exception},
	    insert_report(Fnid, M, AS)
    end.

report_bad_arg(Fnid, Callee, AS) ->
    M = {bad_arg, Fnid, Callee},
    insert_report(Fnid, M, AS).

report_bad_arg(Fnid, Callee, N, AS) ->
    M = {bad_arg, Fnid, Callee, N},
    insert_report(Fnid, M, AS).

report_bad_result(Fnid, Result, AS) ->
    M = {bad_result, Fnid, Result},
    insert_report(Fnid, M, AS).

report_bad_fun_arg(Fnid, AS) ->
    M = {bad_fun_arg, Fnid},
    insert_report(Fnid, M, AS).

report_unused_clause(Fnid, Line, AS) ->
    M = {unused_clause, Fnid, Line},
    insert_report(Fnid, M, AS).

report_unknown_module(Module, AS) ->
    M = {unknown_module, Module},
    insert_report(none, M, AS).

report_bad_dict_entry(Fnid, Entry, AS) ->
    M = {bad_dict_entry, Fnid, Entry},
    insert_report(Fnid, M, AS).

report_bad_dict_key(Fnid, Entry, AS) ->
    M = {bad_dict_key, Fnid, Entry},
    insert_report(Fnid, M, AS).

is_quiet(AS) ->
    as:dict_get(quiet, AS).


insert_report(Fnid, M, AS) ->
    Message = present_message(M, AS),
    case filter(Message, AS) of
	true ->
	    AS;
	false ->
	    as:report(Fnid, Message, AS)
    end.

filter(M, AS) ->
    warn_filter:filter(M, AS).
       
number_of_errors(AS) ->
     gb_trees:size(as:get_reports(AS)).

present(AS) ->
    present(standard_io, AS).

present(H, AS) ->
    io:format(H, "~nNew error report:",[]),
    ReportTable = as:get_reports(AS),
    ReportList = gb_trees:to_list(ReportTable),
    present_report_list(ReportList, H, AS),
    NumberOfErrors = 
	lists:sum(lists:map(fun({_, Table}) -> gb_trees:size(Table) end, ReportList)),
    io:format(H, "~nTotal: ~w errors in ~w functions",
	      [NumberOfErrors, length(ReportList)]).
    


present_report_list([{none, MessageTable} | ReportList], H, AS) ->
    MessageList = lists:map(fun({Key, _}) -> Key end,
			    gb_trees:to_list(MessageTable)),
    present_message_list(MessageList, H, AS),
    present_report_list(ReportList, H, AS);
present_report_list([{Fnid, MessageTable} | ReportList], H, AS) ->
    io:format(H, "~n~nFunction ~s", [program:show_fnid(Fnid)]), 
    MessageList = lists:map(fun({Key, _}) -> Key end,
			    gb_trees:to_list(MessageTable)),
    present_message_list(MessageList, H, AS),
    present_report_list(ReportList, H, AS);
present_report_list([], _, _) ->
    true.

present_message_list([Message | MessageList], H, AS) ->
    % present_message(Message, H, AS),
    io:put_chars(H, Message),
    present_message_list(MessageList, H, AS);
present_message_list([], _, _) ->
    true.


present_message({exception, _, Exception}, AS) ->
    MessString = present:sketch(Exception, AS),
    io_lib:format("~nException: ~n~s", [MessString]);
present_message({bad_arg, _, Callee}, _) ->
    io_lib:format("~nBad argument in call to ~s", [program:show_fnid(Callee)]);
present_message({bad_arg, _, Callee, N}, _) ->
    io_lib:format("~nBad argument ~w in call to ~s", 
                  [N, program:show_fnid(Callee)]);
present_message({bad_result, _, Result}, AS) ->
    io_lib:format("~nBad result: ~n~s", [present:sketch(Result, AS)]);
present_message({unknown_module, M}, _) ->
    io_lib:format("~nUnknown Module: ~s", [M]);
present_message({bad_dict_entry, _, Term}, AS) ->
    io_lib:format("~nBad entry in process dictionary: ~n~s", [present:sketch(Term, AS)]);
present_message({bad_dict_key, _, Term}, AS) ->
    io_lib:format("~nBad key in process dictionary: ~n~s", [present:sketch(Term, AS)]);
present_message(M, _) ->
    io_lib:format("~n~p", [M]).


