-module(show).
						
% Pretty printing interface

-export([struct_to_string/1, show_term/1, nice_var_name/1]).

    
struct_to_string(S) ->
    NameList = present_struct:struct_name_list(S),
    F = fun(Name) ->
		PName = make_printable_name(Name),
		Term = present_struct:lookup_name(Name, S),
		TermTree = show_term(Term),
		Tree = {match, 0, PName, TermTree},
		[erl_pp:exprs([Tree],0,hook), "\n"]
	end,
    A = lists:map(F, NameList),
    IP_list = lists:sort(lists:map(fun(Name) ->
					   {term:numbered_ip_number(Name), Name}
				   end,
				   present_struct:struct_ip_list(S))),
    B = if length(IP_list) > 0 ->
		G = fun({_, IP}) ->
			    IP_name = show_ip_ref_term(IP),
			    Term = present_struct:lookup_ip(IP, S),
			    TermTree = show_term(Term),
			    Tree = {match, 0, IP_name, TermTree},
			    [erl_pp:exprs([Tree],0,hook), "\n"]
		    end,
		"\nwhere \n" ++
		    lists:map(G, IP_list);
	   true  ->
		[]
	end,
    lists:flatten([A,B]).

show_term(Term) ->
    case term:term_tag(Term) of 
	functor ->
	    show_functor_term(Term);
	'fun' ->
	    show_fun_term(Term);
	ip ->
	    show_ip_ref_term(Term);
	join ->
    	    show_join_term(Term);
	universal ->
    	    show_universal_term(Term);
	parameter ->
    	    show_parameter_term(Term);
	pid ->
	    show_pid_term(Term)	   ;
	trunc ->
	    {trunc}
    end.

show_functor_term(Term) ->
    Functor = term:term_functor(Term),
    List = term:term_functor_arg_list(Term),
    FunctorName = list_to_atom(lists:flatten(io_lib:format("~w",[Functor]))),
    show_functor(Functor, List).

show_functor(atomc, [Term]) ->
    case term:term_tag(Term) of
	universal ->
	     {call, 0, {atom, 0, atom},[]};
	functor ->
	    {atoma, Atom} = term:term_functor(Term),
	    {atom, 0, Atom}
    end;
show_functor(intc, [Term]) ->
    case term:term_tag(Term) of
	universal ->
	     {call, 0, {atom, 0, int},[]};
	functor ->
	    {inta, Int} = term:term_functor(Term),
	    {integer, 0, Int};
	_ ->
	    show_general_functor(intc, [Term])
    end;
show_functor(floatc, [Term]) ->
    case term:term_tag(Term) of
	universal ->
	     {call, 0, {atom, 0, float},[]}
    end;
show_functor(tuple, [Term]) ->
     case is_clean_list(Term) of
	true ->
	    {tuple, 0, show_clean_list(Term)};
	false ->
	    {call, 0, {atom, 0, tuple}, [show_term(Term)]}
    end;
show_functor(cons, TermList) ->
    [Hd, Tl] = show_list(TermList),
    {cons, 0, Hd, Tl};
show_functor(nil, []) ->
    {nil, 0};
show_functor({formal, M, N, _}, List) ->
    List1 = show_list(List),
    {call, 0, {remote, 0, {atom, 0, M}, {atom, 0, N}}, List1};
show_functor(FunctorName, List) ->
    show_general_functor(FunctorName, List).

show_general_functor(FunctorName, List) ->
    List1 = show_list(List),
    {call, 0, {atom, 0, FunctorName},List1}.


is_clean_list(Term) ->
    case term:term_tag(Term) of
	functor ->
	    case term:term_functor(Term) of
		cons ->
		    [_, Tl] = term:term_functor_arg_list(Term),
		    is_clean_list(Tl);
		nil ->
		    true;
		_ ->
		    false
	    end;
	_ ->
	    false
    end.

show_clean_list(Term) ->
    case term:term_tag(Term) of
	functor ->
	    case term:term_functor(Term) of
		cons ->
		    [Hd, Tl] = term:term_functor_arg_list(Term),
		    [show_term(Hd) |
		     show_clean_list(Tl)];
		nil ->
		    []
	    end
    end.


	
show_fun_term(Fun) ->
    Context = term:fun_ap(Fun),
    Fnid = context:current_fnid(Context),

    {tuple, 0, [{atom, 0, 'fun'},
		show_fnid(Fnid)]}.

show_fnid(Fnid) ->
    Module = program:fnid_module(Fnid), 
    Name = program:fnid_name(Fnid),
    Arity = program:fnid_arity(Fnid),
    {tuple, 0, [{atom, 0, fnid},
		{atom, 0, Module},
		{atom, 0, Name},
		{integer, 0, Arity}]}.


show_pid_term(Term) ->
    Start = term:pid_location(Term),
    MBox = show_term(term:pid_mbox(Term)),
    {tuple, 0, [{atom, 0, pid}, {atom, 0, Start}, MBox]}.


show_ip_ref_term(Term) ->
    N =  term:numbered_ip_number(Term),
    nice_var_name_tree(N).

nice_var_name_tree(N) ->
    {var, 0, nice_var_name(N)}.

nice_var_name(N) when N =< 25 ->
    [65+N];
nice_var_name(N) ->
    C = N rem 26,
    I = N div 26,
    [65+C] ++ integer_to_list(I).

show_join_term(Term) ->
    case term:join_arg_list(Term) of
	[] ->
	     {call, 0, {atom, 0, none},[]};
	[T] ->
	    show_term(T);
	TermList ->
	    L = lists:map(fun(Term1) ->
				  show_term(Term1)
			  end,
			  TermList), 
	    show_join_1(L)
    end.


show_join_1([T]) ->
    T;
show_join_1([T | L]) ->
    show_join_2(T, L).

show_join_2(T, []) ->
    T;
show_join_2(T, [U | L]) ->
    show_join_2({op,0,'+',T, U}, L).


show_list(L) ->
    lists:map(fun show_term/1, L).

show_universal_term(_) ->
    {call, 0, {atom, 0, any},[]}.

show_parameter_term(T) ->
    {call, 0, {atom, 0, parameter}, [{atom, 0, term:parameter_name(T)}]}.


%% Should really do something clever with hooks for any-atom, any-int,
%% empty, and union.

make_printable_name(Name) ->
    {atom, 0, list_to_atom(lists:flatten(io_lib:format("~W",[Name,5])))}.

%%% 


