%%% File    : parse_spec.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : Read specifiations from file, parse and transform them.
%%% Created : 12 Jan 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(spec_front).
-author('svenolof@harpo.it.uu.se').

-export([read_file/1]).

%% close(Epp)
%% parse_erl_form(Epp)
%% macro_defs(Epp)


open(File, Path, Pdm) ->
    epp:open(File, Path, Pdm).

close(Epp) ->
    epp:close(Epp).

parse_erl_form(Epp) ->
    case epp:scan_erl_form(Epp) of
	{ok,Toks} ->
	    spec_parse:parse_form(Toks);
	Other ->
	    Other
    end.

parse_file(Ifile, Path, Predefs) ->
    case open(Ifile, Path, Predefs) of
	{ok,Epp} ->
	    Forms = parse_file(Epp),
	    close(Epp),
	    {ok,Forms};
	{error,E} ->
	    {error,E}
    end.

parse_file(Epp) ->
    case parse_erl_form(Epp) of
 	{ok,TForm} ->
 	    Form = translate_form(TForm),
	    [Form|parse_file(Epp)];
	{error,E} ->
	    [{error,E}|parse_file(Epp)];
 	{eof,Line} ->
 	    [{eof,Line}]
    end.

%%% Specifications.


read_file(FileName) ->    
    {ok, L} = parse_file(FileName, [], []),
    case lists:any(fun({error, _}) -> true;
		      (_) -> false
		      end,
		      L) of
	false ->
	    {Name,Env, SpecList} = process_spec_module(L),
	    {specmodule, Name, Env, SpecList};
	true ->
	    ErrorList = specification_errors(FileName, L),
	    io:nl(),
	    exit({bad_specification_module, FileName})
    end.

process_spec_module([_, {module, Name} | FormList]) ->
    {Env, FunSpecList} = 
	process_spec_module(FormList, gb_trees:empty(), []),
    {Name, Env, FunSpecList}.

process_spec_module([Form | FormList], Env, FunSpecList) ->
    case spec:spec_form_type(Form) of
	typedef ->
	    Name = spec:typedef_name(Form),
	    Env1 = gb_trees:insert(Name, Form, Env),
	    process_spec_module(FormList, Env1, FunSpecList);
	abstypedef ->
	    Name = spec:typedef_name(Form),
	    Env1 = gb_trees:insert(Name, Form, Env),
	    process_spec_module(FormList, Env1, FunSpecList);
	public_typedef ->
	    Name = spec:typedef_name(Form),
	    Env1 = gb_trees:insert(Name, Form, Env),
	    process_spec_module(FormList, Env1, FunSpecList);
	funspec ->
	    process_spec_module(FormList, Env, [Form | FunSpecList]);
	mfunspec ->
	    process_spec_module(FormList, Env, [Form | FunSpecList]);
	special ->
	    process_spec_module(FormList, Env, [Form | FunSpecList]);
	dict ->
	    process_spec_module(FormList, Env, [Form | FunSpecList]);
	function_attribute ->
	    process_spec_module(FormList, Env, [Form | FunSpecList]);
	module_attribute ->
	    process_spec_module(FormList, Env, [Form | FunSpecList]);
	file ->
	    process_spec_module(FormList, Env, FunSpecList);
	module ->
	    process_spec_module(FormList, Env, FunSpecList);
	eof ->
	    process_spec_module(FormList, Env, FunSpecList) 
    end;
process_spec_module([], Env, FunSpecList) ->
    {Env, lists:reverse(FunSpecList)}.

specification_errors(File, L) ->
    io:format("~nSyntax errors in file ~s", [File]),
    lists:foreach(fun({error, {Line, _, S}}) ->
			  ok = io:format("~nLine ~w: ~s", [Line, S]);
		     (_) ->
			  true
		  end,
		  L).


%%% Translate. Given syntax tree grenerated by parser, produce
%%% internal representation of types and type definitions.

translate_typedef({type, _, def, {TName, TArgList}, TType}) ->
    Name = translate_name(TName),
    ArgList = translate_arg_list(TArgList),
    Type = translate_type(TType),
    spec:make_typedef(Name, ArgList, Type);
translate_typedef({abstype, _, def, {TName, TArgList}, TType}) ->
    Name = translate_name(TName),
    ArgList = translate_arg_list(TArgList),
    Type = translate_type(TType),
    spec:make_abstypedef(Name, ArgList, Type);
translate_typedef({public_type, _, def, {TName, TArgList}, TType}) ->
    Name = translate_name(TName),
    ArgList = translate_arg_list(TArgList),
    Type = translate_type(TType),
    spec:make_public_typedef(Name, ArgList, Type).


translate_name(Atom) ->
    Atom.

translate_arg_list([TArg | TArgList]) ->
    Arg = translate_arg(TArg),
    ArgList = translate_arg_list(TArgList),
    [Arg | ArgList];
translate_arg_list([]) ->
    [].

translate_arg(Arg) ->
    translate_type(Arg).

translate_type({'fun', TArgList, TResult, Pos}) ->
    ArgList = translate_arg_list(TArgList),
    Result = translate_type(TResult),
    spec:make_fun(ArgList, Result, Pos);
translate_type({utype, TTypeList})  ->
    translate_union(TTypeList);
translate_type(Var) when atom(Var) ->
    spec:make_var(Var);
translate_type({alias, Var, T}) ->
    spec:make_alias(Var, translate_type(T));
translate_type({atom, _, Atom}) ->
    spec:make_functor(atomc, [spec:make_functor({atoma, Atom}, [])]);
translate_type({integer, _, Int}) ->
    spec:make_functor(intc, [spec:make_functor({inta, Int}, [])]);
translate_type({int, {integer, _, Int}}) ->
    spec:make_functor(intc, [spec:make_functor({inta, Int}, [])]);
translate_type({tuple, TTypeList}) ->
    TypeList = lists:map(fun translate_type/1, TTypeList),
    spec:make_functor(tuple, [list_to_spec_list(TypeList)]);
translate_type({cons, T1, T2}) ->
    S1 = translate_type(T1),
    S2 = translate_type(T2),
    spec:make_functor(cons,[S1,S2]);
translate_type({cons, _, T1, T2}) ->
    S1 = translate_type(T1),
    S2 = translate_type(T2),
    spec:make_functor(cons,[S1,S2]);
translate_type({nil, _}) ->
    spec:make_functor(nil, []);
translate_type({type, Name, TArgList}) ->
    ArgList = translate_arg_list(TArgList),
    case is_special_struct(Name) of
	true ->
	    translate_special(Name, ArgList);
	false ->
	    spec:make_apply(Name, ArgList, next_apply_pos())
    end;
translate_type({external, Module, Name, TArgList}) ->
    ArgList = translate_arg_list(TArgList),
    spec:make_external(Module, Name, ArgList, next_apply_pos()).
    
list_to_spec_list([Type | List]) ->
    spec:make_functor(cons, [Type, list_to_spec_list(List)]);
list_to_spec_list([]) ->
    spec:make_functor(nil, []).


%% A auick hack to make expansion of type application work.

next_apply_pos() ->
    case get(apply_pos) of
	undefined ->
	    put(apply_pos,1),
	    0;
	N ->
	    put(apply_pos, N+1),
	    N
    end.

%%

translate_union([TType]) ->
    translate_type(TType);
translate_union(L) ->
    spec:make_union_spec(lists:map(fun translate_type/1, L)).
					  
is_special_struct(Name) ->					  
    lists:member(Name, [int, float, atom, tuple, any, none, pid, exception]).

translate_special(atom, []) ->
    spec:make_functor(atomc, [spec:make_universal()]);
translate_special(int, []) ->
    spec:make_functor(intc, [spec:make_universal()]);
translate_special(float, []) ->
    spec:make_functor(floatc, [spec:make_universal()]);
translate_special(tuple, [T]) ->
    spec:make_functor(tuple, [T]);
translate_special(any, []) ->
    spec:make_universal();
translate_special(none, []) ->
    spec:make_union_spec([]);
translate_special(pid,[]) ->
    spec:make_pid_spec();
translate_special(pid, [T]) ->
    spec:make_pid_spec(T);
translate_special(exception, [T]) ->
    spec:make_exception_spec(T).

translate_funspec({function, Line, Name, Arity, [Clause]}) ->
    translate_single_clause(Name, Arity, Clause);
translate_funspec({function, Line, Name, Arity, ClauseList}) ->
    translate_multi_clause(Name, Arity, ClauseList, Line).


translate_single_clause(Name, Arity, {clause, Line, TArgList, TGuard, TType}) ->
    case TType of
	{utype, [{type, special_definition, []}]} ->
	    spec:make_special(Name, Arity);
	_ ->
	    ArgList = translate_arg_list(TArgList),
	    Guard = translate_guard_list(TGuard),
	    Type = translate_type(TType),
	    spec:make_funspec(Name, ArgList, Guard, Type, Line)
    end.
translate_multi_clause(Name, Arity, ClauseList, Line) ->
    ClauseList1 = lists:map(fun translate_clause/1, ClauseList),
    spec:make_mfunspec(Name, Arity, ClauseList1, Line).

translate_clause( {clause, Line, TArgList, [], TType}) ->
    ArgList = translate_arg_list(TArgList),
    Type = translate_type(TType),
    spec:make_clause(ArgList, Type, Line).

translate_guard_list([Guard | GuardList]) ->
    [translate_guard(Guard) | translate_guard_list(GuardList)];
translate_guard_list([]) ->
    [].

translate_guard({guard, Name, ArgList}) ->
    spec:make_apply(Name, translate_arg_list(ArgList), next_apply_pos()).


translate_dict({dictionary_entry, Line, TKey, TType}) ->
    Key = translate_type(TKey),
    Type = translate_type(TType),
    spec:make_dict(Key, Type).

translate_se(Tag, Spec) ->
    spec:make_module_attribute(Tag, translate_type(Spec)).

translate_se(Tag, LFnid, Spec) ->
    spec:make_function_attribute(Tag, LFnid, translate_type(Spec)).

translate_de(Tag, Spec) ->
    spec:make_module_attribute(Tag, translate_clauses(Spec)).

translate_de(Tag, LFnid, Spec) ->
    spec:make_function_attribute(Tag, LFnid, translate_clauses(Spec)).

translate_clauses(L) ->
    lists:map(fun({S1, S2}) ->
		      {translate_type(S1), translate_type(S2)}
	      end,
	      L).



translate_form(Form) ->
    case Form of
	{attribute, _, file, FileName} ->
	    {file, FileName};
	{attribute, _, module, ModuleName} ->
	    {module, ModuleName};
	{eof, _} ->
	    Form;	
	{type, _,_,_,_} ->
	    translate_typedef(Form);
	{abstype, _,_,_,_} ->
	    translate_typedef(Form);
	{public_type, _,_,_,_} ->
	    translate_typedef(Form);
	{function, _, _, _, _, _} ->
	    translate_funspec(Form);
	{function, _, _, _, _} ->
	    translate_funspec(Form);
	{special_entry, Tag, Spec} ->
	    translate_se(Tag, Spec);
	{special_entry, Tag, LFnid, Spec} ->
	    translate_se(Tag, LFnid, Spec);
	{dict_entry, Tag, Spec} ->
	    translate_de(Tag, Spec);
	{dict_entry, Tag, LFnid, Spec} ->
	    translate_de(Tag, LFnid, Spec)
    end.





