%%% File    : spec_generate.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Description : Generate from specification
%%% Created : 21 Aug 2002 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(spec_generate).

-export([generate_list/4, generate/4]).

generate(Spec, Env, Context, AS) ->
    case spec:spec_type(Spec) of
	union ->
	    generate_union(Spec, Env, Context, AS);
	universal ->
	    generate_universal(Spec, Env, Context, AS);
	var ->
	    generate_var(Spec, Env, Context, AS);
	apply ->
	    generate_apply(Spec, Env, Context, AS);
	'fun' ->
	    generate_fun(Spec, Env, Context, AS);
	external ->
	    generate_external(Spec, Env, Context, AS);
	functor ->
	    generate_functor(Spec, Env, Context, AS);
	pid ->
	    generate_pid(Spec, Env, Context, AS)
    end.


generate_list([Spec | SpecList], Env, Context, AS) ->
    {Term, AS1} = generate(Spec, Env, Context, AS),
    {TermList, AS2} = generate_list(SpecList, Env, Context, AS1),
    {[Term | TermList], AS2};
generate_list([], _, Context, AS) ->
    {[], AS}.

generate_functor(Spec, Env, Context, AS) ->
    Functor = spec:spec_functor(Spec),
    SpecList = spec:spec_functor_arg_list(Spec),
    {TermList, AS1} = generate_list(SpecList, Env, Context, AS),
    {term:apply_functor(Functor, TermList), AS1}.

generate_fun(Spec, Env, Context, AS) ->
    Pos = spec:pos(Spec),
    ArgList = spec:fun_arg_list(Spec), 
    Arity = length(spec:fun_arg_list(Spec)),
    FunContext = Context,
    Result = spec:fun_result(Spec),
    TermList = param:invent_fun_arg_list(Arity, FunContext, Pos),
    case spec_match:match_list(ArgList, TermList, Env, Context, AS) of
	{ok, AS1} ->
	    {ResultTerm, AS2} = generate(Result, Env, Context, AS1),
	    AS3 = as:insert(tag:fun_result(FunContext, Pos), ResultTerm, AS2);
	{fail, AS2} ->
	    Fnid = context:current_fnid(Context),
	    AS3 = report:report_bad_fun_arg(Fnid, AS2)
    end,
    Fun = term:make_fun(FunContext, Pos, Arity),
    {Fun, AS3}.

    
generate_union(Spec, Env, Context, AS) ->
    List = spec:union_arg_list(Spec),
    {TermList, AS1} = generate_list(List, Env, Context, AS),
    {term:join(TermList), AS1}.

generate_universal(Spec, Env, Context, AS) ->
    {term:universal(), AS}.

generate_var(Spec, Env, Context, AS) ->
    Name = spec:var_name(Spec),
    Term = env:lookup(Name, Env),
    {Term, AS}.

generate_apply(Spec, Env, Context, AS) ->
    Pos = spec:pos(Spec),
    Module = context:current_module(Context),
    Name = spec:apply_name(Spec),
    ArgList = spec:apply_arg_list(Spec),
    Arity = length(ArgList),
    {TermList, AS1} = generate_list(ArgList, Env, Context, AS),
    case is_special(Name, Arity) of 
	true ->
	    generate_special:generate_special(Name, Arity, TermList, Pos, Context, AS1);
	false ->
	    Tyid = program:make_tyid(Module, Name, Arity), 
	    Form = program:typedef(Tyid),
	    case {spec:spec_form_type(Form), program:spec_mode(Module)} of
		{abstypedef, extern} ->
		    generate_formal(Tyid, Context, AS1, TermList);
		{_, _} ->
		    generate_local(Tyid, Pos, Context, AS1, TermList)
	    end
    end.

is_special(nth,2) ->
    true;
is_special(set_nth,3) ->
    true;
is_special(apply, 3) ->
    true;
is_special(_, _) ->
    false.

generate_external(Spec, Env, Context, AS) ->
    Module = spec:external_module(Spec),
    Name = spec:external_name(Spec),
    ArgList = spec:external_arg_list(Spec),
    Arity = length(ArgList),
    {TermList, AS1} = generate_list(ArgList, Env, Context, AS),
    Tyid = program:make_tyid(Module, Name, Arity), 
    Form = program:typedef(Tyid),
    case spec:spec_form_type(Form) of
	abstypedef ->	
	    generate_formal(Tyid, Context, AS1, TermList);
	typedef ->
	    ThisModule = context:current_module(Context),
	    exit({bad_abstype, program:display_tyid(Tyid), ThisModule});
	public_typedef ->
	    Pos = spec:pos(Spec),
	    generate_local(Tyid, Pos, Context, AS1, TermList)
    end.


generate_formal(Tyid, Context, AS, ArgList) -> 
    Module = program:tyid_module(Tyid),
    Name = program:tyid_name(Tyid),
    Arity = program:tyid_arity(Tyid),
    {term:make_formal(Module, Name, Arity, ArgList), AS}.

generate_local(Tyid, Pos, Context, AS1, TermList) ->
    {NewContext, AS2} = work:call_generation(Tyid, Pos, Context, AS1),
    AS3 = param:insert_arg_list(NewContext, TermList, AS2),
    Term = tag:result(NewContext),
    {Term, AS3}.

generate_pid(Spec, Env, Context, AS) ->
    {term:make_pid(Context), AS}.
