%%% File    : spec_match.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Description : Match specification
%%% Created : 21 Aug 2002 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(spec_match).

-export([match/5, match_list/5, match_list/6, deref/3]).

match(Spec, Term, Env, Context, AS) ->
    case spec:spec_type(Spec) of
	functor ->
	    match_functor(Spec, Term, Env, Context, AS);
	'fun' ->
	    match_fun(Spec, Term, Env, Context, AS);
	pid ->
	    match_pid(Spec, Term, Env, Context, AS);
	union ->
	    match_union(Spec, Term, Env, Context, AS);
	universal ->
	    match_universal(Spec, Term, Env, Context, AS);
	var ->
	    match_var(Spec, Term, Env, Context, AS);
	alias ->
	    match_alias(Spec, Term, Env, Context, AS);
	apply ->
	    match_apply(Spec, Term, Env, Context, AS);
	external ->
	    match_external(Spec, Term, Env, Context, AS)
    end.


match_fun(Spec, Term, Env, Context, AS) ->
    case term:term_tag(Term) of 
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    match_fun(Spec, Term1, Env, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    match_join_list(Spec, TermList, Env, Context, AS);
	'fun' ->
	    SpecArgList = spec:fun_arg_list(Spec),
	    SpecResult = spec:fun_result(Spec) ,
	    AP = term:fun_ap(Term),
	    Arity =  term:fun_arity(Term),
	    Pos = term:fun_pos(Term),
	    TermArgList = param:invent_fun_arg_list(Arity, AP, Pos),
	    TermResult = tag:fun_result(AP, Pos),
	    case Arity == length(SpecArgList) of
		true ->
		    {GenList, AS1} = 
			spec_generate:generate_list(SpecArgList, Env, Context, AS),
		    AS2 = insert_list(TermArgList, GenList, AS1),
		    match(SpecResult, TermResult, Env, Context, AS2);
		false ->
		    {fail, AS}
	    end;
	_ ->
	 {fail, AS}
    end.
   
insert_list([IP | IPList], [Term | TermList], AS) ->
    AS1 = as:insert(IP, Term, AS),
    insert_list(IPList, TermList, AS1);
insert_list([], [], AS) ->
    AS.

match_pid(Spec, Term, Env, Context, AS) ->
   case term:term_tag(Term) of 
       ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    match_pid(Spec, Term1, Env, Context, AS1);
       universal ->
	   {fail, AS};
       join ->
	    TermList = term:join_arg_list(Term),
	    match_join_list(Spec, TermList, Env, Context, AS);
       functor ->
	   {fail, AS};
       pid ->
	   {ok, AS};
       _ ->
	   {fail, AS}
    end.

match_functor(Spec, Term, Env, Context, AS) ->
    case term:term_tag(Term) of 
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    match_functor(Spec, Term1, Env, Context, AS1);
	universal ->
	    PatternList = spec:spec_functor_arg_list(Spec),
	    TermList = lists:map(fun(_) -> term:universal() end, PatternList),
	    case match_list(PatternList, TermList, Env, Context, AS) of
		{ok, AS1} ->
		    {fail, AS1};
		{fail, AS1} ->
		    {fail, AS1}
	    end;
	join ->
	    TermList = term:join_arg_list(Term),
	    match_join_list(Spec, TermList, Env, Context, AS);
	functor ->
	    PC = spec:spec_functor(Spec),
	    TC = term:term_functor(Term),
	    case PC == TC of
		true ->
		    PatternList = spec:spec_functor_arg_list(Spec),
		    KeyList = term:term_functor_arg_list(Term),
		    match_list(PatternList, KeyList, Env, Context, AS);
		false ->
		    {fail, AS}
	    end;
	_ ->
	    {fail, AS}
    end.

match_join_list(Spec, [Term | TermList], Env, Context, AS) ->
    case match(Spec, Term, Env, Context, AS) of
	{ok, AS1} ->
	    match_join_list(Spec, TermList, Env, Context, AS1);
	{fail, AS1} ->
	    {_, AS2} = match_join_list(Spec, TermList, Env, Context, AS1),
	    {fail, AS2}
    end;
match_join_list(_, [], Env, _, AS) ->
    {ok, AS}.

match_list([Spec |SpecList], [Term | TermList], Env, Context, AS) ->
    case match(Spec, Term, Env, Context, AS) of
	{ok, AS1} ->
	    match_list(SpecList, TermList, Env, Context, AS1);
	{fail, AS1} ->
	    {_, AS2} = match_list(SpecList, TermList, Env, Context, AS1),
	    {fail, AS2}
    end;
match_list([], [], Env, _, AS) ->
    {ok, AS}.

match_list([Spec |SpecList], [Term | TermList], N, Env, Context, AS) ->
    case match(Spec, Term, Env, Context, AS) of
	{ok, AS1} ->
	    match_list(SpecList, TermList, N+1, Env, Context, AS1);
	{fail, AS1} ->
	    case match_list(SpecList, TermList, N+1, Env, Context, AS1) of
		{_, AS2} ->
		    {fail, N, AS2};
		{_, _, AS2} ->
		    {fail, N, AS2}
	    end
    end;
match_list([], [], _, Env, _, AS) ->
    {ok, AS}.

%%% Union

match_union(Spec, Term, Env, Context, AS) ->
    SpecList = spec:union_arg_list(Spec),
    match_union_list(SpecList, Term, Env, Context, AS).

% Match the union of a list of specs against a term.

match_union_list([Spec | SpecList], Term, Env, Context, AS) ->
    {Term1, AS1} =  spec_split:extract(Spec, Term, Env, Context, AS),
    case match(Spec, Term1, Env, Context, AS1) of
	{ok, AS2} ->
	    {TermRest, AS3} = spec_split:remove(Spec, Term, Env, Context, AS2),
	    match_union_list(SpecList, TermRest, Env, Context, AS3);
	{fail, AS2} ->
	    {TermRest, AS3} = spec_split:remove(Spec, Term, Env, Context, AS2),	
	    {_, AS4} = match_union_list(SpecList, TermRest, Env, Context, AS3),
	    {fail, AS4}
    end;
match_union_list([], Term, Env, Context, AS) ->
    {Term1, AS1} = deref(Term, Context, AS),
    case term:is_empty(Term1) of
	true ->
	    {ok, AS1};
	false  ->
	    {fail, AS1}
    end.

match_universal(Spec, Term, Env, Context, AS) ->
    {ok, AS}.

match_var(Spec, Term, Env, Context, AS) ->
    Name = spec:var_name(Spec),
    Box = env:lookup(Name, Env),
    case term:term_tag(Box) of 
	ip ->
	    AS1 = as:insert(Box, Term, AS),
	    {ok, AS1};
	parameter ->
	    {Term1, AS1} = deref(Term, Context, AS),
	    case term:included(Term1, Box) of
		true ->
		    {ok, AS1};
		false ->
		    {fail, AS1}
	    end
    end.
   
deref(Term, Context, AS) ->
    case term:term_tag(Term) of 
	ip ->
	    as:lookup(Term, Context, AS);
	_ ->
	    {Term, AS}
    end.

match_alias(Spec, Term, Env, Context, AS) ->
    Name = spec:alias_var_name(Spec),
    Box = env:lookup(Name, Env),
    case term:term_tag(Box) of 
	ip ->
	    AS1 = as:insert(Box, Term, AS)
    end,
    Spec1 = spec:alias_spec(Spec),
    match(Spec1, Term, Env, Context, AS1).


match_apply(Spec, Term, Env, Context, AS) ->
    case term:term_tag(Term) of 
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    match_apply(Spec, Term1, Env, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    match_join_list(Spec, TermList, Env, Context, AS);
	_ ->
	    Module = context:current_module(Context),
	    Name = spec:apply_name(Spec),
	    ArgList = spec:apply_arg_list(Spec),
	    Arity = length(ArgList),
	    Tyid = program:make_tyid(Module,Name, Arity), 
	    Form = program:typedef(Tyid),
	    case (spec:spec_form_type(Form) == abstypedef) 
		and (program:spec_mode(Module) == extern) of
		true ->
		    match_formal(Tyid, ArgList, Term, Env, Context, AS);
		false ->
		    Pos = spec:pos(Spec),
		    match_local(Tyid, ArgList, Term, Pos, Env, Context, AS)
	    end
    end.


match_formal(Tyid, ArgList, Term, Env, Context, AS) -> 
    Arity = program:tyid_arity(Tyid),
    case term:is_formal(Term) of
	true ->
	    case (program:tyid_module(Tyid) == term:formal_module(Term))
		and (program:tyid_name(Tyid) == term:formal_name(Term)) 
		and (Arity == term:formal_arity(Term)) of
		true -> 
		    TermList = term:formal_arg_list(Term),
		    match_list(ArgList, TermList, Env, Context, AS);
		false ->
		    {fail, AS}
	    end;
	false ->
	    case term:is_universal(Term) of
		true ->
		    TermList = lists:map(fun(_) -> term:universal() end, ArgList),
		    case match_list(ArgList, TermList, Env, Context, AS) of
			{ok, AS1} ->
			    {fail, AS1};
			{fail, AS1} ->
			    {fail, AS1}
		    end;
		false ->
		    {fail, AS}
	    end
    end.

	

match_local(Tyid, ArgList, Term, Pos, Env, Context, AS) -> 
    Arity = program:tyid_arity(Tyid),
    {NewContext, AS0} = work:call_matching(Tyid, Pos, Context, AS),
    TermList = create_arg_list(Arity, NewContext),
    case match_list(ArgList, TermList, Env, Context, AS0) of
	{ok,AS1} ->
	    CurrentWorkContext = Context,
						%work:work_context(work:current_work()),
 	    AS4 = as:insert(tag:result(NewContext), Term, AS1),
% 	    AS4 = as:insert(tag:match_result(CurrentWorkContext),
% 			    tag:match_result(NewContext),
% 			    AS3),
%	    AS4 = AS1,
	    case as:lookup(tag:match_result(NewContext), Context, AS4) of
		{MatchRes, AS5} ->
		    case term:is_empty(MatchRes) of
			true ->
			    {ok, AS5};
			false ->
			    {fail, AS5}
		    end
	    end;
	{fail, AS2}  ->
	    {fail, AS2}
    end.
    

match_external(Spec, Term, Env, Context, AS) ->
    case term:term_tag(Term) of 
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    match_external(Spec, Term1, Env, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    match_join_list(Spec, TermList, Env, Context, AS);
	_ ->
	    Module = spec:external_module(Spec),
	    Name = spec:external_name(Spec),
   	    ArgList = spec:external_arg_list(Spec),
	    Arity = length(ArgList),
	    Tyid = program:make_tyid(Module,Name, Arity), 
	    Form = program:typedef(Tyid),
	    case {program:spec_mode(Module),spec:spec_form_type(Form)} of
		{extern, abstypedef} ->
		    match_formal(Tyid, ArgList, Term, Env, Context, AS);
		{local,_} ->
		    Pos = 0,
		    match_local(Tyid, ArgList, Term, Pos, Env, Context, AS);
		{_,public_typedef} ->
		    Pos = 0,
		    match_local(Tyid, ArgList, Term, Pos, Env, Context, AS)
	    end
    end.
    
create_arg_list(Arity, Context) ->
    create_arg_list(1, Arity, Context).

create_arg_list(N, Arity, Context) when N =< Arity ->
    [tag:arg(N, Context) | create_arg_list(N+1, Arity, Context)];
create_arg_list(N, Arity, Context) when N> Arity ->
    [].


