%%% File    : spec_split.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Description : Splitting terms into those that match specification
%%%               and those that don't.
%%% Created : 21 Aug 2002 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(spec_split).
-export([empty_arg_list/3]).
-export([extract_arg_list/5, extract/5]).
-export([remove/5]).
-export([check_head_cover/5]).

%%% Remove a type from a term.

%% Extract a subterm which matches the type

extract_arg_list(SpecList, TermList, Env, Context, AS) ->
    {SkeletList, AS1} = skelet_list(SpecList, Env, Context, AS),
    extract_list_list(SkeletList, TermList, Context, AS1).    


-define(cover_depth, 3).
-define(remove_depth, 2).

%% Head cover


empty_arg_list([Term|TermList], Context, AS) ->
    {Term1, AS1} = spec_match:deref(Term, Context, AS),
    case term:is_empty(Term1) of
	true ->
	    {true, AS1};
	false ->
	    empty_arg_list(TermList, Context, AS1)
    end;
empty_arg_list([], _, AS) ->
    {false, AS}.

check_head_cover([Clause | ClauseList], TermList, Env, Context, AS) ->
    ArgList = spec:clause_arg_list(Clause),
    {SkeletList, AS1} = skelet_list(ArgList, Env, Context, AS),
    {TermList1, AS2} = remove_list_list(SkeletList, TermList, 2, Context, AS1),
    check_head_cover(ClauseList, TermList1, Env, Context, AS2);
check_head_cover([], TermList, Env, Context, AS) ->
    Flag = lists:all(fun (Term) -> term:is_empty(Term) end, TermList),
    {Flag, AS}.

skelet(Spec, Env, Context, AS) ->
    {Term, AS1} = skelet(Spec, history_new(), Env, Context, AS),
    {Term, AS1}.

skelet(Spec, History, Env, Context, AS) ->
    case spec:spec_type(Spec) of 
	universal ->
	    {term:universal(), AS};
	var ->
	    skelet_var(Spec, History, Env, Context, AS);
	union ->
	    skelet_union(Spec, History, Env, Context, AS);
	apply ->
	    skelet_apply(Spec, History, Env, Context, AS);
	external ->
	    skelet_external(Spec, History, Env, Context, AS);
	functor ->
	    skelet_functor(Spec, History, Env, Context, AS)
    end.

skelet_functor(Spec, History, Env, Context, AS) ->
    PC = spec:spec_functor(Spec),
    PatternList = spec:spec_functor_arg_list(Spec),
    {TermList, AS1} = skelet_list(PatternList, history_functor(History), Env, Context, AS),
    {term:apply_functor(PC, TermList), AS1}.

skelet_union(Spec, History, Env, Context, AS) ->
    SpecList = spec:union_arg_list(Spec),
    {TermList, AS1} = skelet_list(SpecList, History, Env, Context, AS),
    {term:join(TermList), AS1}.

skelet_apply(Spec, History, Env, Context, AS) ->
    Module = context:current_module(Context),
    Name = spec:apply_name(Spec),
    ArgList = spec:apply_arg_list(Spec),
    Arity = length(ArgList),
    Tyid = program:make_tyid(Module,Name, Arity), 
    Form = program:typedef(Tyid),
    case (spec:spec_form_type(Form) == abstypedef) 
	and (program:spec_mode(Module) == extern) of
	true ->
	    {TermList, AS1} = skelet_list(ArgList, History, Env, Context, AS),
	    {term:make_formal(Module, Name, Arity, TermList), AS1};
	false ->
	    skelet_call({Tyid, ArgList}, History, Env, Context, AS)
    end.
							
skelet_call({Tyid, ArgList}, History, Env, Context, AS) ->
    case history_call(Tyid, History) of 
	stop ->
	    {term:universal(), AS};
	{next, History1} ->
	    {TermList, AS1} = skelet_list(ArgList, History, Env, Context, AS),
	    Form = program:typedef(Tyid),
	    Body = spec:typedef_body(Form),
	    ParamList = lists:map(fun(Var) -> spec:var_name(Var) end,
				  spec:typedef_arg_list(Form)),
	    Env1 = env:bind_list(ParamList, TermList, env:empty()),
	    skelet(Body, History1, Env1, Context, AS1)
    end.

skelet_external(Spec, History, Env, Context, AS) ->
    Module = spec:external_module(Spec),
    Name = spec:external_name(Spec),
    ArgList = spec:external_arg_list(Spec),
    Arity = length(ArgList),
    Tyid = program:make_tyid(Module,Name, Arity), 
    Form = program:typedef(Tyid),
    case {program:spec_mode(Module),spec:spec_form_type(Form)} of
	{extern, abstypedef} ->
	    {TermList, AS1} = skelet_list(ArgList, History, Env, Context, AS),
	    {term:make_formal(Module, Name, Arity, TermList), AS1};
	{local,_} ->
	    skelet_call({Tyid, ArgList}, History, Env, Context, AS);
	{_,public_typedef} ->
	    skelet_call({Tyid, ArgList}, History, Env, Context, AS)
    end.

skelet_var(Spec, History, Env, Context, AS) ->
    Name = spec:var_name(Spec),
    Value = env:lookup(Name, Env),
    {Value, AS}.

skelet_list(SpecList, Env, Context, AS) ->
    {SkeletList, AS1} = skelet_list(SpecList, history_new(), Env, Context, AS),
    {SkeletList, AS1}.


skelet_list([Spec | SpecList], History, Env, Context, AS) ->
    {Term, AS1} = skelet(Spec, History, Env, Context, AS),
    {TermList, AS2} = skelet_list(SpecList, History, Env, Context, AS1),
    {[Term | TermList], AS2};
skelet_list([], _, _, _, AS) ->
    {[], AS}.
    


%%% Remove

remove(Spec, Term, Env, Context, AS) ->
    {Skelet, AS1} = skelet(Spec, Env, Context, AS),
    remove(Skelet, Term, Context, AS1).

remove(Skelet, Term, Context, AS)  ->
    remove1(Skelet, Term, 2, Context, AS).

remove1(Skelet, Term, N, Context, AS) when N > 0 ->
    case term:term_tag(Skelet) of 
	ip ->
	    {term:join0(), AS};
	join ->
	    SkeletList = term:join_arg_list(Skelet),
	    remove_join_list(SkeletList, Term, N, Context, AS);
	universal ->
	    {term:join0(), AS};
	parameter ->
	    remove_parameter(Skelet, Term, N, Context, AS);
	functor ->
	    remove_functor(Skelet, Term, N, Context, AS)
    end;
remove1(Skelet, Term, N, Context, AS) when N =< 0 ->
    {Term, AS}.
	    
remove_functor(Skelet, Term, N, Context, AS) ->
    case term:term_tag(Term) of
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    remove_functor(Skelet, Term1, N-1, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    {TermList1, AS1} = remove_term_list(Skelet, TermList, N, Context, AS),
	    {term:join(TermList1), AS1};
	universal ->
	    {Term, AS};
	functor ->
	    case term:term_functor(Term) == term:term_functor(Skelet) of
		true ->
		    TC = term:term_functor(Term),
		    TermList = term:term_functor_arg_list(Term),
		    SkeletList = term:term_functor_arg_list(Skelet),
		    {RemoveList, AS1} = remove_list_list(SkeletList, TermList, N, Context, AS),
		    TermListList = combine_arg_lists(RemoveList, TermList),
		    ResultList = lists:map(fun(ArgList) ->
						 term:apply_functor(TC, ArgList)
					 end,
					 TermListList),
		    {term:join(ResultList), AS1};		
		false ->
		    {Term, AS}
	    end;
	_ ->
	   {Term, AS}
    end.

remove_parameter(Skelet, Term, N, Context, AS) ->
    case term:term_tag(Term) of 
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    remove_parameter(Skelet, Term1, N-1, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    {TermList1, AS1} = remove_term_list(Skelet, TermList, N, Context, AS),
	    {term:join(TermList1), AS1};
	parameter ->
	    case Skelet == Term of
		true ->
		    {term:join0(), AS};
		false ->
		    {Term, AS}
	    end;
	_ ->
	    {Term, AS}
    end.


remove_join_list([Skelet|SkeletList], Term, N, Context, AS) ->
    {Term1, AS1} = remove1(Skelet, Term, N, Context, AS),
    remove_join_list(SkeletList, Term1, N, Context, AS1);
remove_join_list([], Term, N, Context, AS) ->
    {Term, AS}.

remove_term_list(Skelet, [Term|TermList], N, Context, AS) ->
    {Term1, AS1} = remove1(Skelet, Term, N, Context, AS),
    {TermList1, AS2} = remove_term_list(Skelet, TermList, N, Context, AS1),
    {[Term1|TermList1],AS2};
remove_term_list(Skelet, [], _, _, AS) ->
    {[], AS}.

remove_list_list([Skelet|SkeletList], [Term|TermList], N, Context, AS) ->
    {Term1, AS1} = remove1(Skelet, Term, N, Context, AS),
    {TermList1, AS2} = remove_list_list(SkeletList, TermList, N, Context, AS1),
    {[Term1|TermList1],AS2};
remove_list_list([], [], _, _, AS) ->
    {[], AS}.

combine_arg_lists([Remove | RemoveList], [Term | TermList]) ->
    TermListList = combine_arg_lists(RemoveList, TermList),
    TermListList1 = 
	lists:map(fun(List) ->
			  [Term | List]
		  end,
		  TermListList),
    [[Remove | TermList] | TermListList1];
combine_arg_lists([], []) ->
    [].
 
% remove_arg_list(SpecList, TermList, Env, Context, AS) ->
%     {SkeletList, AS1} = skelet_list(SpecList, Env, Context, AS),
%     {TermList1, AS2} = remove_list_list(SpecList, TermList, Context, AS1),
%     TermListList = combine_arg_lists(TermList1, TermList),
%     {TermListList, AS2}.

%%% extract

extract(Spec, Term, Env, Context, AS) ->
    {Skelet, AS1} = skelet(Spec, Env, Context, AS),
    extract(Skelet, Term, Context, AS1).

extract(Skelet, Term, Context, AS) ->
    case term:term_tag(Skelet) of 
	ip ->
	    {Term, AS};
	join ->
	    SkeletList = term:join_arg_list(Skelet),
	    {TermList, AS1} = extract_join_list(SkeletList, Term, Context, AS),
	    {term:join(TermList), AS1};
	functor ->
	    extract_functor(Skelet, Term, Context, AS);
	universal ->
	    {Term, AS};
	parameter ->
	    extract_parameter(Skelet, Term, Context, AS)
    end.
	    
extract_functor(Skelet, Term, Context, AS) ->
    case term:term_tag(Term) of
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    extract_functor(Skelet, Term1, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    {TermList1, AS1} = extract_term_list(Skelet, TermList, Context, AS),
	    {term:join(TermList1), AS1};
	universal ->
	    {term:universal(), AS};
	functor ->
	    case term:term_functor(Term) == term:term_functor(Skelet) of
		true ->
		    TC = term:term_functor(Term),
		    TermList = term:term_functor_arg_list(Term),
		    SkeletList = term:term_functor_arg_list(Skelet),
		    {ExtractList, AS1} = extract_list_list(SkeletList, TermList, Context, AS),
		    {term:apply_functor(TC, ExtractList), AS1};
		false ->
		    {term:join0(), AS}
	    end;
	_ ->
	   {term:join0(), AS}
    end.

extract_parameter(Skelet, Term, Context, AS) ->
    case term:term_tag(Term) of 
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    extract_parameter(Skelet, Term1, Context, AS1);
	join ->
	    TermList = term:join_arg_list(Term),
	    {TermList1, AS1} = extract_term_list(Skelet, TermList, Context, AS),
	    {term:join(TermList1), AS1};
	universal ->
	    {term:universal(), AS};
	parameter ->
	    case Skelet == Term of
		true ->
		    {Term, AS};
		false ->
		    {term:join0(), AS}
	    end;
	_ ->
	    {term:join0(), AS}
    end.


extract_join_list([Skelet|SkeletList], Term, Context, AS) ->
    {Term1, AS1} = extract(Skelet, Term, Context, AS),
    {TermList1, AS2} = extract_join_list(SkeletList, Term1, Context, AS1),
    {[Term1|TermList1], AS2};
extract_join_list([], Term, Context, AS) ->
    {[], AS}.

extract_term_list(Skelet, [Term|TermList], Context, AS) ->
    {Term1, AS1} = extract(Skelet, Term, Context, AS),
    {TermList1, AS2} = extract_term_list(Skelet, TermList, Context, AS1),
    {[Term1|TermList1],AS2};
extract_term_list(Skelet, [], _, AS) ->
    {[], AS}.

extract_list_list([Skelet|SkeletList], [Term|TermList], Context, AS) ->
    {Term1, AS1} = extract(Skelet, Term, Context, AS),
    {TermList1, AS2} = extract_list_list(SkeletList, TermList, Context, AS1),
    {[Term1|TermList1],AS2};
extract_list_list([], [], _, AS) ->
    {[], AS}.


%%% History

history_new() ->
    {initial, gb_trees:empty()}.

history_call(Tyid, {initial, History}) ->
    case gb_trees:lookup(Tyid, History) of
	none ->
	    {next, {initial, gb_trees:insert(Tyid, 0, History)}};
	{values, _} ->
	    stop
    end;
history_call(_, finished) ->
    stop.

history_functor(_) ->
    finished.

%%% Cover. Does skeleton cover term?    

% cover(Skelet, Term, Context, AS) ->
%     case term:term_tag(Skelet) of 
% 	universal ->
% 	    {true, AS};
% 	join ->
% 	    SkeletList = term:join_arg_list(Skelet),
% 	    cover_join_list(SkeletList, Term, Context, AS);
% 	functor ->
% 	    cover_functor(Skelet, Term, Context, AS)

%     end.
	    
% cover_functor(Skelet, Term, Context, AS) ->
%     case term:term_tag(Term) of
% 	universal ->
% 	    {false, AS};
% 	ip ->
% 	    {Term1, AS1} = as:lookup(Term, Context, AS),
% 	    cover_functor(Skelet, Term1, Context, AS1);
% 	join ->
% 	    TermList = term:join_arg_list(Term),
% 	    cover_term_list(Skelet, TermList, Context, AS);
% 	functor ->
% 	    case term:term_functor(Term) == term:term_functor(Skelet) of
% 		true ->
% 		    TermList = term:term_functor_arg_list(Term),
% 		    SkeletList = term:term_functor_arg_list(Skelet),
% 		    {Flag, AS1} = cover_list_list(SkeletList, TermList, Context, AS),
% 		    {Flag, AS1};
% 		false ->
% 		    {false, AS}
% 	    end
%     end.

% cover_join_list([Skelet|SkeletList], Term, Context, AS) ->
%     case term:is_empty(Term) of
% 	true ->
% 	    {true, AS};
% 	false ->
% 	    {Term1, AS1} = remove(Skelet, Term, Context, AS),
% 	    cover_join_list(SkeletList, Term1, Context, AS1)
%     end;
% cover_join_list([], Term, Context, AS) ->
%     {term:is_empty(Term), AS}.

% cover_term_list(Skelet, [Term|TermList], Context, AS) ->
%    case cover(Skelet, Term, Context, AS) of
% 	{false, AS1} ->
% 	    {false, AS1};
% 	{true, AS1} ->
% 	    cover_term_list(Skelet, TermList, Context, AS1)
%     end;
% cover_term_list(Skelet, [], _, AS) ->
%     {true, AS}.

% cover_list_list([Skelet|SkeletList], [Term|TermList], Context, AS) ->
%     case cover(Skelet, Term, Context, AS) of
% 	{false, AS1} ->
% 	    {false, AS1};
% 	{true, AS1} ->
% 	    cover_list_list(SkeletList, TermList, Context, AS1)
%     end;
% cover_list_list([], [], _, AS) ->
%     {true, AS}.
