%%% File    : spec_walker.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@csd.uu.se>
%%% Purpose : Analysis of specifications, i.e., type inference
%%% Created : 19 Jan 2000 by Sven-Olof Nystr|m <svenolof@csd.uu.se>

-module(spec_walker).
-author('svenolof@harpo.it.uu.se').

-export([specification/5, 
	 generate_typedef/3, match_typedef/3]).


%%% Evaluate a specification from the analyzer.

specification(Fnid, Spec, Env, Context, AS) ->
    case spec:spec_type(Spec) of
	funspec ->
	    specification_one(Fnid, Spec, Env, Context, AS);
	mfunspec ->
	    specification_many(Fnid, Spec, Env, Context, AS);
	special ->
	    specification_special(Fnid, Context, AS)
    end.

specification_one(Fnid, Spec, Env, Context, AS) ->
    Env1 = bind_universals(Spec, Env, Context),
    SpecArgList = spec:funspec_arg_list(Spec),
    SpecResult = spec:funspec_result(Spec) ,
    Arity = length(SpecArgList),
    Pos = spec:funspec_line(Spec),
    TermArgList = param:invent_arg_list(Arity, Context),
    TermResult = tag:result(Context),
    Caller = context:caller(Context),
    case spec_match:match_list(SpecArgList, TermArgList, 1, Env1, Context, AS) of
	{ok, AS1} ->
	    {Term, AS2} = 
		spec_generate:generate(SpecResult, Env1, Context, AS1),
	    AS3 = as:insert(TermResult, Term, AS2),
	    AS3;
	{fail, N, AS1} ->
	    AS11 = report:report_bad_arg(Caller, Fnid, N, AS1),
	    {Term, AS2} = spec_generate:generate(SpecResult, Env1, Context, AS11),
	    AS3 =  as:insert(TermResult, Term, AS2)
    end,
    Cont = fun(_, BS) ->
		   report:report_bad_arg(Caller, Fnid, BS)
	   end,
    as:lookup_cont(tag:match_result(Context), Cont, Context, AS3).

specification_many(Fnid, Spec, Env, Context, AS) ->
     Env1 = bind_universals(Spec, Env, Context),
    ClauseList = spec:mfunspec_clause_list(Spec),
    Arity = program:fnid_arity(Fnid),
    TermArgList = param:invent_arg_list(Arity, Context),
    Caller = context:caller(Context),
    case spec_split:check_head_cover(ClauseList, TermArgList, Env1, Context, AS) of
	{true, AS1} ->
	    try_clause_list(ClauseList, TermArgList, Fnid, Env1, Context, AS1);
	{false, AS1} ->
	    AS2 = report:report_bad_arg(Caller, Fnid, AS1),
	    try_clause_list(ClauseList, TermArgList, Fnid, Env1, Context, AS2)
    end.

try_clause_list([Clause|ClauseList], TermList, Fnid, Env, Context, AS) ->
    Env1 = bind_universals(Clause, Env, Context),
    SpecList = spec:clause_arg_list(Clause),
    SpecResult = spec:clause_result(Clause),
    {TermList1, AS1} 
	= spec_split:extract_arg_list(SpecList, TermList, Env, Context, AS),
    case spec_split:empty_arg_list(TermList1, Context, AS1) of
	{false, AS11} ->
	    case spec_match:match_list(SpecList, TermList1, 1, Env1, Context, AS11) of
		{ok, AS2} ->
		    AS3 = AS2;
		{fail, N, AS2} ->
		    Caller = context:caller(Context),
		    AS3 = report:report_bad_arg(Caller, Fnid, N, AS2)
	    end,
	    {Term, AS3b} 
		= spec_generate:generate(SpecResult, Env1, Context, AS3),
	    AS4 = as:insert(tag:result(Context), Term, AS3b),
	    try_clause_list(ClauseList, TermList, Fnid, Env, Context, AS4);
	{true, AS11} ->
	    try_clause_list(ClauseList, TermList, Fnid, Env, Context, AS11)
    end;
try_clause_list([], _, Fnid, Env, Context, AS) ->
    AS.
    

specification_special(Fnid, Context, AS) ->
    special_definition:call(Fnid, Context, AS).

%%%    

bind_universals(Spec, Env, Context) ->
    case spec:spec_type(Spec) of
	funspec ->
	    ArgList = spec:funspec_arg_list(Spec),
	    Result = spec:funspec_result(Spec),
	    bind_universal_list([Result | ArgList], Env, Context);
	mfunspec ->
	    ClauseList = spec:mfunspec_clause_list(Spec),
	    bind_universal_clause_list(ClauseList, Env, Context);
	functor ->
	    bind_universal_list(spec:spec_functor_arg_list(Spec), Env, Context);
	'fun' ->
	    ArgList = spec:fun_arg_list(Spec),
	    Result = spec:fun_result(Spec),
	    bind_universal_list([Result | ArgList], Env, Context);
	pid ->
	    ArgList = spec:pid_arg_list(Spec),
	    bind_universal_list(ArgList, Env, Context);
	union ->
	    List = spec:union_arg_list(Spec),
	    bind_universal_list(List, Env, Context);
	universal ->
	    Env;
	var ->
	    Name = spec:var_name(Spec),
	    case env:defined(Name, Env) of
		true ->
		    Env;
		false ->
		    env:bind(Name, tag:var(Name, 0, Context), Env)
	    end;
	alias ->
	    Name = spec:alias_var_name(Spec),
	    case env:defined(Name, Env) of
		true ->
		    Env1 = Env;
		false ->
		    Env1 = env:bind(Name, tag:var(Name, 0, Context), Env)
	    end,
	    Spec1 = spec:alias_spec(Spec),
	    bind_universals(Spec1, Env1, Context);
	apply ->
	    ArgList = spec:apply_arg_list(Spec),
	    bind_universal_list(ArgList, Env, Context);
	external ->
	    ArgList = spec:external_arg_list(Spec),
	    bind_universal_list(ArgList, Env, Context);
	clause ->
	    ArgList = spec:clause_arg_list(Spec),
	    Result = spec:clause_result(Spec),
	    bind_universal_list([Result|ArgList], Env, Context)
    end.

bind_universal_clause_list([Clause|ClauseList], Env, Context) ->
    ArgList = spec:clause_arg_list(Clause),
    Result = spec:clause_result(Clause),
    Env1 = bind_universal_list([Result | ArgList], Env, Context),
    bind_universal_clause_list(ClauseList, Env1, Context);
bind_universal_clause_list([], Env, _) ->
    Env.
  

bind_universal_list([Spec | SpecList], Env, Context) ->
    Env1 = bind_universals(Spec, Env, Context),
    bind_universal_list(SpecList, Env1, Context);
bind_universal_list([], Env, _) ->
    Env.


%%% Generate

generate_typedef(Typedef, Context, AS) ->
    GetName = fun (Var) -> 
		      spec:var_name(Var)
	      end,
    ArgList = lists:map(GetName, spec:typedef_arg_list(Typedef)),
    Body = spec:typedef_body(Typedef),
    Env = param:initialize_environment(ArgList, Context),
    {Term, AS1} = spec_generate:generate(Body, Env, Context, AS),
    as:insert(tag:result(Context), Term, AS1).


%%% Match

match_typedef(Typedef, Context, AS) ->
    GetName = fun (Var) -> 
		      spec:var_name(Var)
	      end,
    ArgList = lists:map(GetName, spec:typedef_arg_list(Typedef)),
    Body = spec:typedef_body(Typedef),
    Env = param:initialize_environment(ArgList, Context),
    case spec_match:match(Body, tag:result(Context), Env, Context, AS) of
	{fail, AS1} ->
	    CurrentWorkContext = Context, 
	    as:insert(tag:match_result(CurrentWorkContext),
		      term:make_fail(),
		      AS1);
	{ok, AS1} ->
	    AS1
    end.
    
	    
