%%% File    : special_definition.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : Handle functions in 'erlang' module that cannot be
%%%           decribed using ordinary function type specifications.
%%% Created : 30 Oct 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(special_definition).
-author('svenolof@harpo.it.uu.se').

%-compile(export_all).
-export([call/3, send/4, rse/1, call/6, result/1]).

call(Fnid, Context, AS) ->
    case program:fnid_module(Fnid) of
	erlang ->
	    call_erlang(Fnid, Context, AS);
	'This is not the primop module' ->
	    call_primop(Fnid, Context, AS)
    end.

call_erlang(Fnid, Context, AS) ->
    case {program:fnid_name(Fnid), program:fnid_arity(Fnid)} of
	{apply, 3} ->
	    apply_3(Context, AS);
	{exit_current, 1} ->
	    exit_current_1(Context, AS);
	{exit, 1} ->
	    exit_current_1(Context, AS);
	{fault,1} ->
	    fault_1(Context, AS);
	{fault,2} ->
	    fault_1(Context, AS);
	{throw, 1} ->
	    throw_1(Context, AS);
	{self, 0} ->
	    self_0(Context, AS);
	{spawn, 1} ->
	    spawn_1(Context, AS);
	{spawn, 3} ->
	    spawn_3(Context, AS);
	{spawn_opt, 4} ->
	    spawn_3(Context, AS);
	{spawn_link, 3} ->
	    spawn_3(Context, AS); % Should also make link.
	{'!',2} ->
	    '!_2'(Context, AS);
	{put, 2} ->
	    put_2(Context, AS);
	{get, 1} ->
	    get_1(Context, AS)
    end.

call_primop(Fnid, Context, AS) ->
    case {program:fnid_name(Fnid), program:fnid_arity(Fnid)} of
	{match_fail, 1} ->
	    exit_current_1(Context, AS)
    end.

%% Special definitions of 'Erlang' module 


apply_3(Context, AS) ->
    [Module, Name, ArgList] = param:invent_arg_list(3, Context),
    call(Module, Name, ArgList, rse(Context), Context, AS).

exit_current_1(Context, AS) ->
    [Arg1] = param:invent_arg_list(1, Context),
    exception('EXIT',Arg1, Context, AS).

fault_1(Context, AS) ->
    [Arg1] = param:invent_arg_list(1, Context),
    M = term:make_tuple([Arg1, term:universal()]),
    exception(M, Context, AS).

throw_1(Context, AS) ->
    [Arg1] = param:invent_arg_list(1, Context),
    exception('THROW',Arg1, Context, AS).
    
self_0(Context, AS) ->
    return(self(Context), Context, AS).

spawn_1(Context, AS) ->
    [Fun0] = param:invent_arg_list(1, Context),
    Result = tag:spawn_result(0, Context),
    Env = param:initialize_environment([], Context),
    Self = new_process(Context),
    Env1 = env:bind('$self', Self, Env),
    F = fun([Fun], BS)->
		{_, BS1} = walker:apply_fun(Fun, [], Env1, Context, BS),
		BS1
	end,
    AS1 = wait:wait([Fun0], [universal],F, Context, AS),
    return(Self, Context, AS1).


spawn_3(Context, AS) ->
    [Module, Name, ArgList] = param:invent_arg_list(3, Context),
    Result = tag:spawn_result(0, Context),
    Self = new_process(Context),
    Except = except(Context),
    RSE = {Result, Self, Except},
    AS1 = call(Module, Name, ArgList, RSE, Context, AS),
    return(Self, Context, AS1).

'!_2'(Context, AS) ->
    [Dest, Mess] = param:invent_arg_list(2, Context),
    AS1 = return(Mess, Context, AS),
    send(Dest, Mess, Context, AS1).

send(Pid, Message, Context, AS) ->
     F = fun([Dest1], BS) ->
		 case term:is_universal(Dest1) of
		     true ->
			 BS;
		     false ->
			 BS1 = as:insert(term:pid_mbox(Dest1), Message, BS),
			 BS1
		 end
	end,
    wait:wait([Pid], [pid], F, Context, AS).

%% Process dictionary

put_2(Context, AS) ->
    [Key, Val] = param:invent_arg_list(2, Context),
    Caller = context:caller(Context),
    Context1 = context:set_current_fnid(Caller, Context),
    Term = term:make_tuple([Key, Val]),
    case program:attribute(dict, Caller) of
	{value, DictClauseList} ->
	    F = fun({KeySpec, ValSpec}, {Acc, BS}) ->
			case spec_match:match(KeySpec, Key, env:empty(), Context1, BS) of
			    {ok, BS1} ->
				Acc1 = [ValSpec | Acc],
				{Acc1, BS1};
			    {fail, BS1} ->
				{Acc, BS1}
			end
		end,
	    case lists:foldl(F, {[], AS}, DictClauseList) of
		{[], AS1} ->
		    AS2 = report:report_bad_dict_entry(Caller, Term, AS1);
		{ValSpecList, AS1} ->
		    G = fun(ValSpec) ->
				case spec_match:match(ValSpec, Val, env:empty(), Context1, AS1) of
				    {ok, _} -> true;
				    _ -> false
				end
			end,
		    case lists:all(G, ValSpecList) of
			true ->
			    AS2 = AS1;
			false ->
			    AS2 = report:report_bad_dict_entry(Caller, Term, AS1)
		    end
	    end,
	    ValSpec = spec:make_union_spec(lists:map(fun({_, Spec}) -> Spec end, DictClauseList)),
	    {Res, AS3} = spec_generate:generate(ValSpec, env:empty(), Context1, AS2),
	    return(Res, Context, AS3);
	none ->
	    return(term:universal(), Context, AS)
    end.
    
get_1(Context, AS) ->
    [Key] = param:invent_arg_list(1, Context),
    Caller = context:caller(Context),
    Context1 = context:set_current_fnid(Caller, Context),
    case program:attribute(dict, Caller) of
	{value, DictClauseList} ->
%	    io:format("~nFound dcl: ~w",[DictClauseList]),
	    F = fun({KeySpec, ValSpec}, {Acc, BS}) ->
			case spec_match:match(KeySpec, Key, env:empty(), Context1, BS) of
			    {ok, BS1} ->
				Acc1 = [ValSpec | Acc],
				{Acc1, BS1};
			    {fail, BS1} ->
				{Acc, BS1}
			end
		end,
	    case lists:foldl(F, {[], AS}, DictClauseList) of
		{[], AS1} ->
		    AS2 = report:report_bad_dict_key(Caller, Key, AS1);
		{ValSpecList, AS1} ->
%		    io:format("~nFound vsl: ~w",[ValSpecList]),
		    ValSpec = spec:make_union_spec(ValSpecList),
		    {Res, AS3} = spec_generate:generate(ValSpec, env:empty(), Context1, AS1),
%		    io:format("~n generated ~w",[Res]),
		    return(Res, Context, AS3)
	    end;
	none ->
	    return(term:universal(), Context, AS)
    end.

    
%% Useful help functions

result(Context) ->
    tag:result(Context).

self(Context) ->
    tag:self(Context).

except(Context) ->
     tag:except(Context).

exception(Message, Context, AS) -> 
    Except = tag:except(Context),
    AS1 = as:insert(Except, Message, AS),
    Caller = context:caller(Context),
    report:report_exception(Caller, Message, AS1).

exception(Tag, Message, Context, AS) -> 
    Except = tag:except(Context),
    Term1 = term:make_tuple([term:make_atom(Tag), Message]),
    AS1 = as:insert(Except, Term1, AS),
    Caller = context:caller(Context),
    report:report_exception(Caller, Term1, AS1).

return(Term, Context, AS) ->
    as:insert(tag:result(Context), Term, AS).

rse(Context) ->
    {tag:result(Context), tag:self(Context), tag:except(Context)}.

new_process(Context) ->
    term:make_pid(Context).

call(Module, Name, ArgList, RSE, Context, AS) ->
    F = fun([Module1, Name1], BS) ->
		case term:is_universal(Module1) or term:is_universal(Name1) of	
		    false ->
			ModuleAtom = term:atom_val(Module1),
			NameAtom = term:atom_val(Name1),
			ArityList = program:arity_list(ModuleAtom, NameAtom),
			G = fun(Arity, CS) ->
				    Fnid = program:make_fnid(ModuleAtom, NameAtom, Arity),
				    H = fun([ArgList1], DS) ->
						call:call_fnid(Fnid, ArgList1, RSE, Context, DS)
					end,
				    wait:wait([ArgList], [{list, Arity}], H, Context, CS)
			    end,
			lists:foldl(G, BS, ArityList);
		    true ->
			io:format("~n Uni",[]),
			{ResultIP, _, _} = RSE,
			BS1 = call:treat_escaped(ArgList, Context, BS),
			BS2 = as:insert(ResultIP, term:universal(), BS1),
			io:format("~n OK",[]),
			BS2
		end;
	   (X, BS) ->
		io:format("~nHej hopp:w",[X]),
		BS
	end,
    wait:wait([Module, Name], [atom, atom], F, Context, AS).
