%%% File    : store.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@csd.uu.se>
%%% Purpose : 
%%% Created : 15 Feb 1999 by Sven-Olof Nystr|m <svenolof@csd.uu.se>

-module(store).
-author('svenolof@csd.uu.se').

-export([empty_store/0, lookup_ip_term/2, insert/3, depend/3,
	 make_context_table/2]).


empty_store() ->
    gb_trees:empty().

insert(IP_val, Term, Store) ->
    case term:term_tag(Term) of
	ip ->
	    {Store1, WL} =
		insert(IP_val, lookup_ip_term(Term, Store), Store),
	    {Store2, WL2} = 
		install_link(Term, IP_val, Store1),
	    {Store2, WL ++ WL2};
	join ->
	    TermList = term:join_arg_list(Term),
	    insert_list(IP_val, TermList, Store);
	universal ->
	    insert_universal(IP_val, Term, Store);
	_  ->
	    insert_term(IP_val, Term, Store)
    end.

%% insert_term returns {NewStore, WorkList}

insert_term(IP_val, Term, Store) ->
    case  lookup_ip(IP_val, Store) of
	not_found -> 
	    Size = gb_trees:size(Store) + 1,
	    Data1 = update_ip_term(new_ip_data(), Term),
	    Store1 = gb_trees:insert(IP_val, Data1, Store),
	    {Store1, []};	
	{found, Data} ->
	    PresentTerm = ip_term(Data),
	    case term:included(Term, PresentTerm) of
		true ->
		    {Store, []};
		false  ->
		    Data1 = update_ip_term(Data, term:join2(PresentTerm, Term)),
		    Store1 = gb_trees:update(IP_val, Data1, Store),
		    WorkList = ip_dependent(Data),
		    ForwardList = ip_links(Data),
		    {Store2, WorkList2} = multi_insert(ForwardList, Term, [], Store1),
		    {Store2, WorkList++WorkList2}
	    end
    end.

multi_insert([IP_val | ForwardList], Term, WL, Store) ->
    {Store1, WorkList1} = insert_term(IP_val, Term, Store),
    multi_insert(ForwardList, Term, WorkList1 ++ WL, Store1);
multi_insert([], _, WL, Store) ->
    {Store, WL}.


	
inc(X) ->
    case get({count, X}) of
	undefined ->
	    put({count, X}, 1);
	N when N rem 1000 == 0 ->
	    io:format("~nCount ~w. Done ~w ",[X,N]),
	    put({count, X}, N+1);
	N ->
	    put({count, X}, N+1)
    end.



insert_universal(IP_val, Term, Store) ->
     case  lookup_ip(IP_val, Store) of
	 not_found -> 
	     insert_term(IP_val, Term, Store);
	 {found, Data} ->
	     PresentTerm = ip_term(Data),
	     case term:term_tag(PresentTerm) of
		 universal ->
		    {Store, []};
		 _  ->
		    Data1 = update_ip_term(Data, Term),
		    Store1 = gb_trees:update(IP_val, Data1, Store),
		    WorkList = ip_dependent(Data),
		    ForwardList = ip_links(Data),
		    {Store2, WorkList2} = multi_insert(ForwardList, Term, [], Store1),
		    {Store2, WorkList++WorkList2}
	     end
     end.

insert_list(IP, [], Store) ->
    {Store, []};
insert_list(IP, [Term|TermList], Store) ->
    {Store1, WL1} = insert(IP, Term, Store),
    {Store2, WL2} = insert_list(IP, TermList, Store1),
    {Store2, WL1 ++ WL2}.


install_link(A, A, Store) ->
    {Store, []};
install_link(A, B, Store) ->
    case  lookup_ip(A, Store) of
	not_found -> 
	    Data1 = update_ip_links(new_ip_data(), [B]),
	    Store1 = gb_trees:insert(A, Data1, Store),
	    {Store1, []};
	{found, Data} ->
	    LinkList = ip_links(Data),
	    case lists:member(B, LinkList) of
		true ->
		    {Store, []};
		false -> 
		    Term = ip_term(Data),
		    Data1 = update_ip_links(Data, [B | LinkList]),
		    Store1 = gb_trees:update(A, Data1, Store),
		    {Store2, WL} = insert(B, Term, Store1)
	    end
    end.




depend(IP, AP, Store) ->
    Data = lookup_ip_safe(IP, Store),
    DependentList = ip_dependent(Data),
    case lists:member(AP, DependentList) of
	true ->
	    Store;
	false ->
	    Data1 = update_ip_dependent(Data, [AP|DependentList]),
	    gb_trees:enter(IP, Data1, Store)
    end.

lookup_ip(IP, Store) ->	
    case gb_trees:lookup(IP, Store) of
	{value, X} ->
	    {found, X};
	none -> 
	    not_found
    end.

lookup_ip_safe(IP, Store) ->
    case gb_trees:lookup(IP, Store) of
	none -> 
	    new_ip_data();
	{value, Data} ->
	    Data
    end.
	

lookup_ip_term(IP, Store) ->
    case lookup_ip(IP, Store) of
	not_found -> 
	    term:join0();
	{found, Data} ->
	    ip_term(Data)
    end.

%%% IP Data

%% {ip_data, Term, Links, Dependents}

new_ip_data() ->
    {ip_data, term:join0(), [], []}.

ip_term({ip_data, Term, _Links, _Dependent}) ->
    Term.

update_ip_term({ip_data, _, Links, Dependent}, Term) ->
    {ip_data, Term, Links, Dependent}.

ip_links({ip_data, _Term, Links, _Dependent}) ->
    Links.

update_ip_links({ip_data, Term, _, Dependent}, Links) ->
    {ip_data, Term, Links, Dependent}.

ip_dependent({ip_data, _Term, _Links, Dependent}) ->
    Dependent.

update_ip_dependent({ip_data, Term, Links, _}, NewDependentList) ->
    {ip_data, Term, Links, NewDependentList}.


%%%

make_context_table(Store, Interesting) ->
    make_context_table(gb_trees:to_list(Store), gb_trees:empty(), Interesting).

make_context_table([{IP, _} | L], Table, I) ->
    Context = term:ip_context(IP),
    Name = term:ip_name(IP),
    case I(Name) of
	{ok, RP} ->
	    case gb_trees:lookup(RP, Table) of 
		none ->
		    Table1 = gb_trees:insert(RP, {[Context], [Name]}, Table);
		{value, {ContextList, NameList}} ->
		    case lists:member(Context, ContextList) 
			and lists:member(Name, NameList) of
			true ->
			    Table1 = Table;
			false ->
			    case lists:member(Context, ContextList) of
				true ->
				    ContextList1 = ContextList;
				false ->
				    ContextList1 = [Context | ContextList]
			    end,
			    case lists:member(Name, NameList) of
				true ->
				    NameList1 = NameList;
				false ->
				    NameList1 = [Name | NameList]
			    end,
			    Table1 = 
				gb_trees:enter(RP, {ContextList1, NameList1}, Table)
		    end
	    end;
	_ ->
	    Table1 = Table
    end,
    make_context_table(L, Table1, I);
make_context_table([], Table, _) ->
%    io:format("~nContext table:~n~p",[gb_trees:to_list(Table)]),
    Table.


