%%% File    : summary.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Description : Summarize analysis statistics.
%%% Created : 15 Oct 2002 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(summary).
-export([summary/2]).

summary(File,  M) ->
    List = read_file(File),
    NameTable = collect_names(List, M),
    Q = quant(List, NameTable),
    show_quant(Q).

show_quant(Q) ->
    lists:foreach(fun show_strategy/1, Q).

show_strategy({strategy, Tag, List}) ->
    io:format("~n~n=== Strategy: ~w ===",[Tag]),
    G = fun ({Label, N}) ->
		io:format("~n~16w:~8w",[Label, N])
	end,
    lists:foreach(G, List),
    H = fun({any, N},{Tot1, Any1}) ->
		{Tot1+N, Any1+N};
	   ({_, N},{Tot1, Any1}) ->
		{Tot1+N, Any1}
	end,
    {Tot, Any} = lists:foldl(H, {0,0}, List),
    io:format("~nTotal: ~w, percentage determined: ~5.1f", [Tot, (Tot-Any)*100/Tot]).

read_file(File) ->
    {ok, H} = file:open(File, [read]),
    List = collect_file(H),
    file:close(H),
    List.

collect_file(H) ->
    case io:read(H, '') of
	{ok, T} ->
	    [T|collect_file(H)];
	eof ->
	    []
    end.


collect_names(L, M) ->
    collect_names(L, M, gb_trees:empty()).

collect_names([{measurement, _, {module, M, FList}}|L], M, Table) ->
    Table1 = npl_function_list(FList, Table),
    collect_names(L, M, Table1);
collect_names([{measurement, _, {module, M1, FList}}|L], M, Table) ->
    collect_names(L, M, Table);
collect_names([], M, Table) ->
    Table.


npl_function_list([{function, Fnid, NameDataList} | FList], Table) ->
    F = fun ({name, Name, _}, Table0) ->
		gb_trees:enter(Name, true, Table0)
	end,
    Table1 = lists:foldl(F, Table, NameDataList),
    npl_function_list(FList, Table1);
npl_function_list([], Table) ->
    Table.

quant(L, NameTable) ->
    lists:map(fun(Mea) ->
		      quant_measurement(Mea, NameTable)
	      end,
	      L).


quant_measurement({measurement, Tag, {module, _, FList}}, NameTable) ->
    Table = quant_flist(FList, NameTable),
    Table2 = add_nonames(gb_trees:to_list(NameTable), Table),
    Inv = invert(Table2),
    S = {strategy, Tag, gb_trees:to_list(Inv)},
	S.

quant_flist(List, NameTable) ->
    F = fun({name, Name, Data}, Table) ->
	       case gb_trees:is_defined(Name, NameTable) of
		   true ->
		       gb_trees:insert(Name, q(Data), Table);
		   false ->
		       Table
	       end
       end,
    G = fun ({function, _, NameDataList}, Table) ->
		lists:foldl(F, Table, NameDataList)
	end,
    lists:foldl(G, gb_trees:empty(), List).
    
add_nonames(NameTrueList, Table) ->
    F = fun({Name, true}, Table0) ->
		case gb_trees:is_defined(Name, Table0) of
		    true ->
			Table0;
		    false ->
			gb_trees:insert(Name, q(none), Table0)
		end
	end,
    lists:foldl(F, Table, NameTrueList).


invert(Table) ->
    F = fun({Key, Data}, Table0) ->
		case gb_trees:lookup(Data, Table0) of
		    none ->
			gb_trees:insert(Data, 1, Table0);
		    {value, N} ->
			gb_trees:update(Data, N+1, Table0)
		end
	end,
    lists:foldl(F, gb_trees:empty(), gb_trees:to_list(Table)).
		
q(any) ->
    any;
q(X) ->
    X.
