%%% File    : term.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@csd.uu.se>
%%% Purpose : Represent Erlang terms
%%% Created : 15 Feb 1999 by Sven-Olof Nystr|m <svenolof@csd.uu.se>

-module(term).
-author('svenolof@csd.uu.se').

-export([make_atom/1, make_atom_any/0, is_atom/1, atom_val/1, 
	 make_int/1, make_int_any/0, is_int/1, make_float/1, float_val/1,
	 make_nil/0, is_nil/1, make_cons/2, is_cons/1, cons_car/1, 
         cons_cdr/1, 
	 make_string/1, 
	 make_tuple/1, make_tuple_clean/1, is_tuple/1,
	 make_vector/1, is_vector/1, vector_arg_list/1,
	 unsafe/2,
	 make_ip/2, ip_context/1, ip_name/1,
	 make_numbered_ip/1, numbered_ip_number/1,
	 make_fun/3, fun_ap/1, fun_arity/1, fun_pos/1, 
	 parameter/1, parameter_name/1,
	 make_formal/4, is_formal/1, formal_module/1, formal_name/1, 
         formal_arg_list/1, formal_arity/1,
	 universal/0, is_universal/1,
	 make_pattern_var/1, pattern_var_name/1,
	 make_alias_pattern/2,
	 alias_pattern_var_name/1,
	 alias_pattern_pattern/1,
	 apply_functor/2, term_functor/1, term_functor_arg_list/1,
	 join/1, join0/0, join2/2, join_arg_list/1, 
	 included/2, is_empty/1,
	 do_all/4,
	 make_pid/1, make_pid/3, make_pid_with_mbox/1, 
	 pid_location/1, pid_mbox/1, 
	 make_fail/0,
	 type/1, term_tag/1,
	 trunc/0]).

%% Atom

is_atom(Term) ->
    case is_functor_term(Term) of
	true ->
	    atomc == term_functor(Term);
	false  ->
	    false
    end.

make_atom(Name) ->
    apply_functor(atomc, [apply_functor({atoma, Name}, [])]).

make_atom_any() ->
    apply_functor(atomc, [universal()]).

atom_val(Term) ->
    case term_functor(Term) of
	atomc ->
	    case term_functor_arg_list(Term) of
		[Arg] ->
		    case term_functor(Arg) of
			{atoma, A} ->
			    A
		    end
	    end
    end.
			    
%% Int

make_int(Int) ->
    apply_functor(intc, [apply_functor({inta, Int}, [])]).

make_int_any() ->
    apply_functor(intc, [universal()]).
    
is_int(Term) ->
    case is_functor_term(Term) of
	true ->
	    intc == term_functor(Term);
	false  ->
	    false
    end.

%% Float

make_float(_g) ->
    make_float_any().

make_float_any() ->
    apply_functor(floatc, [universal()]).

float_val(Term) ->
    case Term of
	{atomic, Term1} ->
	    true;
	Term1 ->
	    true
    end,
    [Arg] = term_functor_arg_list(Term1),
    {floata, Val} = term_functor(Arg),
    Val.


%%

make_pid(Context) ->
    make_pid(0, 0, Context).
make_pid(Loc, _, Context) ->
    MBox = tag:spawn_mbox(Loc, Context),
    Dict = tag:pid_dict(Loc, Context),
    apply_functor({pid, Loc, MBox, Dict},[]).

make_pid_with_mbox(MBox) ->
    apply_functor({pid, 0, MBox, term:join0()},[]).


pid_mbox(Term) ->
    case is_functor_term(Term) of
	true ->
	    case term_functor(Term) of
		{pid, _Location, MBox, _Dict} ->
		    MBox
	    end
    end.

pid_location(Term) ->
      case is_functor_term(Term) of
	true ->
	    case term_functor(Term) of
		{pid, Location, _MBox, _Dict} ->
		    Location
	    end
    end.
    

%%

make_tuple(TermList) ->
    apply_functor(tuple, [list_to_ablist(TermList)]).

make_tuple_clean(Term) ->
    apply_functor(tuple, [Term]).



is_tuple(Term) ->
    case is_functor_term(Term) of
	true ->
	    case term_functor(Term) of
		tuple ->
		    true;
		_ ->
		    false
	    end;
	false ->
	    false
    end.


list_to_ablist([]) ->
    make_nil();
list_to_ablist([X|L]) ->
    make_cons(X, list_to_ablist(L)).

% Cons

make_cons(CarTerm, CdrTerm) ->
    apply_functor(cons, [CarTerm, CdrTerm]).

is_cons(Term) ->
    case term_tag(Term) of
	functor ->
	    cons == term_functor(Term);
	_->
	    false 
    end.

cons_car(Term) ->
    [Car, _] = term_functor_arg_list(Term),
    Car.

cons_cdr(Term) ->
    [_, Cdr] = term_functor_arg_list(Term),
    Cdr.

make_nil() ->
    apply_functor(nil, []).

is_nil(Term) ->
    nil == term_functor(Term).


make_string(_) ->
    tag:any_string().

%%% Vectors. Not for now.

make_vector([Term]) ->
    Term;
make_vector(TermList) ->
    apply_functor({vector, length(TermList)}, TermList).

is_vector(Term) ->
    case is_functor_term(Term) of
	true ->
	    case term_functor(Term) of
		{vector, _} ->
		    true;
		_ ->
		    false
	    end;
	false ->
	    false
    end.
		    
vector_arg_list(Term) ->
    term_functor_arg_list(Term).

%%% Unsafes

unsafe(Term, AS) ->
    case as:dict_get(safe, AS) of
	true ->
	    Term;
	false ->
	    universal()
    end.


%%%

term_tag({ip, _, _}) ->
    ip;
term_tag({join, _}) ->
    join;
term_tag({universal}) ->
    universal;
term_tag({functor, {pid, _, _, _}, _}) ->
    pid;
term_tag({functor, _, _}) ->
    functor;
term_tag({'fun',_,_,_}) ->
    'fun';
term_tag({pattern_var, _}) ->
    pattern_var;
term_tag({alias_pattern, _, _}) ->
    alias_pattern;
term_tag({parameter, _}) ->
    parameter;
term_tag({trunc}) ->
    trunc.
    

%%% Formal types

make_formal(Module, Name, Arity, ArgList) when Arity == length(ArgList) ->
    apply_functor({formal, Module, Name, Arity}, ArgList).

is_formal(Term) ->
    case term_tag(Term) of
	functor ->
	    case term_functor(Term) of
		{formal, _, _, _} ->
		    true;
		_ ->
		    false
	    end;
	_ ->
	    false
    end.

formal_module(Term) ->
    {formal, Module, _Name, _Arity} = term_functor(Term),
    Module.

formal_name(Term) ->
    {formal, _Module, Name, _Arity} = term_functor(Term),
    Name.

formal_arity(Term) ->
    {formal, _Module, _Name, Arity} = term_functor(Term),
    Arity.

formal_arg_list(Term) ->
    term_functor_arg_list(Term).

%%% Fun

make_fun(AP, Pos, Arity) ->
    {'fun', AP, Pos, Arity}.

fun_ap({'fun', AP, _Arity}) ->
    AP;
fun_ap({'fun', AP, _, _Arity}) ->
    AP.


fun_arity({'fun', _AP, Arity}) ->
    Arity;
fun_arity({'fun', _AP, _, Arity}) ->
    Arity.

fun_pos({'fun', _AP, Pos, _}) ->
    Pos.

%% Pattern variables. Not supposed to show up in regular analysis
%% terms.

make_pattern_var(Name) ->
    {pattern_var, Name}.

pattern_var_name({pattern_var, Name}) ->
    Name.

%% Pattern alias. No, we don't expect them to occur in analysis terms.

make_alias_pattern(Name, Pattern) ->
    {alias_pattern, Name, Pattern}.

alias_pattern_var_name({alias_pattern, Name, _Pattern}) ->
    Name.

alias_pattern_pattern({alias_pattern, _Name, Pattern}) ->
    Pattern.

%% Join


is_empty({join, U}) ->
    gb_trees:is_empty(U);
is_empty(_) ->
    false.
    
join0() ->
    {join, gb_trees:empty()}.

join2({univeral}, _) ->
    {universal};
join2(_, {universal}) ->
    {universal};
join2({join, U1}, {join, U2}) ->
    make_join(union(U1, U2));
join2({join, U1}, T) ->
    make_join(insert_term(T, U1));
join2(T, {join, U1}) ->
    make_join(insert_term(T, U1));
join2(T1, T2) ->
    make_join(insert_term(T1, insert_term(T2, gb_trees:empty()))).

make_join(Table) ->
    case gb_trees:size(Table) == 1 of
	true ->
	    [{Term,_}] = gb_trees:to_list(Table),
	    Term;	    
	false ->
	    {join, gb_trees:balance(Table)}
    end.

join([T]) ->
    T;
join([T|L]) ->
    join2(T, join(L));
join([]) ->
    join0().

insert_term(T, U) ->
    case gb_trees:lookup(T, U) of
	none ->
	    gb_trees:insert(T, true, U);
	{value, _} ->
	    U
    end.

union(U1, U2) ->
    L = gb_trees:to_list(U1),
    lists:foldl(fun({Key, true}, S) ->
		  insert_term(Key, S)
		end,
		U2,
		L).

join_arg_list({join, U}) ->
    lists:map(fun({Key, _}) -> Key end, gb_trees:to_list(U)).

%% Inclusion. At present, a fairly crude test. To be improved.

included(_T, {universal}) ->
    true;
included(T, T) ->
    true;
included({join, U}, T) ->
    case gb_trees:size(U) of
	0 -> 
	    true;
	1 ->
	    [{T1,_}] = gb_trees:to_list(U),
	    included(T1, T);
	_ ->
	    false
    end;
included(T, {join, U}) ->
    case gb_trees:lookup(T, U) of
	{value, _} ->
	    true;
	none ->
	    false
    end;
included(_, _) ->
    false.
    
	




%%% Type

type(Term) ->
    case term_tag(Term) of
	functor ->
	    case term_functor(Term) of
		{pid, _, _, _} ->
		    pid;
		_ ->
		    constructor
	    end;
	'fun' ->
	    'fun';
	ip ->
	    ip;
	universal ->
	    universal;
	parameter ->
	    parameter;
	join ->
	    join;
	pid ->
	    pid;
	node ->
	    node    
    end.

%% do all

do_all(Term, F, Context, AS) ->
    case type(Term) of
	join ->
	    TermList = join_arg_list(Term),
	    lists:foldl(F, AS, TermList);
	ip ->
		 as:lookup_cont(Term, F, Context, AS);
	_ ->
		 F(Term, AS)
    end.

%%% Functors...

is_functor_term({functor, _, _}) ->
    true;
is_functor_term(_) ->
    false.

apply_functor(F, ArgList) ->
    case lists:any(fun is_empty/1, ArgList) of
	true -> join0();
	false -> {functor, F, ArgList}
    end.

term_functor({functor, F, _ArgList}) ->
    F.

term_functor_arg_list({functor, _F, ArgList}) ->
    ArgList.
    
%% Universal

universal() ->
    {universal}.

is_universal({universal}) ->
    true;
is_universal(_) ->
    false.

%% Parameter
%
% A parameter is generated during analysis of a function specified
% with a polymorphic type. Thus, a parameter represents an
% all-quantified variable. It follows that the analysis may not make
% any assumptions about a parameter, so a parameter may not be
% instantiated or matched against.

parameter(Name) ->
    {parameter, Name}.

parameter_name({parameter, Name}) ->
    Name.




%%% Fail. To report result of matching.

make_fail() ->
    apply_functor(fail, []).

% ip

make_ip(Name, Context) ->
    {ip, Name, Context}.

ip_context({ip, _, Context}) ->
    Context.

ip_name({ip, Name, _}) ->
    Name.

% Numbered ip's. Hopefully, we'll receive a new nuber each time.

make_numbered_ip(N) ->
    {ip, numbered, N}.

numbered_ip_number({ip, numbered, N}) ->
    N.

trunc() ->
    {trunc}.
