%% File    : trav.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@csd.uu.se>
%%% Purpose : Traversing Core Erlang expressions
%%% Created :  8 Jul 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(trav).
-author('svenolof@csd.uu.se').

-export([trav_mdef/2, trav_fdef/2]).

% Sketching how to traverse a Core Erlang expression. 

trav(Exp, {P, F}) ->
    Exp1 =
	case kinds:exp_type(Exp) of 
	    var ->
		Exp;
	    constructor ->
		trav_constructor(Exp, {P, F});
	    control ->
		trav_control(Exp, {P, F});
	    dynamic ->
		trav_dynamic(Exp, {P, F});
	    call ->
		trav_call(Exp, {P, F});
	    local_call ->
		trav_apply(Exp, {P, F});
	    apply ->
		trav_apply(Exp, {P, F});
	    'fun' ->
		trav_fun(Exp, {P, F});
	    fref ->
		Exp
	end,
    case P(Exp1) of
	true ->
	    F(Exp1);
	false ->
	    Exp1    
    end.

trav_constructor(Exp, PF) ->
    C = kinds:exp_constructor(Exp),
    L = kinds:constructor_exp_arg_list(Exp),
    L1 = trav_list(L, PF),
    kinds:copy_ann(kinds:make_constructor_exp(C, L1), Exp).

trav_list([Exp | ExpList], PF) ->
    Exp1 = trav(Exp, PF),
    ExpList1 = trav_list(ExpList, PF),
    [Exp1 | ExpList1];
trav_list([], _) ->
    [].
	    
trav_control(Exp, PF) ->
    case kinds:control_exp_type(Exp) of
	'case' ->
	    trav_case(Exp, PF);
	'let' ->
	    trav_let(Exp, PF);
	letrec ->
	    trav_letrec(Exp, PF);
	'receive' ->
	    trav_receive(Exp, PF);
	seq ->
	    trav_seq(Exp, PF);
	'catch' ->
	    trav_catch(Exp, PF);
	'try' ->
	   trav_try(Exp, PF)
    end.


trav_case(Exp, PF) ->
    Arg = kinds:case_exp_arg(Exp),
    ClauseList = kinds:case_exp_clause_list(Exp),
    Arg1 = trav(Arg, PF),
    ClauseList1 = 
	seq_map(fun(Clause) ->
			  trav_clause(Clause, PF)
		  end,
		  ClauseList),
    kinds:copy_ann(kinds:make_case(Arg1, ClauseList1), Exp).

trav_seq(Exp, PF) ->
    Arg = kinds:seq_arg(Exp),
    Body = kinds:seq_body(Exp),
    Arg1 = trav(Arg, PF),
    Body1 = trav(Body, PF),
    kinds:copy_ann(kinds:make_seq(Arg1, Body1), Exp).

trav_catch(Exp, PF) ->
    Body = kinds:catch_body(Exp),
    Body1 = trav(Body, PF),
    kinds:copy_ann(kinds:make_catch(Body1), Exp).
    
trav_try(Exp, PF) ->
    Arg = kinds:try_arg(Exp),
    VarList = kinds:try_var_list(Exp),
    Body = kinds:try_body(Exp),

    Arg1 = trav(Arg, PF),
    VarList1 = trav_list(VarList, PF),
    Body1 = trav(Body, PF),
    
    Evars = kinds:try_e_var_list(Exp),
    Handler = kinds:try_handler(Exp),

    kinds:copy_ann(
      kinds:make_try(Arg1, VarList1, Body1, Evars, Handler), Exp).

trav_receive(Exp, PF) ->
    ClauseList = kinds:receive_exp_clause_list(Exp),
    Timeout = kinds:receive_exp_timeout(Exp),
    Action = kinds:receive_exp_action(Exp),
    ClauseList1 = 
	seq_map(fun(Clause) ->
			  trav_clause(Clause, PF)
		  end,
		  ClauseList),
    Timeout1 = trav(Timeout, PF),
    Action1 = trav(Action, PF),
    kinds:copy_ann(kinds:make_receive(ClauseList1, Timeout1, Action1), Exp).
    
trav_clause(Clause, {P, F}) ->
    PatternList = kinds:clause_pattern_list(Clause),
    Guard = kinds:clause_guard(Clause),
    Body = kinds:clause_body(Clause),
    PatternList1 = lists:map(fun(Pattern) ->
				     trav_pattern(Pattern,  {P, F})
			     end, PatternList),
    Guard1 = trav_guard(Guard, {P, F}),
    Body1 = trav(Body, {P, F}),
    Clause1 = kinds:copy_ann(kinds:make_clause(PatternList1, Guard1, Body1), Clause),
    case P(Clause1) of
	true ->
	    F(Clause1);
	false ->
	    Clause1
    end.

trav_pattern(Pattern, {P, F}) ->
    Pattern1 =
	case kinds:pattern_type(Pattern) of 
	    var ->
		Pattern;
	    alias ->
		Var = kinds:alias_var(Pattern),
		Var1 = trav_pattern(Var, {P, F}),
		Pat = kinds:alias_pattern(Pattern),
		Pat1 = trav_pattern(Pat, {P, F}),
		kinds:make_alias(Var1, Pat1);
	    constructor ->
		trav_constructor_pattern(Pattern, {P, F})
	end,
    case P(Pattern1) of
	true ->
	    F(Pattern1);
	false ->
	    Pattern1    
    end.

trav_constructor_pattern(Pattern, PF) ->
    C = kinds:pattern_constructor(Pattern),
    L = kinds:constructor_pattern_arg_list(Pattern),
    L1 = seq_map(fun(P) ->
			   trav_pattern(P, PF)
		   end,
		   L),
    kinds:copy_ann(kinds:make_constructor_pattern(C, L1), Pattern).

trav_guard(G, PF) ->
    G1 = unlet(G, gb_trees:empty()),
    trav(G1, PF).

%%% In a sane world, code such as this wouldn't be necessary...

unlet(Exp, Env) ->
    case kinds:exp_type(Exp) of 
	 var ->
	    Var = kinds:name_atom(Exp),
	    case gb_trees:lookup(Var, Env) of
		{value, Exp1} ->
		    Exp1;
		none ->
		    Exp
	    end;
	constructor ->
	    L = kinds:constructor_exp_arg_list(Exp),
	    C = kinds:exp_constructor(Exp),
	    L1 = unlet_list(L, Env),
	    kinds:make_constructor_exp(C, L1);
	control ->
	    unlet_control(Exp, Env);
	call ->
	    Module = kinds:call_module(Exp),
	    Name = kinds:call_name(Exp),
	    ArgList = kinds:call_arg_list(Exp),
	    Module1 = unlet(Module, Env),
	    Name1 = unlet(Name, Env),
	    ArgList1 = unlet_list(ArgList, Env),
	    kinds:make_call(Module1, Name1, ArgList1);
	apply ->
	    Op = kinds:apply_op(Exp),
	    ArgList = kinds:apply_arg_list(Exp),
	    Op1 = unlet(Op, Env),
	    ArgList1 = unlet_list(ArgList, Env),
	    kinds:make_apply(Op1, ArgList1);
	fref ->
	    Exp
	end.

unlet_control(Exp, Env) ->
    case kinds:control_exp_type(Exp) of
	'let' ->
	    case kinds:let_exp_var_list(Exp) of
		[Var] ->
		    Arg = kinds:let_exp_arg(Exp),
		    Body = kinds:let_exp_body(Exp),
		    Arg1 = unlet(Arg, Env),
		    Body1 = unlet(Body, Env),

		    Env1 = gb_trees:insert(kinds:name_atom(Var), Arg1, Env),
		    unlet(Body1, Env1);
		VarList ->
		    Arg = kinds:let_exp_arg(Exp),
		    Body = kinds:let_exp_body(Exp),
		    Arg1 = unlet(Arg, Env),
		    Body1 = unlet(Body, Env),
		    kinds:make_let(VarList, Arg1, Body1)
	    end;
	'try' ->
	    Arg = kinds:try_arg(Exp),
	    VarList = kinds:try_var_list(Exp),
	    Body = kinds:try_body(Exp),

	    Arg1 = unlet(Arg, Env),
	    Body1 = unlet(Body, Env),
            
            Evars = kinds:try_e_var_list(Exp),
            Handler = kinds:try_handler(Exp),
            Handler1 = unlet(Handler, Env),

	    kinds:make_try(Arg1, VarList, Body1, Evars, Handler1)
    end.
    
unlet_list(L, Env) ->
    lists:map(fun(Exp) ->
		      unlet(Exp, Env)
	      end,
	      L).    

trav_let(Exp, PF) ->
    VarList = kinds:let_exp_var_list(Exp),
    Arg = kinds:let_exp_arg(Exp),
    Body = kinds:let_exp_body(Exp),
    VarList1 = lists:map(fun(Var) ->trav(Var, PF) end, VarList),
    Arg1 = trav(Arg, PF),
    Body1 = trav(Body, PF),
    kinds:copy_ann(kinds:make_let(VarList1, Arg1, Body1), Exp).

trav_letrec(Exp, PF) ->
    DefList = kinds:letrec_def_list(Exp),
    Body = kinds:letrec_body(Exp),
    DefList1 = lists:map(fun({Var, Fun}) ->
				 {trav(Var, PF), trav(Fun, PF)}
			 end, DefList),
    Body1 = trav(Body, PF),
    kinds:copy_ann(kinds:make_letrec(DefList1, Body1), Exp).

trav_dynamic(Exp, PF) ->
    Mod = kinds:dynamic_mod(Exp),
    Name = kinds:dynamic_name(Exp),
    Arity = kinds:dynamic_arity(Exp),
    Mod1 = trav(Mod, PF),
    Name1 = trav(Name, PF),
    Arity1 = trav(Arity, PF),
    kinds:copy_ann(kinds:make_dynamic(Mod1, Name1, Arity1), Exp).

trav_fun(Exp, PF) ->
    VarList = kinds:fun_var_list(Exp),
    Body = kinds:fun_body(Exp),
    VarList1 = seq_map(fun(Var) ->
				 trav(Var, PF)
			 end,
			 VarList),
    Body1 = trav(Body, PF),
    kinds:copy_ann(kinds:make_fun(VarList1, Body1), Exp).

trav_apply(Exp, PF) ->
    Op = kinds:apply_op(Exp),
    ArgList = kinds:apply_arg_list(Exp),
    Op1 = trav(Op, PF),
    ArgList1 = seq_map(fun(Arg) ->
				trav(Arg, PF)
			end,
			ArgList),
    kinds:copy_ann(kinds:make_apply(Op1, ArgList1), Exp).

trav_call(Exp, PF) ->
    Module = kinds:call_module(Exp),
    Name = kinds:call_name(Exp),
    ArgList = kinds:call_arg_list(Exp),
    Module1 = trav(Module, PF),
    Name1 = trav(Name, PF),
    ArgList1 = seq_map(fun(Arg) ->
				trav(Arg, PF)
			end,
			ArgList),
    kinds:copy_ann(kinds:make_call(Module1, Name1, ArgList1), Exp).


%%% Traverse a module

trav_mdef(Module, PF) ->
    Name = kinds:mdef_name(Module),
    Body = kinds:mdef_body(Module),
    ExportList = kinds:mdef_export_list(Module),
    ExportList1 = lists:map(fun(Fref) ->
				    trav(Fref, PF)
			    end,
			    ExportList),
    Body1 = seq_map(fun(FDef) ->
			      trav_fdef(FDef, PF)
		      end,
		      Body),
    MDef1 = kinds:copy_ann(kinds:make_mdef(Name, ExportList1, Body1), Module),
    {P, F} = PF,
    case P(MDef1) of
	true ->
	    F(MDef1);
	false ->
	    MDef1  
    end.

%%% Traverse a function definition

trav_fdef(FDef, PF) ->
    Functor = kinds:fdef_name(FDef),
    Arity = kinds:fdef_arity(FDef),
    Body = kinds:fdef_body(FDef),
    Body1 = trav(Body, PF),
    FDef1 = kinds:copy_ann(kinds:make_fdef(Functor, Arity, Body1), FDef),
    {P, F} = PF,
    case P(FDef1) of
	true ->
	    F(FDef1);
	false ->
	    FDef1  
    end.


% A map which evaluates the element in order!

seq_map(F, [X | L]) ->
    Y = F(X),
    [Y | seq_map(F, L)];
seq_map(_, []) ->
    [].
    


