%%% File    : walker.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@csd.uu.se>
%%% Purpose : The analysis of Erlang expressions. 
%%% Created : 17 Feb 1999 by Sven-Olof Nystr|m <svenolof@csd.uu.se>

%%% Expression-walking functions typically take four arguments;
%%%
%%% 1) the expression to be analyzed,
%%% 2) the environment,
%%% 3) the current context, and
%%% 4) the state of the analysis,
%%%
%%% and return two values;analyze:check_module(lists)
%%%
%%% 1) a term representing an abstraction of the value returned by the
%%%    expression, 
%%% 2) the new analysis state.
%%%
%%% The program being analyzed and the name of the current module are
%%% stored as global variables.

-module(walker).

-author('svenolof@csd.uu.se').

-export([definition/3, exp/4, exp_list/4, apply_fun/5,
	 set_current_fnid/1, current_fnid/0]).

%%% Function definitions

definition(Def, Context, AS) ->
    Body = kinds:fdef_body(Def),
    GetName = 
	fun(Var) -> 
		kinds:name_atom(Var) 
	end,
    ArgList = lists:map(GetName, kinds:fun_var_list(Body)),
    Exp = kinds:fun_body(Body),
    Env = param:initialize_environment(ArgList, Context),
    {Term, AS1} = exp(Exp, Env, Context, AS),
    AS2 = as:insert(tag:result(Context), Term, AS1),
    AS2.

%%% Expressions

exp(Exp, Env, Context, AS) ->
    case kinds:exp_type(Exp) of
	var ->
	    case kinds:is_fref(Exp) and not env:defined(kinds:var_name(Exp), Env) of
		false ->
		    Var = kinds:name_atom(Exp),
		    Term = env:lookup(Var, Env),
		    {Term, AS};
		true ->
		    fref_exp(Exp, Env, Context, AS)
	    end;
	constructor ->
	    constructor_exp(Exp, Env, Context, AS);
	control ->
	    control_exp(Exp, Env, Context, AS);
	apply ->
	    apply_exp(Exp, Env, Context, AS);
	call ->
	    call:call_exp(Exp, Env, Context, AS);
	'fun' ->
	    fun_exp(Exp, Env, Context, AS);
 	fref ->
	    fref_exp(Exp, Env, Context, AS)
    end.


exp_list([], _Env, _Context, AS) ->
    {[],  AS};
exp_list([E|EList], Env, Context, AS) ->
    {Term,  AS1} = exp(E, Env, Context, AS),
    {TermList, AS2} = exp_list(EList, Env, Context, AS1),
    {[Term|TermList], AS2}.

%%% Constructor expressions

constructor_exp(Exp, Env, Context, AS) ->
    case kinds:is_literal(Exp) of
	true ->
	    {literal_exp(Exp), AS};
	false ->
	    C = kinds:exp_constructor(Exp),
	    case kinds:constructor_type(C) of
		tuple ->
		    tuple_exp(Exp, Env, Context, AS);
		vector ->
		    vector_exp(Exp, Env, Context, AS);
		cons ->
		    cons_exp(Exp, Env, Context, AS)
	    end
    end.

tuple_exp(Exp, Env, Context, AS) ->
    ElementList = kinds:constructor_exp_arg_list(Exp),
    {ABElementList, AS1} = exp_list(ElementList, Env, Context, AS),
    {term:make_tuple(ABElementList),  AS1}.

vector_exp(Exp, Env, Context, AS) ->
    ElementList = kinds:constructor_exp_arg_list(Exp),
    {TermList, AS1} = exp_list(ElementList, Env, Context, AS),
    {term:make_vector(TermList), AS1}.

cons_exp(Cons, Env, Context, AS) ->
    [Head, Tail] = kinds:constructor_exp_arg_list(Cons),
    {[ABHead, ABTail], AS1} = exp_list([Head, Tail], Env, Context, AS),
    {term:make_cons(ABHead, ABTail),  AS1}.

%%% Control expressions

control_exp(E, Env, Context, AS) ->
    case kinds:control_exp_type(E) of 
	'let' ->
	    let_exp(E, Env, Context, AS);
	letrec ->
	    letrec_exp(E, Env, Context, AS);
	'case' ->
	    case_exp(E, Env, Context, AS);
	'receive' ->
	    receive_exp(E, Env, Context, AS);
	seq ->
	    seq_exp(E, Env, Context, AS);
	'catch' ->
	    catch_exp(E, Env, Context, AS)
    end.

let_exp(E, Env, Context, AS) ->
    VarList = lists:map(fun(V) ->
				kinds:name_atom(V)
			end,
			kinds:let_exp_var_list(E)),			
    Arg = kinds:let_exp_arg(E),
    Body = kinds:let_exp_body(E),
    DestList = tag:vector(Context, length(VarList), kinds:pos(E)),
    {Term, AS1} = exp(Arg, Env, Context, AS),
    AS2 = wait:wait_vector_r(Term, DestList, Context, AS1),
    Env1 = env:bind_list(VarList, DestList, Env),
    exp(Body, Env1, Context, AS2).

letrec_exp(E, Env, Context, AS) ->
    DefList = kinds:letrec_def_list(E),
    Body = kinds:letrec_body(E),
    VarList = lists:map(fun({V, _}) ->
				kinds:name_atom(V)
			end,
			DefList),
    FunList = lists:map(fun({_, Fun}) ->
				Fun
			end,
			DefList),
    
    Pos = kinds:pos(E),
    IPList = param:invent_fun_arg_list(length(VarList), Context, Pos),
    Env1 = env:bind_list(VarList, IPList, Env),
    {TermList, AS1} = exp_list(FunList, Env1, Context, AS),
    AS2 = bind_var_list(IPList, TermList, AS1),
    exp(Body, Env1, Context, AS2).

bind_var_list([IP|IPList],[Term|TermList],AS) ->
    AS1 = as:insert(IP, Term, AS),
    bind_var_list(IPList, TermList, AS1);
bind_var_list([], [], AS) ->
    AS.

case_exp(E, Env, Context, AS) ->
    Arg = kinds:case_exp_arg(E),
    ClauseList = kinds:case_exp_clause_list(E),
    Pos = kinds:pos(E),
    {Key, AS1} = exp(Arg, Env, Context, AS),
    clause_select(ClauseList, Key, Pos, 1, Env, Context, AS1).

clause_select([Clause | ClauseList], Key, Pos, Index, Env, Context, AS) ->
    {Term1, Key1, AS1} = match:clause(Clause, Key, Env, Context, AS),
    {Term2, AS2} = 
        clause_select(ClauseList, Key1, Pos, Index+1, Env, Context, AS1),
    {term:join2(Term1, Term2), AS2};
clause_select([], _Key, _Pos, _Index, _Env, _Context, AS) ->
    {term:join0(), AS}.


%% Receive

receive_exp(E, Env, Context, AS0) ->
    ClauseList = kinds:receive_exp_clause_list(E),
    TimeOut = kinds:receive_exp_timeout(E),
    Action = kinds:receive_exp_action(E),
    {Self, AS} = 
	case as:dict_get(optimistic, AS0) of
	    true ->
		case program:attribute(mbox,  current_fnid()) of
		    {value, Spec} ->
			{MBox, AS01} = 
                            spec_generate:generate(Spec, env:empty(), 
                                                   Context, AS0),
			{term:make_pid_with_mbox(MBox), AS01};
		    none ->
			{env:lookup('$self', Env), AS0}
		end;
	    false ->
		{env:lookup('$self', Env), AS0}
	end,
    Pos = kinds:pos(E),
    Result_IP = tag:receive_result(Pos, Context),
    F = fun (Self1, BS) ->
		Key = term:unsafe(term:pid_mbox(Self1), BS),
		{Result, BS1} = 
		    clause_select(ClauseList, Key, Pos, 1, Env, Context, BS),
		as:insert(Result_IP, Result, BS1)
	end,
    AS1 = term:do_all(Self, F, Context, AS),
    case kinds:is_int(TimeOut) of
	true ->
	    {TimeoutResult,AS2} = exp(Action, Env, Context, AS1),
	    AS3 = as:insert(Result_IP, TimeoutResult, AS2);
	false ->
	    AS3 = AS1
    end,
    {Result_IP, AS3}.

seq_exp(E, Env, Context, AS) ->
    Arg = kinds:seq_arg(E),
    Body = kinds:seq_body(E),
    {_, AS1} = exp(Arg, Env, Context, AS),
    exp(Body, Env, Context, AS1).

catch_exp(Exp, Env, Context, AS) ->
    Body = kinds:catch_body(Exp),
     Pos = 42, %kinds:pos(Exp),
     NewExcept = tag:try_except(Pos, Context),
     Exce = term:universal(),
     Env0 = env:bind('$except', NewExcept, Env),
     {Term1, AS1} = exp(Body, Env0, Context, AS),
    {term:join([Term1, term:make_tuple([term:make_atom('EXIT'),Exce]), Exce]), AS1}.

% Semantics of catch: An exception consists of two components; the
% first indicates its class, one of the atoms 'EXIT', 'ERROR' and
% 'THROW', and the second is an arbitrary term indicating the reason
% for and context of the eexception. An expression catch E has the
% same semantics as the try-expression
% 
% try E
% catch (v1, v2) ->
%     case v1 of
%     'THROW' -> v2
%     _ -> {'EXIT',v2}
%     end


%%% Apply


apply_exp(Exp, Env, Context, AS) ->
    Operator = kinds:apply_op(Exp),
    case kinds:is_fref(Operator) 
        and not env:defined(kinds:var_name(Operator), Env) of
	true ->
	    call:local_call(Exp, Env, Context, AS);
	false ->
	    ExpList = kinds:apply_arg_list(Exp),
	    {ArgList, AS1} = exp_list(ExpList, Env, Context, AS),
	    {Op, AS2} = exp(Operator, Env, Context, AS1),
	    Pos = kinds:pos(Exp),
	    Result_IP = tag:apply_result(Pos, Context),
	    F = 
                fun (Term, AS_a) ->
			{Result, AS_b} = 
                            apply_fun(Term, ArgList, Env, Context, AS_a), 
			as:insert(Result_IP, Result, AS_b)
		end,
	    AS3 = term:do_all(Op, F, Context, AS2),
	    {Result_IP, AS3}
    end.
			    
apply_fun(Fun, ArgList, Env, _Context, AS) ->
    case term:type(Fun) of
	'fun' ->
	    case term:fun_arity(Fun) == length(ArgList) of
		true ->
		    AP = term:fun_ap(Fun),
		    Label = term:fun_pos(Fun),
		    AS2 = param:insert_fun_arg_list(AP, Label, ArgList, AS),
		    AS3 = as:insert(env:lookup('$except', Env), tag:except(AP), AS2),
		    AS4 = as:insert(tag:self(AP), env:lookup('$self', Env), AS3),
		    Term = tag:fun_result(AP, Label),
		    {Term, AS4};
		false ->
		    {term:join0(), AS}
	    end;
	universal ->
	    {term:universal(), AS};
	_ ->
	    {term:join0(), AS}
    end.
    

%%% fun_exp 

fun_exp(Exp, Env, Context, AS) ->
    Pos = kinds:pos(Exp),
    VarList = lists:map(fun(Var) ->
				kinds:name_atom(Var) 
			end,
			kinds:fun_var_list(Exp)),
    Body = kinds:fun_body(Exp),
    Arity = length(VarList),
    FunContext = Context,
    Fun = term:make_fun(Context, Pos, Arity),
    TermList = param:invent_fun_arg_list(length(VarList), FunContext, Pos),
    Env1 = env:bind_list(VarList, TermList, Env),
    {Result, AS1} = exp(Body, Env1, FunContext, AS),
    AS2 = as:insert(tag:fun_result(FunContext, Pos), Result, AS1),
    {Fun, AS2}.

    
%%% fref_exp

fref_exp(Exp, Env, Context, AS) ->
    Name = kinds:fref_name(Exp),
    Arity = kinds:fref_arity(Exp),
    Pos = kinds:pos(Exp),
    Self = env:lookup('$self', Env),
    Except = env:lookup('$except', Env),
    ArgList = param:invent_fun_arg_list(Arity, Context, Pos),
    F = program:make_fnid(context:current_module(Context), Name, Arity),
    {AP, AS1} = work:call_analysis(F, Pos, Context, AS),
    AS2 = param:insert_arg_list(AP, ArgList, AS1),
    AS3 = as:insert(Except, tag:except(AP), AS2),
    AS4 = as:insert(tag:self(AP), Self, AS3),
    Result = tag:result(AP),
    AS5 = as:insert(tag:fun_result(Context, Pos), Result, AS4),
    Fun = term:make_fun(Context, Pos, Arity),
    {Fun, AS5}.

%%% Walking literals

-define(min_string_length, 5).

literal_exp(Exp) ->
    C = kinds:concrete(Exp),
    literal(C).

literal(C= [_|_]) ->
    literal_cons(C);
literal(C) ->
    literal_other(C).

literal_cons(C) ->
    case (is_string(C)) andalso (length(C) >= ?min_string_length) of
	true ->
	    term:make_string(C);
	false ->
	    literal_ordinary_list(C)
    end.

is_string([C|L]) when is_integer(C), C band -256 == 0 ->
    is_string(L);
is_string([]) ->
    true;
is_string(_) ->
    false.


literal_ordinary_list([X|Y]) ->
    TermX = literal(X),
    TermY = literal_ordinary_list(Y),
    term:make_cons(TermX, TermY);
literal_ordinary_list(X) ->
    literal_other(X).

literal_other(C) when is_atom(C) ->
    term:make_atom(C);
literal_other(C) when is_integer(C) ->
    term:make_int(C);
literal_other(C) when is_float(C) ->
    term:make_float(C);
literal_other([])  ->
    term:make_nil();
literal_other(C) when is_tuple(C) ->
    L = tuple_to_list(C),
    TermL = literal_ordinary_list(L),
    term:make_tuple_clean(TermL).

%%%

current_fnid() ->
    get(current_fnid).

set_current_fnid(Name) ->
    put(current_fnid, Name),
    true.
    



	


