%%% File    : warn_filter.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : Filter warnings
%%% Created : 30 Jun 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(warn_filter).
-author('svenolof@harpo.it.uu.se').

-export([init/0, filter/2]).

init() ->
    put(warn_module_table, gb_trees:empty()).

filter({exception, Fnid, Exception}, AS) ->
    case filter_exception(Fnid, Exception, AS) of
	{fail, _} ->
	    false;
	{ok, _} ->
	    true
    end;
filter(_, _) ->
    false.

filter_exception(Fnid, Term, AS) ->
    filter_exception(Fnid, Term, c, AS).

filter_exception(Fnid, Term, Context, AS) ->
    Module = program:fnid_module(Fnid),
    case module_table(Module) of 
% Currently, table only stores exception declarations.
	none ->
	    {fail, AS};
	{value, Table} ->
	    case gb_trees:lookup(Fnid, Table) of
		none ->
		    {fail, AS};
		{value, PList} ->
		    match_union_list(PList, Term, Context, AS)
	    end
    end.

match(P, Term, Context, AS) ->
    case spec:spec_type(P) of
	functor ->
	    match_functor(P, Term, Context, AS);
	union ->
	    match_union(P, Term, Context, AS);	    
	var ->
	    {ok, AS}
    end.

match_functor(P, Term, Context, AS) ->
    case term:term_tag(Term) of 
	ip ->
	    {Term1, AS1} = as:lookup(Term, Context, AS),
	    match_functor(P, Term1, Context, AS1);
	any ->
	    {fail, AS};
	join ->
	    TermList = term:join_arg_list(Term),
	    match_join_list(P, TermList, Context, AS);
	functor ->
	    PC = spec:spec_functor(P),
	    TC = term:term_functor(Term),
	    case PC == TC of
		true ->
		    PatternList = spec:spec_functor_arg_list(P),
		    KeyList = term:term_functor_arg_list(Term),
		    match_list(PatternList, KeyList, Context, AS);
		false ->
		    {fail, AS}
	    end;
	_ ->
	    {fail, AS}
    end.
  
match_join_list(P, [Term | TermList],  Context, AS) ->
    case match(P, Term,  Context, AS) of
	{ok, AS1} ->
	    match_join_list(P, TermList, Context, AS1);
	{fail, AS1} ->
	    {fail, AS1}
    end;
match_join_list(_, [], _, AS) ->
    {ok, AS}.

match_list([P |PList], [Term | TermList],  Context, AS) ->
    case match(P, Term, Context, AS) of
	{ok, AS1} ->
	    match_list(PList, TermList, Context, AS1);
	{fail, AS1} ->
	    {fail, AS1}
    end;
match_list([], [], _, AS) ->
    {ok, AS}.

match_union(P, Term,  Context, AS) ->
    PList = spec:union_arg_list(P),
    match_union_list(PList, Term, Context, AS).


match_union_list([P | PList], Term, Context, AS) ->
    case match(P, Term, Context, AS) of
	{ok, AS2} ->
	    {ok, AS2};
	{fail, AS2} ->
	    match_union_list(PList, Term, Context, AS2)
    end;
match_union_list([], _, _, AS) ->
    {fail, AS}.

module_table(Module) ->
    ensure_present(Module),
    MT = get(warn_module_table),
    gb_trees:lookup(Module, MT).

ensure_present(Module) ->
    MT = get(warn_module_table),
    case gb_trees:is_defined(Module, MT) of
	true ->
	    true;
	false ->
	    Path = program:path(),
	    case sa_file:find_module(Path, Module, warn) of 
	       {ok, FileName} ->
		   load_warn_module(FileName);
		not_found ->
		    true
	   end
    end.


load_warn_module(FileName) ->
    case warn_front:read_file(FileName) of
	{warn_module, ModuleName, FunctionList} ->
	    FunctionTable = 
		lists:foldr(
		  fun({funspec, Name, Arity, DecList}, Table) ->
			  Fnid = program:make_fnid(ModuleName, Name, Arity),
			  ExeList = lists:map(fun({exception, P}) -> P end,
					      DecList),
			  gb_trees:insert(Fnid, ExeList, Table)
		  end,
		  gb_trees:empty(),
		  FunctionList),
	    MT = get(warn_module_table),
	    MT1 = gb_trees:insert(ModuleName, FunctionTable, MT),
	    put(warn_module_table, MT1);
	bad_module ->
	    exit({bad_warn_module, FileName})
    end.



