%%% File    : warn_front.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>
%%% Purpose : Read specifiations from file, parse and transform them.
%%% Created : 12 Jan 2000 by Sven-Olof Nystr|m <svenolof@harpo.it.uu.se>

-module(warn_front).
-author('svenolof@harpo.it.uu.se').

-export([read_file/1]).

%% open(FileName, IncludePath, PreDefMacros)
%% close(Epp)
%% scan_erl_form(Epp)
%% parse_erl_form(Epp)


open(File, Path, Pdm) ->
    Epp = spawn_link(epp, server, [self(),File,Path,Pdm]),
    wait_epp_reply(Epp).

close(Epp) ->
    epp_request(Epp, close),
    wait_epp_reply(Epp).

parse_erl_form(Epp) ->
    epp_request(Epp, scan_erl_form),
    case wait_epp_reply(Epp) of
	{ok,Toks} ->
	    parse_warn:parse_form(Toks);
	Other ->
	    Other
    end.


parse_file(Ifile, Path, Predefs) ->
    case open(Ifile, Path, Predefs) of
	{ok,Epp} ->
	    Forms = parse_file(Epp),
	    close(Epp),
	    Forms;
	{error,E} ->
	    [{error,E}]
    end.

%% parse_file(Epp) ->
%%	[Form]

parse_file(Epp) ->
    case parse_erl_form(Epp) of
	{ok,Form} ->
	    [Form|parse_file(Epp)];
	{error,E} ->
	    [{error,E}|parse_file(Epp)];
	{eof,Line} ->
	    [{eof,Line}]
    end.

%% epp_request(Epp, Request)
%% wait_epp_reply(Epp)
%%  Handle communication with the epp.

epp_request(Epp, Req) ->
    Epp ! {epp_request,self(),Req}.

wait_epp_reply(Epp) ->
    receive
	{epp_reply,Epp,Rep} -> Rep
    end.

%%% Specifications.



read_file(FileName) ->    
    L = parse_file(FileName, [], []),
    case lists:any(fun({error, _}) -> true;
		      (_) -> false
		      end,
		      L) of
	false ->
	    {Name, FunctionList} = process_warn_module(L),
	    {warn_module, Name, FunctionList};
	true ->
	    warn_errors(FileName, L),
	    bad_module
    end.

process_warn_module([_, {attribute, _, module, Name} | FormList]) ->
    FunctionList = 
	process_warn_module(FormList, []),
    {Name, FunctionList}.

process_warn_module([Form | FormList], FunSpecList) ->
    case element(1, Form) of
	funspec ->
	    Form1 = translate_form(Form),
	    process_warn_module(FormList, [Form1 | FunSpecList]);
	attribute ->
	    case element(3, Form) of
		file ->
		    process_warn_module(FormList, FunSpecList);
		module ->
		    process_warn_module(FormList, FunSpecList)
	    end;
	eof ->
	    process_warn_module(FormList, FunSpecList)
    end;
process_warn_module([], FunSpecList) ->
    lists:reverse(FunSpecList).

warn_errors(File, L) ->
    io:format("~nSyntax errors in file ~s", [File]),
    lists:foreach(fun({error, {Line, _, S}}) ->
		ok = io:format("~nLine ~w: ~s", [Line, S]);
	   (_) ->
		true
	 end,
	 L).

%%%% Translate. Given syntax tree grenerated by parser, produce
%%%% internal representation of types and type definitions.

translate_arg_list([TArg | TArgList]) ->
    Arg = translate_arg(TArg),
    ArgList = translate_arg_list(TArgList),
    [Arg | ArgList];
translate_arg_list([]) ->
    [].

translate_arg(Arg) ->
    translate_type(Arg).


translate_type({utype, TTypeList})  ->
    translate_union(TTypeList);
translate_type(Var) when atom(Var) ->
    spec:make_var(Var);
translate_type({atom, _, Atom}) ->
    spec:make_functor(atomc, [spec:make_functor({atoma, Atom}, [])]);
translate_type({integer, _, Int}) ->
    spec:make_functor(intc, [spec:make_functor({inta, Int}, [])]);
translate_type({tuple, TTypeList}) ->
    TypeList = lists:map(fun translate_type/1, TTypeList),
    spec:make_functor(tuple, [list_to_spec_list(TypeList)]);
translate_type({cons, T1, T2}) ->
    S1 = translate_type(T1),
    S2 = translate_type(T2),
    spec:make_functor(cons,[S1,S2]);
translate_type({cons, _, T1, T2}) ->
    S1 = translate_type(T1),
    S2 = translate_type(T2),
    spec:make_functor(cons,[S1,S2]);
translate_type({nil, _}) ->
    spec:make_functor(nil, []);
translate_type({type, Name, TArgList}) ->
    ArgList = translate_arg_list(TArgList),
    case is_special_struct(Name) of
	true ->
	    translate_special(Name, ArgList);
	false ->
	    spec:make_apply(Name, ArgList, next_apply_pos())
    end;
translate_type({external, Module, Name, TArgList}) ->
    ArgList = translate_arg_list(TArgList),
    spec:make_external(Module, Name, ArgList, next_apply_pos()).
    
list_to_spec_list([Type | List]) ->
    spec:make_functor(cons, [Type, list_to_spec_list(List)]);
list_to_spec_list([]) ->
    spec:make_functor(nil, []).



%%% A auick hack to make expansion of type application work.

next_apply_pos() ->
    case get(apply_pos) of
	undefined ->
	    put(apply_pos,1),
	    0;
	N ->
	    put(apply_pos, N+1),
	    N
    end.

%%%

translate_union([TType]) ->
    translate_type(TType);
translate_union(L) ->
    spec:make_union_spec(lists:map(fun translate_type/1, L)).
					  
is_special_struct(Name) ->					  
    lists:member(Name, [int, float, atom, any, none]).

translate_special(atom, []) ->
    spec:make_functor(atomc, [spec:make_universal()]);
translate_special(int, []) ->
    spec:make_functor(intc, [spec:make_universal()]);
translate_special(float, []) ->
    spec:make_functor(floatc, [spec:make_universal()]);
translate_special(any, []) ->
    spec:make_universal();
translate_special(none, []) ->
    spec:make_union_spec([]).


translate_form(Form) ->
	case Form of
	    {funspec, _Lineno, Name, Arity, Body} ->
		NewBody = lists:map(fun translate_point/1, Body),
		{funspec, Name, Arity, NewBody}
	end.

translate_point({point, _Lineno, exception, Type}) ->
    NewType = translate_type(Type),
    {exception, NewType};
translate_point({point, _Lineno, exit, Type}) ->
    NewType = translate_type(Type),
    Exit = spec:make_functor(atomc, [spec:make_functor({atoma,'EXIT'},[])]),
    WithExit = 
	spec:make_functor(
	  tuple,
	  [spec:make_functor(
	     cons,
	     [Exit, 
	      spec:make_functor(
		cons,
		[NewType,
		 spec:make_functor(nil,[])]
		)]
	     )]),
    {exception, WithExit};
translate_point({point, Lineno, K, _Type}) ->
    exit({line, Lineno, bad_keyword_in_warn_file, K});
translate_point(P) ->
    exit({bad_declaration_in_warn_file, P}).





